/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ifs_keywords.h"
#include "sph_dataset.h"
#include "sph_dataset_stats.h"
#include "sph_ifs_astrometry.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ifs_tags.h"
#include "sph_ldt.h"
#include "sph_framecombination.h"
#include "sph_framesorter.h"
#include "sph_keyword_manager.h"
#include "sph_create_super_flat.h"
#include "sph_filemanager.h"
#include "sph_ifs_science.h"
#include <strings.h>
/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_ASTROMETRY_GENERAL;
extern sph_error_code SPH_IFS_ASTROMETRY_FRAMES_MISSING;
const double SPH_IFS_ASTROMETRY_DIST_TOLERANCE = 4.0;
const double SPH_IFS_ASTROMETRY_ESTIMATED_PIXEL_SCALE = 0.01;
const double SPH_IFS_ASTROMETRY_NORTH_DR_ANGLE = 0.0;

static
cpl_propertylist*
sph_ird_astrometry_find_companion(
        sph_master_frame* mframe,
        double sigma,
        double min_radius,
        double max_radius ) {
    cpl_apertures*            aps = NULL;
    cpl_image*                image = NULL;
    int                        nobs = 0;
    int                        nn     = 0;
    int                        ngoodobs = 0;
    int                        idbrightest = 0;
    double                    x = 0.0;
    double                    y = 0.0;
    double                    cx = 0.0;
    double                    cy = 0.0;
    double                    flux = 0.0;
    cpl_vector*                idgoods = NULL;
    cpl_propertylist*        plist  = NULL;
    SPH_RAISE_CPL_RESET
    image = sph_master_frame_get_image(mframe,1);
    cpl_image_fill_rejected(image,0.0);
    cx = cpl_image_get_size_x(image)/2.0;
    cy = cpl_image_get_size_y(image)/2.0;
    idgoods = cpl_vector_new(1);
    if ( image ) {
        plist = cpl_propertylist_new();
        aps = cpl_apertures_extract_sigma(image,sigma);
        if ( aps ) nobs = cpl_apertures_get_size(aps);
        if ( nobs < 1 ) {
            SPH_ERROR_RAISE_ERR( CPL_ERROR_DATA_NOT_FOUND,
                    "Found no object "
                    "on the reduced image using a sigma of %f.",
                    sigma );
            cpl_propertylist_delete(plist); plist = NULL;
            cpl_vector_delete(idgoods); idgoods = NULL;
            cpl_image_delete(image); image = NULL;
            return NULL;
        }
        for (nn = 0; nn < nobs; ++nn) {
            x = cpl_apertures_get_centroid_x(aps,nn+1);
            y = cpl_apertures_get_centroid_y(aps,nn+1);
            flux = cpl_apertures_get_flux(aps,nn+1);
            if ( (x-cx)*(x-cx)+(y-cy)*(y-cy) <
                    max_radius * max_radius &&
                 (x-cx)*(x-cx)+(y-cy)*(y-cy) >
                    min_radius * min_radius )
            {
                cpl_vector_set(idgoods,ngoodobs,nn+1);
                ngoodobs++;
                if ( idbrightest < 1 ) idbrightest = nn+1;
                else if ( cpl_apertures_get_flux(aps,idbrightest) < flux ) {
                    idbrightest = nn+1;
                }
                cpl_vector_set_size(idgoods,ngoodobs);
            }
        }
        if ( ngoodobs < 1 ) {
            SPH_ERROR_RAISE_ERR( CPL_ERROR_DATA_NOT_FOUND,
                    "Found no object "
                    "on the reduced image using a sigma of %f at"
                    " the radius between %f and %f.",
                    sigma, min_radius, max_radius );
            cpl_propertylist_delete(plist); plist = NULL;
            cpl_vector_delete(idgoods); idgoods = NULL;
            cpl_image_delete(image); image = NULL;
            return NULL;
        }
        if ( ngoodobs > 2 ) {
            SPH_ERROR_RAISE_INFO( CPL_ERROR_INCOMPATIBLE_INPUT,
                    "Found %d objects "
                    "on the reduced image using a sigma of %f at"
                    " distance of less than %f from center. Am assuming the brightest one "
                    " (at coords %f,%f) "
                    "is the companion to use for flux calibration.",
                    ngoodobs, sigma, max_radius,
                    cpl_apertures_get_centroid_x(aps,idbrightest),
                    cpl_apertures_get_centroid_y(aps,idbrightest));
        }
        cpl_propertylist_append_double(plist,SPH_COMMON_KEYWORD_SPH_COMANION_X,
                cpl_apertures_get_centroid_x(aps,idbrightest)-cx);
        cpl_propertylist_append_double(plist,SPH_COMMON_KEYWORD_SPH_COMANION_Y,
                cpl_apertures_get_centroid_y(aps,idbrightest)-cy);
    }
    SPH_RAISE_CPL;
    return plist;
}


static
cpl_propertylist*
sph_ifs_astrometry_solve( sph_ifs_astrometry* self,
        sph_master_frame* masterframe) {
    cpl_propertylist*        pl = NULL;
    double                    rmin = 0.0;
    double                    rmax = 0.0;
    double                    r = 0.0;
    double                    rin = 0.0;
    double                    dr_ang = 0.0; //later make parameter!
    double                    measured_ang = 0.0;
    double                    xc = 0.0;
    double                    yc = 0.0;
    double                    scale = 0.0;
    sph_error_code            rerr = CPL_ERROR_NONE;

    rin = sqrt(self->companion_x * self->companion_x + self->companion_y * self->companion_y);
    r = rin / SPH_IFS_ASTROMETRY_ESTIMATED_PIXEL_SCALE;
    rmin = r - SPH_IFS_ASTROMETRY_DIST_TOLERANCE;
    rmax = r + SPH_IFS_ASTROMETRY_DIST_TOLERANCE;
    pl = sph_ird_astrometry_find_companion(masterframe,10.0,rmin,rmax);
    if ( pl ) {
        xc = cpl_propertylist_get_double(pl,SPH_COMMON_KEYWORD_SPH_COMANION_X);
        yc = cpl_propertylist_get_double(pl,SPH_COMMON_KEYWORD_SPH_COMANION_Y);
        measured_ang = atan2( yc, xc ) / CPL_MATH_PI * 180.0 - dr_ang;
        if ( measured_ang < 0.0 ) measured_ang = 360.0 + measured_ang;
        rerr = cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_SPH_ANGLE_TO_NORTH, measured_ang);
        if ( rerr ) SPH_RAISE_CPL;
        if ( xc && yc ) scale = 0.5 *( self->companion_x / xc + self->companion_y / yc );
        rerr = cpl_propertylist_update_double(pl, SPH_COMMON_KEYWORD_SPH_PLATE_SCALE, scale);
        if ( rerr ) SPH_RAISE_CPL;

    }

    return pl;
}
/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_astrometry_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_astrometry.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
cpl_error_code
sph_ifs_astrometry_run( sph_ifs_astrometry* self )
{
    sph_ldt*        ldt    = NULL;
    sph_master_frame*                    result = NULL;
    sph_ifs_lenslet_model*    lensmodel = NULL;
    cpl_propertylist*        plist = NULL;
    cpl_propertylist*        pl = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE

    plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->wave_calib_frame),0);
    lensmodel = sph_ifs_lenslet_model_new_from_propertylist(plist);

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE
    if ( self->dither == 0 ) {
        ldt = sph_ifs_science_reduce_no_dither_no_adi(
                self->raw_science_frameset,
                self->coll_alg,
                self->framecomb_parameterlist,
                lensmodel,
                self->master_dark_frame,NULL,
                self->wave_calib_frame,NULL,
                self->master_ifu_flat_frame,
                self->master_dff_long1_frame,
                self->master_dff_long2_frame,
                self->master_dff_long3_frame,
                self->master_dff_long4_frame,
                self->master_dff_longbb_frame,
                self->master_dff_short_frame );
        result = sph_ldt_collapse(ldt,0.0,-1.0);
        sph_ldt_delete(ldt); ldt = NULL;
    }
    else {
        ldt = sph_ifs_science_reduce_dither_no_adi(
                self->raw_science_frameset,
                self->coll_alg,
                self->framecomb_parameterlist,
                lensmodel,
                self->master_dark_frame,NULL,
                self->wave_calib_frame,NULL,
                self->master_ifu_flat_frame,
                self->master_dff_long1_frame,
                self->master_dff_long2_frame,
                self->master_dff_long3_frame,
                self->master_dff_long4_frame,
                self->master_dff_longbb_frame,
                self->master_dff_short_frame );
        result = sph_ldt_collapse(ldt,0.0,-1);
        sph_ldt_delete(ldt); ldt = NULL;
    }
    SPH_RAISE_CPL_RESET
    if ( result ) {
        pl = sph_ifs_astrometry_solve( self, result );
        if ( !pl ) {
            SPH_RAISE_CPL_RESET;
        }
        sph_master_frame_save_dfs(result, self->astrometry_filename, self->inframes, NULL,
                self->inparams,
                SPH_IFS_TAG_ASTROMETRY_CALIB,
                SPH_RECIPE_NAME_IFS_ASTROMETRY,
                SPH_PIPELINE_NAME_IFS ,pl );
        sph_master_frame_delete( result ); result = NULL;
        cpl_propertylist_delete(pl); pl = NULL;
    }

    return cpl_error_get_code();
}
/**@}*/
