/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ifs_keywords.h"
#include "sph_dataset.h"
#include "sph_dataset_stats.h"
#include "sph_ifs_sky_cal.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_ifs_tags.h"
#include "sph_ldt.h"
#include "sph_framecombination.h"
#include "sph_framesorter.h"
#include "sph_keyword_manager.h"
#include "sph_create_super_flat.h"
#include "sph_filemanager.h"
#include "sph_ifs_science.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "sph_fitting.h"
#include "sph_fits.h"

/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IFS_SKY_CAL_GENERAL;
extern sph_error_code SPH_IFS_SKY_CAL_FRAMES_MISSING;
const int SPH_IFS_SKY_CAL_MAX_TRIES = 1000;

static
sph_master_frame*
sph_ifs_create_polyfit(
        sph_master_frame* mframe, int order, int nsamples ) {
    sph_master_frame*            polframe = NULL;
    int                            nx    = 0;
    int                            ny    = 0;
    cpl_vector*                    xvals = NULL;
    cpl_vector*                    yvals = NULL;
    cpl_vector*                    zvals = NULL;
    int                            xx = 0;
    int                            yy = 0;
    double                        val = 0.0;
    gsl_rng*                    pRNG = NULL;
    int                            cc    = 0;
    int                            bpix    = 0;
    cpl_polynomial*                poly = NULL;
    cpl_vector*                    vv = NULL;
    int                            ntries = 0;
    nx = cpl_image_get_size_x(mframe->image);
    ny = cpl_image_get_size_y(mframe->image);

    pRNG = gsl_rng_alloc(gsl_rng_taus);
    if ( !pRNG ) return NULL;
    polframe = sph_master_frame_duplicate(mframe);
    if ( !polframe ) return NULL;
    xvals = cpl_vector_new(nsamples);
    yvals = cpl_vector_new(nsamples);
    zvals = cpl_vector_new(nsamples);
    if ( !xvals || !yvals || !zvals ) return NULL;

    cpl_image_delete(polframe->ncombmap);
    polframe->ncombmap = cpl_image_new(nx,ny,CPL_TYPE_DOUBLE);
    cpl_image_delete(polframe->badpixelmap);
    polframe->badpixelmap = cpl_image_new(nx,ny,CPL_TYPE_INT);

    xx = (int)gsl_ran_flat(pRNG,0.0,nx);
    yy = (int)gsl_ran_flat(pRNG,0.0,ny);


    while ( cc < nsamples )
    {
        xx = (int)gsl_ran_flat(pRNG,0.0,nx);
        yy = (int)gsl_ran_flat(pRNG,0.0,ny);
        ntries = 0;
        while ( cpl_image_get(mframe->badpixelmap,xx + 1,yy + 1,&bpix) > 0 &&
                ntries < SPH_IFS_SKY_CAL_MAX_TRIES )
        {
            xx = (int)gsl_ran_flat(pRNG,0.0,nx);
            yy = (int)gsl_ran_flat(pRNG,0.0,ny);
            ntries++;
        }
        if ( ntries < SPH_IFS_SKY_CAL_MAX_TRIES ) {
            val = cpl_image_get(mframe->image,xx + 1,yy + 1,&bpix);
            cpl_vector_set(xvals,cc,xx);
            cpl_vector_set(yvals,cc,yy);
            cpl_vector_set(zvals,cc,val);
            val = cpl_image_get(polframe->ncombmap,xx + 1,yy + 1,&bpix);
            cpl_image_set( polframe->ncombmap, xx + 1, yy + 1,val + 1);
            cc++;
        }
        else {
            cc = nsamples;
        }
    }

    poly = sph_fitting_fit_poly2d(xvals,yvals,zvals,order,order);
    if ( poly ) {
        vv = cpl_vector_new(2);
        for (xx = 0; xx < nx; ++xx) {
            for (yy = 0; yy < ny; ++yy) {
                cpl_vector_set(vv,0,xx);
                cpl_vector_set(vv,1,yy);
                val = cpl_polynomial_eval(poly,vv);
                cpl_image_set( polframe->image, xx + 1, yy + 1,val);
                val = val - cpl_image_get(mframe->image,xx + 1,yy + 1,&bpix);
                val = val * val;
                cpl_image_set( polframe->rmsmap, xx + 1, yy + 1,val);
            }
        }
        cpl_vector_delete(vv);vv=NULL;
    }
    SPH_RAISE_CPL_RESET;
    return polframe;
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ifs_sky_cal_run Run the Instrument Flat Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * instrument flat recipe for IFS.
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ifs_sky_cal.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/
cpl_error_code
sph_ifs_sky_cal_run( sph_ifs_sky_cal* self )
{
    sph_ldt*        result    = NULL;
    sph_ifs_lenslet_model*    lensmodel = NULL;
    cpl_propertylist*        plist = NULL;
    int                        nplanes = 0;
    int                        pp        = 0;
    sph_master_frame*        aplane    = 0;
    sph_master_frame*        fitplane = 0;
    cpl_imagelist*           imlist = NULL;

    plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->wave_calib_frame),0);
    lensmodel = sph_ifs_lenslet_model_new_from_propertylist(plist);

    SPH_RAISE_CPL_RESET
    if ( self->use_adi == 0 && self->dither == 0 ) {
        result = sph_ifs_science_reduce_no_dither_no_adi(
                self->raw_science_frameset,
                self->coll_alg,
                self->framecomb_parameterlist,
                lensmodel,
                self->master_dark_frame,NULL,
                self->wave_calib_frame,NULL,
                self->master_ifu_flat_frame,
                self->master_dff_long1_frame,
                self->master_dff_long2_frame,
                self->master_dff_long3_frame,
                self->master_dff_long4_frame,
                self->master_dff_longbb_frame,
                self->master_dff_short_frame );
    }
    else if ( self->use_adi == 0 && self->dither == 1 ) {
        result = sph_ifs_science_reduce_dither_no_adi(
                self->raw_science_frameset,
                self->coll_alg,
                self->framecomb_parameterlist,
                lensmodel,
                self->master_dark_frame,NULL,
                self->wave_calib_frame,NULL,
                self->master_ifu_flat_frame,
                self->master_dff_long1_frame,
                self->master_dff_long2_frame,
                self->master_dff_long3_frame,
                self->master_dff_long4_frame,
                self->master_dff_longbb_frame,
                self->master_dff_short_frame );
    }

    if ( result ) {
        sph_ldt_save_cube(
                result, self->sky_cal_filename,
                self->inframes, NULL, self->inparams,
                SPH_IFS_TAG_SKY_CAL_CALIB,
                SPH_RECIPE_NAME_IFS_SKY_CAL,
                SPH_PIPELINE_NAME_IFS, plist );
        nplanes = cpl_vector_get_size(result->reflambdas);
        sph_ldt_delete( result ); result = NULL;
        for (pp = 0; pp < nplanes; ++pp) {
            aplane = sph_master_frame_load(self->sky_cal_filename,pp);
            fitplane = sph_ifs_create_polyfit(aplane,self->order,1000);
            if ( pp == 0 ) {
                imlist = cpl_imagelist_new();
                cpl_propertylist_update_string( plist,
                        SPH_COMMON_KEYWORD_PRO_CATG, SPH_IFS_TAG_SKY_CAL_FIT_CALIB );
                cpl_imagelist_set(imlist,cpl_image_duplicate(fitplane->image),0);
                cpl_dfs_save_imagelist(self->inframes,NULL,self->inparams,
                        self->inframes,NULL,imlist,
                        CPL_TYPE_DOUBLE,
                        SPH_RECIPE_NAME_IFS_SKY_CAL,plist,
                        NULL,SPH_PIPELINE_NAME_IFS,self->fit_filename);
                cpl_imagelist_delete(imlist);imlist = NULL;
                imlist = cpl_imagelist_new();
                cpl_imagelist_set(imlist,cpl_image_duplicate(fitplane->badpixelmap),0);
                cpl_imagelist_save(imlist,self->fit_filename,
                        CPL_BPP_8_UNSIGNED,plist,CPL_IO_EXTEND);
                cpl_imagelist_delete(imlist);imlist = NULL;
                imlist = cpl_imagelist_new();
                cpl_imagelist_set(imlist,cpl_image_duplicate(fitplane->ncombmap),0);
                cpl_imagelist_save(imlist,self->fit_filename,
                        CPL_TYPE_DOUBLE,plist,CPL_IO_EXTEND);
                cpl_imagelist_delete(imlist);imlist = NULL;
                imlist = cpl_imagelist_new();
                cpl_imagelist_set(imlist,cpl_image_duplicate(fitplane->rmsmap),0);
                cpl_imagelist_save(imlist,self->fit_filename,
                        CPL_TYPE_DOUBLE,plist,CPL_IO_EXTEND);
                cpl_imagelist_delete(imlist);imlist = NULL;
            }
            else {
                sph_cube_append_image(
                        self->fit_filename,
                        fitplane->image,
                        NULL,
                        SPH_MASTER_FRAME_IMAGE_EXT);
                sph_cube_append_image(
                        self->fit_filename,
                        fitplane->badpixelmap,
                        NULL,
                        SPH_MASTER_FRAME_BPIX_EXT);
                sph_cube_append_image(
                        self->fit_filename,
                        fitplane->ncombmap,
                        NULL,
                        SPH_MASTER_FRAME_NC_EXT);
                sph_cube_append_image(self->fit_filename,
                        fitplane->rmsmap,
                        NULL,
                        SPH_MASTER_FRAME_RMS_EXT);
            }
        }
    }
    return cpl_error_get_code();
}
/**@}*/
