/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include "sph_common_keywords.h"
#include "sph_ird_keywords.h"
#include "sph_ird_tags.h"
#include "sph_ird_distortion_map.h"
#include "sph_ird_instrument_model.h"
#include "sph_master_frame.h"
#include "sph_error.h"
#include "sph_keyword_manager.h"
#include "sph_cube.h"
#include "sph_utils.h"
#include "sph_filemanager.h"
#include "sph_polynomial_distortion_model.h"
#include "sph_framecombination.h"
#include "sph_point_pattern.h"
#include "sph_transform.h"

#define SPH_IRD_DISTORTION_MAP_POINT_LIMIT 20
/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/

extern sph_error_code SPH_IRD_DISTORTION_MAP_GENERAL;
extern sph_error_code SPH_IRD_DISTORTION_MAP_FRAMES_MISSING;

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_input_pattern(
        sph_point_pattern* pp,sph_ird_instrument_model* model,
        double opt_cent_l_x, double opt_cent_l_y,
        double opt_cent_r_x, double opt_cent_r_y);

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_distortion(
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_ird_instrument_model* model,
        cpl_propertylist* qclist);

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_pp(
        sph_ird_distortion_map* self,
        sph_ird_instrument_model* model,
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_point_pattern* obs_left,
        sph_point_pattern* obs_right);
static
sph_error_code
__sph_ird_distortion_map_create_qc_of_accuracy(
        sph_ird_distortion_map * self,
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_point_pattern* ppin);
static
sph_master_frame*
create_science( sph_ird_distortion_map* self,
        sph_master_frame* dark,
        sph_master_frame* flat,
        cpl_frameset* rawframes )
{
    sph_master_frame*                mframe = NULL;

    mframe = sph_framecombination_master_frame_from_cpl_frameset(
            rawframes,
            self->coll_alg,
            self->framecomb_parameterlist);

    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
    	sph_master_frame_delete(mframe); mframe = NULL;
    }

    if ( !mframe ) return NULL;

    if ( dark ) sph_master_frame_subtract_master_frame(mframe,dark);
    if ( flat ) sph_master_frame_divide_master_frame(mframe,flat);

    if ( sph_master_frame_interpolate_bpix(mframe) == CPL_ERROR_NONE ) {
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Interpolating badpixels on reduced science image.");
    }
    else {
        SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                "No interpolation of badpixels possible.");
        cpl_error_reset();
    }
    return mframe;
}


/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_ird_distortion_map_run Create Master Dark Recipe
 *
 * This module provides the algorithm inplementation for the creation of the
 * master dakr
 *
 * @par Synopsis:
 * @code
 *   #include "sph_ird_distortion_map.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_ird_distortion_map recipe. The error
 code returned is always a cpl error code (to allow maximal compatibility with
 esorex, gasgano, etc.) even if during recipe execution an error in the SPHERE
 API is the cause. In this case (and if the underlying error is not a cpl error)
 the cpl error code is set to the cpl_error_code that matches the failure
 reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
cpl_error_code sph_ird_distortion_map_run( sph_ird_distortion_map* self )
{
    sph_master_frame*           dark         = NULL;
    sph_master_frame*            flat     = NULL;
    sph_master_frame*            result = NULL;
    sph_master_frame*            result_left = NULL;
    sph_master_frame*            result_right = NULL;
    sph_point_pattern*            point_pattern_left = NULL;
    sph_point_pattern*            point_pattern_right = NULL;
    sph_point_pattern*            point_pattern = NULL;
    sph_point_pattern*            obs_pp_left = NULL;
    sph_point_pattern*            obs_pp_right = NULL;
    sph_polynomial_distortion_model*    distmap_left = NULL;
    sph_polynomial_distortion_model*    distmap_right = NULL;
    sph_ird_instrument_model*   model = NULL;
    cpl_image*                  imleft = NULL;
    cpl_image*                  imright = NULL;
    cpl_propertylist*           plist = NULL;
    cpl_propertylist*           qclist = NULL;
    cpl_image*                  qcim = NULL;
    cpl_image*                  qcim_l = NULL;
    cpl_image*                  qcim_r = NULL;
    double                      opt_cent_l_x = -10000.0;
    double                      opt_cent_l_y = -10000.0;
    double                      opt_cent_r_x = -10000.0;
    double                      opt_cent_r_y = -10000.0;
    sph_transform*              trans = NULL;

    /*------------------------------------------------------------------
    -  Placing master dark in cube structure for later use
    --------------------------------------------------------------------*/


    /*------------------------------------------------------------------
     -  Now process all sets of wavelength frames separately
     --------------------------------------------------------------------*/
    if ( self->dark_frame )
    {
        dark = sph_master_frame_load( cpl_frame_get_filename(self->dark_frame),0);
    }
    if ( self->flat_frame )
    {
        flat = sph_master_frame_load( cpl_frame_get_filename(self->flat_frame),0);
        plist = sph_keyword_manager_load_properties(cpl_frame_get_filename(self->flat_frame),0);
        model = sph_ird_instrument_model_new_from_propertylist(plist);
    }
    else {
        model = sph_ird_instrument_model_new();
        plist = sph_ird_instrument_model_get_as_propertylist(model);
    }
    result = create_science(self,dark,flat,self->rawframes);

    if ( self->user_cent ) {
        opt_cent_l_x = self->cent_left_x;
        opt_cent_l_y = self->cent_left_y;
        opt_cent_r_x = self->cent_right_x;
        opt_cent_r_y = self->cent_right_y;
    }
    if ( result ) {
        if ( !self->point_pattern_frame ) {
            SPH_INFO_MSG("No point pattern provided. "
                    "Cant proceed to create distortion map and will"
                    " create point pattern from raw images instead.");
            result_left = sph_ird_instrument_model_extract_left_master_frame(model,result);
            imleft = sph_master_frame_get_image(result_left,1);
            result_right = sph_ird_instrument_model_extract_right_master_frame(model,result);
            imright = sph_master_frame_get_image(result_right,1);
            point_pattern_left = sph_point_pattern_new_from_image(imleft,NULL,self->threshold_sigma,3.0);
            point_pattern_right = sph_point_pattern_new_from_image(imright,NULL,self->threshold_sigma,3.0);
            cpl_image_delete(imleft); imleft = NULL;
            cpl_image_delete(imright); imright = NULL;
            if ( !point_pattern_left && !point_pattern_right ) {
                SPH_ERR("Could not create point pattern. Cant proceed.");
                goto EXIT;
            }
            if ( !point_pattern_left && point_pattern_right ) {
                SPH_INFO_MSG("Could not create left point pattern. "
                        "Am using only right point pattern.");
                point_pattern = point_pattern_right;
            }
            if ( point_pattern_left && !point_pattern_right ) {
                SPH_INFO_MSG("Could not create right point pattern. "
                        "Am using only left point pattern.");
                point_pattern = point_pattern_left;
            }
            if ( point_pattern_left && point_pattern_right ) {
                if ( sph_point_pattern_get_size(point_pattern_left) <= sph_point_pattern_get_size(point_pattern_right) ) {
                    SPH_INFO_MSG("Am taking the left point pattern as point pattern to save.");
                    point_pattern = point_pattern_left;
                    sph_point_pattern_delete(point_pattern_right);
                }
                else {
                    SPH_INFO_MSG("Am taking the right point pattern as point pattern to save.");
                    point_pattern = point_pattern_right;
                    sph_point_pattern_delete(point_pattern_left);
                }
            }

            sph_point_pattern_centralise_wrt_central_point(point_pattern);
            sph_point_pattern_save_dfs(point_pattern,
                    self->pp_outfilename,
                    self->inframes,
                    NULL,
                    self->inparams,
                    SPH_IRD_TAG_POINT_PATTERN,
                    SPH_RECIPE_NAME_IRD_DISTORTION_MAP,
                    SPH_PIPELINE_NAME_IRDIS, NULL );
        }
        if ( point_pattern == NULL ) {
            point_pattern = sph_point_pattern_new_from_table(self->point_pattern_frame,1);
            if ( point_pattern == NULL ) {
                cpl_error_reset();
                point_pattern = sph_point_pattern_load_ascii(cpl_frame_get_filename(self->point_pattern_frame));
            }
        }
        if ( !point_pattern ) {
            SPH_ERR("Could not load point pattern. Cant proceed.");
            goto EXIT;
        }


        if ( sph_point_pattern_get_size(point_pattern) < SPH_IRD_DISTORTION_MAP_POINT_LIMIT ) {
            SPH_ERR("Point pattern has too few points. Cant proceed.");
            goto EXIT;
        }
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Determining distortion of left FOV.");
        if ( !result_left ) result_left = sph_ird_instrument_model_extract_left_master_frame(model,result);
        qclist = cpl_propertylist_new();

        cpl_propertylist_update_int(qclist,
                SPH_IRD_KEYWORD_DISTMAP_NPOINTS_IN,
                sph_point_pattern_get_size(point_pattern));

        distmap_left = sph_polynomial_distortion_model_from_grid_image(
                result_left,
                point_pattern,
                self->polyfit_order,
                self->threshold_sigma,
                self->threshold_length,
                qclist,
                &opt_cent_l_x,
                &opt_cent_l_y,
                &obs_pp_left );


        if ( distmap_left ) {
            cpl_propertylist_update_int(qclist,
                    SPH_IRD_KEYWORD_DISTMAP_NPOINTS_OBS,
                    sph_point_pattern_get_size(obs_pp_left));

            cpl_propertylist_update_double(
                    qclist,
                    SPH_IRD_KEYWORD_DISTMAP_OPTICAL_AXIS_X,
                    opt_cent_l_x);
            cpl_propertylist_update_double(qclist,
                    SPH_IRD_KEYWORD_DISTMAP_OPTICAL_AXIS_Y,
                    opt_cent_l_y);

            sph_polynomial_distortion_model_save_dfs(
            		distmap_left,
                    self->outfilename,
                    self->inframes,
                    NULL,
                    self->inparams,
                    SPH_IRD_TAG_DISTORTION_MAP_CALIB,
                    SPH_RECIPE_NAME_IRD_DISTORTION_MAP,
                    SPH_PIPELINE_NAME_IRDIS,
                    qclist,
                    cpl_image_get_size_x(result_left->image),
                    cpl_image_get_size_y(result_left->image), 1.0 );

            SPH_ERROR_RAISE_INFO(
                    SPH_ERROR_GENERAL,
                    "Creating QC image by "
                    "applying distortionmap to input image itself.");

            trans = sph_transform_new_cpl_warp();
            if ( sph_transform_apply_to_images(trans,
                    &result_left->image,
                    NULL, NULL, NULL, NULL,
                    opt_cent_l_x, opt_cent_l_y,
                    0.0, 1.0, model) == CPL_ERROR_NONE) {
                SPH_ERROR_RAISE_INFO(
                        SPH_ERROR_GENERAL,
                        "Shifted QC image to %f, %f before "
                        "applying distortion.",
                        opt_cent_l_x, opt_cent_l_y);

            }
            else {
                SPH_ERROR_RAISE_WARNING(
                        cpl_error_get_code(),
                        "Could not shift QC image: %s",
                        cpl_error_get_message());
                cpl_error_reset();
            }

            sph_transform_delete(trans);
            qcim_l = sph_polynomial_distortion_model_apply(
                    distmap_left,
                            result_left->image,
                            CPL_KERNEL_DEFAULT,
                            2 * CPL_KERNEL_DEF_WIDTH);
            SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                    "Left distortion map saved.");
        }


        cpl_propertylist_delete(qclist); qclist = NULL;
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Determining distortion of right FOV.");
        if ( !result_right ) result_right = sph_ird_instrument_model_extract_right_master_frame(model,result);
        qclist = cpl_propertylist_new();

        cpl_propertylist_update_int(qclist,
                SPH_IRD_KEYWORD_DISTMAP_NPOINTS_IN,
                sph_point_pattern_get_size(point_pattern));

        distmap_right = sph_polynomial_distortion_model_from_grid_image(
                result_right,
                point_pattern,
                self->polyfit_order,
                self->threshold_sigma,
                self->threshold_length,
                qclist,
                &opt_cent_r_x, &opt_cent_r_y, &obs_pp_right );

        if ( distmap_right ) {
            cpl_propertylist_update_int(qclist,
                    SPH_IRD_KEYWORD_DISTMAP_NPOINTS_OBS,
                    sph_point_pattern_get_size(obs_pp_right));

            cpl_propertylist_update_double(
                    qclist,
                    SPH_IRD_KEYWORD_DISTMAP_OPTICAL_AXIS_X,
                    opt_cent_r_x);
            cpl_propertylist_update_double(qclist,
                    SPH_IRD_KEYWORD_DISTMAP_OPTICAL_AXIS_Y,
                    opt_cent_r_y);

            sph_polynomial_distortion_model_save(
                    distmap_right,
                    self->outfilename,
                    cpl_image_get_size_x(result_right->image),
                    cpl_image_get_size_y(result_right->image), CPL_IO_EXTEND,
                    qclist );

            SPH_ERROR_RAISE_INFO(
                    SPH_ERROR_GENERAL,
                    "Creating QC image by "
                    "applying distortionmap to input image itself.");

            trans = sph_transform_new_cpl_warp();
            if ( sph_transform_apply_to_images(trans,
                    &result_right->image,
                    NULL, NULL, NULL, NULL,
                    opt_cent_r_x, opt_cent_r_y,
                    0.0, 1.0, model) == CPL_ERROR_NONE)
            {
                SPH_ERROR_RAISE_INFO(
                        SPH_ERROR_GENERAL,
                        "Shifted QC image to %f, %f before "
                        "applying distortion.",
                        opt_cent_r_x, opt_cent_r_y);

            }
            else {
                SPH_ERROR_RAISE_WARNING(
                        cpl_error_get_code(),
                        "Could not shift QC image: %s",
                        cpl_error_get_message());
                cpl_error_reset();
            }

            sph_transform_delete(trans);

            qcim_r = sph_polynomial_distortion_model_apply(
                    distmap_right,
                            result_right->image,
                            CPL_KERNEL_DEFAULT,
                            CPL_KERNEL_DEF_WIDTH);

        }

        if ( self->full_qc ) {
            __sph_ird_distortion_map_create_qc_of_input_pattern(
                    point_pattern,
                    model,
                    opt_cent_l_x, opt_cent_l_y,
                    opt_cent_r_x, opt_cent_r_y);

            __sph_ird_distortion_map_create_qc_of_distortion(
                    qcim_l, qcim_r, model,
                    qclist);

            __sph_ird_distortion_map_create_qc_of_pp(
                    self,
                    model,
                    qcim_l,
                    qcim_r,
                    obs_pp_left, obs_pp_right);

            __sph_ird_distortion_map_create_qc_of_accuracy(
                    self,
                    qcim_l,
                    qcim_r,
                    point_pattern);
        }
EXIT:
        cpl_image_delete(qcim_l); qcim_l = NULL;
        cpl_image_delete(qcim_r); qcim_r = NULL;
        cpl_image_delete(qcim); qcim = NULL;


        cpl_propertylist_delete(qclist); qclist = NULL;
        sph_point_pattern_delete(point_pattern); point_pattern = NULL;
        sph_master_frame_delete(result);result=NULL;
        sph_master_frame_delete(result_left);result_left=NULL;
        sph_master_frame_delete(result_right);result_right=NULL;
    }

     if ( dark ) {
        sph_master_frame_delete(dark); dark = NULL;
    }
    if ( flat ) {
        sph_master_frame_delete(flat); flat = NULL;
    }
    sph_polynomial_distortion_model_delete(distmap_left); distmap_left = NULL;
    sph_polynomial_distortion_model_delete(distmap_right); distmap_right = NULL;
    sph_ird_instrument_model_delete(model); model = NULL;
    cpl_propertylist_delete(plist); plist = NULL;
    sph_point_pattern_delete(obs_pp_left);
    sph_point_pattern_delete(obs_pp_right);
    //sph_filemanager_clean();
    return (int)cpl_error_get_code();
}

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_input_pattern(
        sph_point_pattern* pp,sph_ird_instrument_model* model,
        double opt_cent_l_x, double opt_cent_l_y,
        double opt_cent_r_x, double opt_cent_r_y)
{
    cpl_image*          testim_left = NULL;
    cpl_image*          testim_right = NULL;
    cpl_image*          testim = NULL;
    sph_point*          point = NULL;
    double              dx = 0.0;
    double              dy = 0.0;
    point = sph_point_pattern_get_point(
            pp,
            sph_point_pattern_get_central_point(pp,
                    &dx,
                    &dy) );
    sph_point_pattern_add_offset(
            pp,
            model->window_size_y / 2 - point->x,
            model->window_size_y / 2 - point->y);

    testim_left = sph_point_pattern_create_image(
            pp,
            model->window_size_y,
            model->window_size_y,1.5);
    cpl_ensure_code(testim_left != NULL,cpl_error_get_code());

    cpl_image_save(testim_left,
            "qc_input_pattern_left.fits",
            CPL_TYPE_DOUBLE,
            NULL,
            CPL_IO_CREATE);

    testim_right = sph_point_pattern_create_image(
            pp,
            model->window_size_y,
            model->window_size_y,1.5);
    cpl_ensure_code(testim_right != NULL,cpl_error_get_code());

    cpl_image_save(testim_right,
            "qc_input_pattern_right.fits",
            CPL_TYPE_DOUBLE,
            NULL,
            CPL_IO_CREATE);

    testim = sph_ird_instrument_model_assemble_image(
            model,
            testim_left,
            testim_right);

    cpl_image_save(testim,
            "qc_input_pattern.fits",
            CPL_TYPE_DOUBLE,
            NULL,
            CPL_IO_CREATE);

    cpl_image_delete(testim_left); testim_left = NULL;
    cpl_image_delete(testim_right); testim_right = NULL;
    cpl_image_delete(testim); testim = NULL;
    sph_point_delete(point);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_distortion(
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_ird_instrument_model* model,
        cpl_propertylist* qclist)
{
    cpl_image*          qcim = NULL;
    if ( model && qcim_l && qcim_r ) {
        qcim = sph_ird_instrument_model_assemble_image(model,qcim_l,qcim_r);
        cpl_image_save(qcim,"qc_image.fits",CPL_TYPE_DOUBLE,
                qclist,CPL_IO_CREATE);
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Full QC image  saved.");
    }

    if ( qcim_r ) {
        cpl_image_save(qcim_r,"qc_image_right.fits",CPL_TYPE_DOUBLE,
                qclist,CPL_IO_CREATE);
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Right QC image saved.");
    }
    if ( qcim_l ) {
        cpl_image_save(qcim_l,"qc_image_left.fits",CPL_TYPE_DOUBLE,
                qclist,CPL_IO_CREATE);
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "Left QC image  saved.");
    }
    cpl_image_delete(qcim);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
static
sph_error_code
__sph_ird_distortion_map_create_qc_of_accuracy(
        sph_ird_distortion_map * self,
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_point_pattern* ppin)
{
    sph_point_pattern* pp_left_corrected = NULL;
    sph_point_pattern* pp_right_corrected = NULL;
    sph_distortion_map*         dmap = NULL;
    sph_distortion_vector*         dv = NULL;
    sph_dataset*                dat = NULL;
    int                 ii      = 0;
    int                 nn      = 0;
    double              rr = 0.0;
    cpl_image*          im = NULL;
    cpl_image*          imdx = NULL;
    cpl_image*          imdy = NULL;
    cpl_image*          nim = NULL;
    double              cx = 512.0;
    double              cy = 512.0;
    int                 bpix = 0;
    double              opt_cent_l_x;
    double              opt_cent_l_y;
    double              opt_cent_r_x;
    double              opt_cent_r_y;
    if ( self->user_cent ) {
        opt_cent_l_x = self->cent_left_x;
        opt_cent_l_y = self->cent_left_y;
        opt_cent_r_x = self->cent_right_x;
        opt_cent_r_y = self->cent_right_y;
    }

    if ( qcim_l ) {
        pp_left_corrected = sph_point_pattern_new_from_image(
                    qcim_l,
                    NULL,
                    self->threshold_sigma,
                    3.0);
        sph_point_pattern_centralise_wrt_central_point(
                pp_left_corrected);
        sph_point_pattern_centralise_wrt_central_point(
                ppin);

        dmap = sph_distortion_map_new(pp_left_corrected,ppin);

        dat = sph_dataset_new(100,-1.0,9.5);
        im = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        imdx = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        imdy = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        nim = cpl_image_new(65,65,CPL_TYPE_DOUBLE);

        sph_distortion_map_remove_offset(dmap,NULL,NULL);

        SPH_RAISE_CPL_RESET;
        for (ii = 0; ii < sph_distortion_map_get_size(dmap); ++ii) {
            dv = sph_distortion_map_get(dmap,ii);
            rr = dv->dx * dv->dx + dv->dy * dv->dy;
            rr = sqrt(rr);
            cpl_image_set( im, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    rr);
            cpl_image_set( imdx, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    dv->dx);
            cpl_image_set( imdy, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    dv->dy);

            nn = cpl_image_get( nim,
                    (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    &bpix);

            cpl_image_set( nim, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    nn + 1);
            if ( rr > 9.0 ) rr = 9.0;
            sph_dataset_add_value(dat,rr,1.0);
        }
        SPH_RAISE_CPL_RESET;

        sph_dataset_save_ascii(dat,
                "qc_residuals_left_hist.txt");

        cpl_image_threshold(nim,-0.5,0.5,1.0,1.0);
        cpl_image_divide(im,nim);
        cpl_image_divide(imdx,nim);
        cpl_image_divide(imdy,nim);

        cpl_image_save(im,
                "qc_residuals_left.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
        cpl_image_save(imdx,
                "qc_residuals_left_x.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
        cpl_image_save(imdy,
                "qc_residuals_left_y.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);

        sph_point_pattern_delete(pp_left_corrected);
        sph_distortion_map_delete(dmap); dmap = NULL;
        sph_dataset_delete(dat); dat = NULL;
        cpl_image_delete(im); im = NULL;
        cpl_image_delete(imdx); imdy = NULL;
        cpl_image_delete(imdy); imdx = NULL;
        cpl_image_delete(nim); nim = NULL;
        pp_left_corrected = NULL;
    }
    if ( qcim_r ) {
        SPH_RAISE_CPL_RESET;
        pp_right_corrected = sph_point_pattern_new_from_image(
                    qcim_r,
                    NULL,
                    self->threshold_sigma,
                    3.0);

        sph_point_pattern_centralise_wrt_central_point(
                pp_right_corrected);
        sph_point_pattern_centralise_wrt_central_point(
                ppin);
        dmap = sph_distortion_map_new(
                pp_right_corrected,
                ppin);

        dat = sph_dataset_new(100,-1.0,9.5);

        im = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        imdx = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        imdy = cpl_image_new(65,65,CPL_TYPE_DOUBLE);
        nim = cpl_image_new(65,65,CPL_TYPE_DOUBLE);

        sph_distortion_map_remove_offset(dmap,NULL,NULL);

        for (ii = 0; ii < sph_distortion_map_get_size(dmap); ++ii) {
            dv = sph_distortion_map_get(dmap,ii);
            rr = dv->dx * dv->dx + dv->dy * dv->dy;
            rr = sqrt(rr);
            cpl_image_set( im, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    rr);
            cpl_image_set( imdx, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    dv->dx);
            cpl_image_set( imdy, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    dv->dy);
            nn = cpl_image_get( nim,
                    (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    &bpix);

            cpl_image_set( nim, (dv->x + cx) * 60.0 / 1024.0 + 1,
                    (dv->y + cy) * 60.0 / 1024.0 + 1,
                    nn + 1);
            if ( rr > 9.0 ) rr = 9.0;
            sph_dataset_add_value(dat,rr,1.0);
        }

        sph_dataset_save_ascii(dat,
                "qc_residuals_right_hist.txt");
        cpl_image_threshold(nim,-0.5,0.5,1.0,1.0);
        cpl_image_divide(im,nim);
        cpl_image_divide(imdx,nim);
        cpl_image_divide(imdy,nim);

        cpl_image_save(im,
                "qc_residuals_right.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
        cpl_image_save(imdx,
                "qc_residuals_right_x.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
        cpl_image_save(imdy,
                "qc_residuals_right_y.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
        sph_distortion_map_delete(dmap); dmap = NULL;
        sph_dataset_delete(dat); dat = NULL;
        cpl_image_delete(im); im = NULL;
        cpl_image_delete(nim); nim = NULL;
        sph_point_pattern_delete(pp_right_corrected);
        cpl_image_delete(imdx); imdy = NULL;
        cpl_image_delete(imdy); imdx = NULL;
        pp_right_corrected = NULL;
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_ird_distortion_map_create_qc_of_pp(
        sph_ird_distortion_map * self,
        sph_ird_instrument_model* model,
        cpl_image* qcim_l,
        cpl_image* qcim_r,
        sph_point_pattern* obs_left,
        sph_point_pattern* obs_right)
{
    sph_point_pattern* pp_left_corrected = NULL;
    sph_point_pattern* pp_right_corrected = NULL;
    cpl_image*          imleft = NULL;
    cpl_image*          imright = NULL;
    cpl_image*          im = NULL;

    if ( qcim_l ) {
        pp_left_corrected = sph_point_pattern_new_from_image(
                    qcim_l,
                    NULL,
                    self->threshold_sigma,
                    3.0);
        sph_point_pattern_centralise_wrt_central_point(pp_left_corrected);
        sph_point_pattern_save_ascii(pp_left_corrected,
                "qc_pp_left_corrected.txt");
        sph_point_pattern_delete(pp_left_corrected);
        pp_left_corrected = NULL;
    }
    if ( qcim_r ) {
        pp_right_corrected = sph_point_pattern_new_from_image(
                    qcim_r,
                    NULL,
                    self->threshold_sigma,
                    3.0);
        sph_point_pattern_centralise_wrt_central_point(pp_right_corrected);
        sph_point_pattern_save_ascii(pp_right_corrected,
                "qc_pp_right_corrected.txt");
        sph_point_pattern_delete(pp_right_corrected);
        pp_right_corrected = NULL;
    }
    if ( obs_left ) {
        sph_point_pattern_save_ascii(
                obs_left,
                "qc_pp_left.txt");

        SPH_ERROR_RAISE_INFO(
                SPH_ERROR_GENERAL,
                "QC point pattern (left) saved");
        sph_point_pattern_add_offset(
                obs_left,
                model->window_size_y / 2.0,
                model->window_size_y / 2.0);
        imleft = sph_point_pattern_create_image(
                obs_left,
                model->window_size_y,
                model->window_size_y,
                3.0);
        cpl_image_save(imleft,"qc_pp_left.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
    }
    if ( obs_right ) {
        sph_point_pattern_save_ascii(obs_right,"qc_pp_right.txt");
        SPH_ERROR_RAISE_INFO(SPH_ERROR_GENERAL,
                "QC point pattern (right) saved");

        sph_point_pattern_add_offset(
                obs_right,
                model->window_size_y / 2.0,
                model->window_size_y / 2.0);
        imright = sph_point_pattern_create_image(
                obs_right,
                model->window_size_y,
                model->window_size_y,
                3.0);
        cpl_image_save(imright,
                "qc_pp_right.fits",
                CPL_TYPE_DOUBLE,
                NULL,
                CPL_IO_CREATE);
    }
    if ( imright && imleft ) {
        im = sph_ird_instrument_model_assemble_image( model, imleft, imright);
        cpl_image_save(im,"qc_pp_full.fits",CPL_TYPE_DOUBLE,NULL,CPL_IO_CREATE);
        cpl_image_delete(im); im = NULL;
    }
    cpl_image_delete(imleft);
    cpl_image_delete(imright);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


/**@}*/
