/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <cpl.h>
#include <math.h>
#include "sph_zpl_aoc_offset.h"
#include "sph_framecombination.h"
#include "sph_zpl_exposure.h"
#include "sph_quad_image.h"
#include "sph_double_image.h"
#include "sph_master_frame.h"
#include "sph_framelist.h"
#include "sph_error.h"
#include "sph_utils.h"
#include "sph_zpl_utils.h"
#include "sph_common_keywords.h"
#include "sph_zpl_keywords.h"
#include "sph_zpl_tags.h"

/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_zpl_aoc_offset_run Create Polarization AOC Offset Recipe
 *
 * This module provides the algorithm implementation for the creation of the
 * AOC polarizattion offset (polarimetric modes)
 *
 * @par Synopsis:
 * @code
 *   #include "sph_zpl_aoc_offset.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_zpl_aoc_offset recipe
 (polarimetric mode). The error  code returned is always a cpl error code
 (to allow maximal compatibility with  esorex, gasgano, etc.) even if during
 recipe execution an error in the SPHERE API is the cause. In this case
 (and if the underlying error is not a cpl error)  the cpl error code is set
 to the cpl_error_code that matches the failure reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/
static void
__sph_zpl_aoc_offset_delete( sph_quad_image* master_bias_quadimage,
		sph_quad_image* master_dark_quadimage,
		sph_master_frame* intensity_flat_masterframe,
		sph_quad_image* intensity_flat_quadimage,
		sph_double_image* polarization_flat_doubleimage,
		sph_master_frame* modem_efficiency_masterframe,
		sph_framelist* framelist)
{
    if ( master_bias_quadimage ) {
    	sph_quad_image_delete( master_bias_quadimage ); master_bias_quadimage=NULL;
    }
    if ( master_dark_quadimage ) {
    	sph_quad_image_delete( master_dark_quadimage ); master_dark_quadimage=NULL;
    }
    if ( intensity_flat_masterframe ){
    	sph_master_frame_delete( intensity_flat_masterframe ); intensity_flat_masterframe=NULL;
    }
    if ( intensity_flat_quadimage ) {
    	sph_quad_image_delete( intensity_flat_quadimage ); intensity_flat_quadimage=NULL;
    }
    if ( polarization_flat_doubleimage ) {
    	sph_double_image_delete( polarization_flat_doubleimage ); polarization_flat_doubleimage=NULL;
    }
    if ( modem_efficiency_masterframe ) {
    	sph_master_frame_delete( modem_efficiency_masterframe ); modem_efficiency_masterframe=NULL;
    }
    if ( framelist ) {
    	sph_framelist_delete( framelist ); framelist=NULL;
    }
}



static
sph_double_image* sph_zpl_aoc_offset_create_offset( sph_zpl_aoc_offset* self, cpl_frameset* frames,
		sph_quad_image* master_bias_quadimage, sph_quad_image* master_dark_quadimage,
		sph_master_frame* intensity_flat_masterframe, sph_quad_image* intensity_flat_quadimage,
		sph_double_image* polarization_flat_doubleimage, sph_master_frame* modem_efficiency_masterframe )
{
	sph_quad_image*		collapsed_quadimage		= NULL;
	sph_double_image*	stokes					= NULL;
    cpl_frameset*		ovsc_subtruct_frames	= NULL;
	int					rerr					= CPL_ERROR_NONE;

    if ( frames ){

        if ( self->subtract_overscan ) {
        	SPH_INFO_MSG("Subtract overscans from rawframes...")
        	ovsc_subtruct_frames = sph_zpl_utils_subtract_overscans( frames );
        	if ( ovsc_subtruct_frames ){
        		collapsed_quadimage = sph_framecombination_quad_image_from_frameset(
        									ovsc_subtruct_frames, self->coll_alg, self->framecomb_parameterlist);
        		cpl_frameset_delete( ovsc_subtruct_frames );
        	} else {
        		SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Overscans could not be subtracted, trying to proceed further.")
               	collapsed_quadimage = sph_framecombination_quad_image_from_frameset(
                                              frames, self->coll_alg, self->framecomb_parameterlist);
        	}
        } else {
        	collapsed_quadimage = sph_framecombination_quad_image_from_frameset(
                                      frames, self->coll_alg, self->framecomb_parameterlist);
        }

    } else {
        SPH_ERR("no rawframes are set up to combine; please verify TAGS")
        return NULL;
    }


    if ( !collapsed_quadimage ){
    	SPH_ERR("Couldn't collapse input frames.")
    	return NULL;
    }

	if ( master_bias_quadimage ) {
		rerr = sph_quad_image_subtract_quad_image( collapsed_quadimage, master_bias_quadimage);
		if ( rerr != CPL_ERROR_NONE ){
			sph_error_raise( SPH_ERROR_GENERAL,
					__FILE__, __func__, __LINE__,
					SPH_ERROR_ERROR,
					"Error is raised by subtracting master bias:\n"
					"cpl error code is: %d", cpl_error_get_code() );
			if (collapsed_quadimage) {
				sph_quad_image_delete( collapsed_quadimage );
				collapsed_quadimage = NULL;
			}
			return NULL;
		}
	}


	if ( master_dark_quadimage ) {
		rerr = sph_quad_image_subtract_quad_image( collapsed_quadimage, master_dark_quadimage);
		if ( rerr != CPL_ERROR_NONE ){
			sph_error_raise( SPH_ERROR_GENERAL,
					__FILE__, __func__, __LINE__,
					SPH_ERROR_ERROR,
					"Error is raised by subtracting dark:\n"
					"cpl error code is: %d", cpl_error_get_code() );
			if (collapsed_quadimage) {
				sph_quad_image_delete( collapsed_quadimage );
				collapsed_quadimage = NULL;
			}
			return NULL;
		}
	}


   if ( intensity_flat_masterframe ){
			   rerr = sph_quad_image_divide_master_frame( collapsed_quadimage, intensity_flat_masterframe );
			   if ( rerr != CPL_ERROR_NONE ){
							sph_error_raise( SPH_ERROR_GENERAL,
									__FILE__, __func__, __LINE__,
									SPH_ERROR_ERROR,
									"Error is raised by applying(dividing) intensity flat masterframe:\n"
									"cpl error code is: %d", cpl_error_get_code() );
							if (collapsed_quadimage) {
								sph_quad_image_delete( collapsed_quadimage );
								collapsed_quadimage = NULL;
							}
							return NULL;
				}
	} else 	if ( intensity_flat_quadimage ) {
				rerr = sph_quad_image_divide_quad_image( collapsed_quadimage,intensity_flat_quadimage);
				if ( rerr != CPL_ERROR_NONE ){
					sph_error_raise( SPH_ERROR_GENERAL,
							__FILE__, __func__, __LINE__,
							SPH_ERROR_ERROR,
							"Error is raised by applying (dividing) intensity flat:\n"
							"cpl error code is: %d", cpl_error_get_code() );
					if (collapsed_quadimage) {
						sph_quad_image_delete( collapsed_quadimage );
						collapsed_quadimage = NULL;
					}
					return NULL;
				}
	}

	stokes = sph_zpl_utils_calculate_stokes_param_double_image_from_quad_image( collapsed_quadimage );
    if (!stokes){
    	SPH_ERR("stokes parameters is NULL");
		if (collapsed_quadimage) {
			sph_quad_image_delete( collapsed_quadimage );
			collapsed_quadimage = NULL;
		}
		return NULL;
    }


    if ( polarization_flat_doubleimage ){
    	rerr = sph_zpl_utils_apply_polarization_flat( stokes, polarization_flat_doubleimage );
		if ( rerr != CPL_ERROR_NONE ){
			sph_error_raise( SPH_ERROR_GENERAL,
					__FILE__, __func__, __LINE__,
					SPH_ERROR_ERROR,
					"Error is raised by applying polarization flat:\n"
					"sph error code is: %d", sph_error_get_last_code() );
			if (collapsed_quadimage) {
				sph_quad_image_delete( collapsed_quadimage );
				collapsed_quadimage = NULL;
			}
			if (stokes) {
				sph_double_image_delete( stokes );
				stokes = NULL;
			}
			return NULL;
		}

    }

    if ( modem_efficiency_masterframe ){
    	rerr = sph_zpl_utils_apply_modem_efficiency( stokes, modem_efficiency_masterframe );
		if ( rerr != CPL_ERROR_NONE ){
			sph_error_raise( SPH_ERROR_GENERAL,
					__FILE__, __func__, __LINE__,
					SPH_ERROR_ERROR,
					"Error is raised by applying modem efficiency:\n"
					"sph error code is: %d", sph_error_get_last_code() );
			if (collapsed_quadimage) {
				sph_quad_image_delete( collapsed_quadimage );
				collapsed_quadimage = NULL;
			}
			if (stokes) {
				sph_double_image_delete( stokes );
				stokes = NULL;
			}
			return NULL;
		}
    }
    if (collapsed_quadimage) {
    	sph_quad_image_delete( collapsed_quadimage ); collapsed_quadimage = NULL;
    }

	return stokes;

}




cpl_error_code sph_zpl_aoc_offset_run( sph_zpl_aoc_offset* self )
{
	sph_zpl_exposure*	zplexp1							= NULL;
	sph_zpl_exposure*	zplexp2							= NULL;
    sph_quad_image*		master_bias_quadimage			= NULL;
    sph_quad_image*		master_dark_quadimage			= NULL;
    sph_quad_image*		intensity_flat_quadimage		= NULL;
    sph_master_frame*	intensity_flat_masterframe		= NULL;
    sph_double_image*	polarization_flat_doubleimage	= NULL;
    sph_master_frame*	modem_efficiency_masterframe	= NULL;

    sph_framelist*		framelist						= NULL;
    int					nset_first						= 0;
    int					nset_step						= 1;

    cpl_propertylist*	pl						= NULL;

    int					recipe_error			= CPL_ERROR_NONE;


    SPH_INFO_MSG("Starting sph_zpl_aoc_offset_run...");
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE


    //if master bias exists load it
    if ( self->master_bias_frame ){
    	SPH_INFO_MSG("Loading master bias...\n");
        master_bias_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->master_bias_frame ), 0);
    }

    //if master dark exists load it
    if ( self->master_dark_frame){
       	SPH_INFO_MSG("Loading master dark...\n");
        master_dark_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->master_dark_frame ), 0);
    }

    if ( self->intensity_flat_frame_master){
        //if intensity flat field master frame exists load
       	SPH_INFO_MSG("Loading intensity flat field masterframe...\n");
        intensity_flat_masterframe = sph_master_frame_load( cpl_frame_get_filename( self->intensity_flat_frame_master ), 0);
    } else {
        //if intensity flat field quad image exists load
    	if ( self->intensity_flat_frame){
    		SPH_INFO_MSG("Loading intensity flat field quad image...\n");
    		intensity_flat_quadimage = sph_quad_image_load( cpl_frame_get_filename( self->intensity_flat_frame ), 0);
    	}
    }

    //if polarization flat field exists load it
    if ( self->polarization_flat_frame){
       	SPH_INFO_MSG("Loading polarization flat field...\n");
        polarization_flat_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->polarization_flat_frame ), 0);
    }

    //if modulation/demodulation efficiency exists load it
    if ( self->modem_efficiency_frame){
       	SPH_INFO_MSG("Loading modulation/demodulation efficiency...\n");
        modem_efficiency_masterframe = sph_master_frame_load( cpl_frame_get_filename( self->modem_efficiency_frame ), 0);
    }


    framelist = sph_framelist_create( self->rawframes, SPH_ZPL_KEYWORD_HWP2_SETTING, 44.0);
    if (!framelist){
    	SPH_ERR("Framelist is NULL, can't proceed further.")
         __sph_zpl_aoc_offset_delete(
                		master_bias_quadimage,
                		master_dark_quadimage,
                		intensity_flat_masterframe,
                		intensity_flat_quadimage,
                		polarization_flat_doubleimage,
                		modem_efficiency_masterframe,
                		framelist);
		SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
    }

    if ( framelist->nsets != 2){
    	SPH_ERROR_RAISE_WARNING( CPL_ERROR_ILLEGAL_INPUT,
    	                "Bad number of HWP2 offset angles for +Q/+U and -Q/-U. "
    	                "It is usually needed two different HWP2 offset angles [0.0(+Q)/22.5(+U) "
    	                "and 45.0(-Q)/67.5(-U)] degrees "
    	                "in the input. Make sure the %s keyword is present in all the "
    	                "raw frames. nsets = %d", SPH_ZPL_KEYWORD_HWP2_SETTING, framelist->nsets );
    	SPH_INFO_MSG("Recipe will consider all rawframes having one polarization direction, which is correct for the "
    			"V->Q crosstalk!" );
        self->aoc_offset_di_plus = sph_zpl_aoc_offset_create_offset( self, self->rawframes,
        						master_bias_quadimage, master_dark_quadimage,
        						intensity_flat_masterframe, intensity_flat_quadimage,
        						polarization_flat_doubleimage,
        						modem_efficiency_masterframe );

        if( !self->aoc_offset_di_plus ) {
        	SPH_ERR("Can't create aoc_efficiency_di_plus...");
            __sph_zpl_aoc_offset_delete(
                   		master_bias_quadimage,
                   		master_dark_quadimage,
                   		intensity_flat_masterframe,
                   		intensity_flat_quadimage,
                   		polarization_flat_doubleimage,
                   		modem_efficiency_masterframe,
                   		framelist);
            SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
        }


        //calculate quality parameters and save +U/+Q as a double image
        recipe_error = sph_double_image_quality_check( self->aoc_offset_di_plus );
        if ( recipe_error ){
            sph_error_raise( SPH_ERROR_GENERAL,
                    __FILE__, __func__, __LINE__,
                    SPH_ERROR_ERROR,
                    "sph_double_image_quality_check returns cpl error.\n"
                    "cpl error code is: %d", recipe_error );
            __sph_zpl_aoc_offset_delete(
                   		master_bias_quadimage,
                   		master_dark_quadimage,
                   		intensity_flat_masterframe,
                   		intensity_flat_quadimage,
                   		polarization_flat_doubleimage,
                   		modem_efficiency_masterframe,
                   		framelist);
            return recipe_error;
        }

        recipe_error = sph_double_image_save_dfs( self->aoc_offset_di_plus,
                                                self->aoc_offset_plus_outfilename,
                                                self->inframes,
                                                cpl_frameset_get_first( self->rawframes ),
                                                self->inparams,
                                                SPH_ZPL_TAG_AOC_OFFSET_PLUS_CALIB,
                                                SPH_RECIPE_NAME_ZPL_AOC_OFFSET,
                                                SPH_PIPELINE_NAME_ZIMPOL,
                                                pl);

        self->aoc_offset_master = sph_double_image_calculate_pframe_slash_iframe_master_frame( self->aoc_offset_di_plus );
        if( !self->aoc_offset_master){
        		SPH_ERR("Can't create masterframe aoc_offset_master from aoc_offset_di_plus");
                __sph_zpl_aoc_offset_delete(
                       		master_bias_quadimage,
                       		master_dark_quadimage,
                       		intensity_flat_masterframe,
                       		intensity_flat_quadimage,
                       		polarization_flat_doubleimage,
                       		modem_efficiency_masterframe,
                       		framelist);
                SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
        }


        //quality check (calculate statistical moments for the aoc offset master frame)
        recipe_error = sph_master_frame_quality_check( self->aoc_offset_master );
        if ( recipe_error ){
            sph_error_raise( SPH_ERROR_GENERAL,
                    __FILE__, __func__, __LINE__,
                    SPH_ERROR_ERROR,
                    "sph_master_frame_quality_check returns cpl error.\n"
                    "cpl error code is: %d", recipe_error );
            __sph_zpl_aoc_offset_delete(
                   		master_bias_quadimage,
                   		master_dark_quadimage,
                   		intensity_flat_masterframe,
                   		intensity_flat_quadimage,
                   		polarization_flat_doubleimage,
                   		modem_efficiency_masterframe,
                   		framelist);
            return recipe_error;
        }
        // save final aoc offset as a master frame
        recipe_error = sph_master_frame_save_dfs( self->aoc_offset_master,
                                                self->aoc_offset_outfilename,
                                                self->inframes,
                                                cpl_frameset_get_first( self->rawframes ),
                                                self->inparams,
                                                SPH_ZPL_TAG_AOC_OFFSET_CALIB,
                                                SPH_RECIPE_NAME_ZPL_AOC_OFFSET,
                                                SPH_PIPELINE_NAME_ZIMPOL,
                                                pl);
        if ( recipe_error ){
            sph_error_raise( SPH_ERROR_GENERAL,
                    __FILE__, __func__, __LINE__,
                    SPH_ERROR_ERROR,
                    "Could not create aoc offset offset:\n"
                    "sph_double_image_save_dfs returns error code.\n"
                    "cpl error code is: %d", recipe_error );
        }
        __sph_zpl_aoc_offset_delete(
               		master_bias_quadimage,
               		master_dark_quadimage,
               		intensity_flat_masterframe,
               		intensity_flat_quadimage,
               		polarization_flat_doubleimage,
               		modem_efficiency_masterframe,
               		framelist);

        SPH_INFO_MSG("sph_zpl_aoc_offset_run...End");
        SPH_ERROR_CHECK_STATE_RETURN_ERRCODE


    } else {

    	// To calculate aoc_offset one doesn't need to know if a first frameset belong to plus or minus
    	// because the final product is a sum. However we do set up it properly in order to provide corrected
    	// intermediate product filename
    	SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO, "framelist->values[0] = %5.2f\n",  framelist->values[0])
    	SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO, "framelist->values[1] = %5.2f\n",  framelist->values[1])
    	zplexp1 = sph_zpl_exposure_load( cpl_frameset_get_first( framelist->framesets[0]), 0);
    	zplexp2 =  sph_zpl_exposure_load( cpl_frameset_get_first( framelist->framesets[1]), 0);

    	if ( ( isSame( cpl_propertylist_get_double( zplexp1->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ),
    				 SPH_ZPL_KEYWORD_VALUE_HWP2_QPLUS ) &&
    		   isSame( cpl_propertylist_get_double( zplexp2->properties, SPH_ZPL_KEYWORD_HWP2_SETTING),
    				 SPH_ZPL_KEYWORD_VALUE_HWP2_QMINUS) )  // +Q and -Q
    				 	 	 	 	 ||
    		 ( isSame( cpl_propertylist_get_double( zplexp1->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ),
    				  SPH_ZPL_KEYWORD_VALUE_HWP2_UPLUS ) &&
    		   isSame( cpl_propertylist_get_double( zplexp2->properties, SPH_ZPL_KEYWORD_HWP2_SETTING),
    				     				 SPH_ZPL_KEYWORD_VALUE_HWP2_UMINUS) )  // +U and -U
    				     				 	 	 	 	 	 	 	 	 	 	 )
    	{

    		nset_first = 0; nset_step = 1;

    	} else if ( ( isSame( cpl_propertylist_get_double( zplexp2->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ),
    		    			 SPH_ZPL_KEYWORD_VALUE_HWP2_QPLUS ) &&
    		    	 isSame( cpl_propertylist_get_double( zplexp1->properties, SPH_ZPL_KEYWORD_HWP2_SETTING),
    		    			 SPH_ZPL_KEYWORD_VALUE_HWP2_QMINUS) )   // +Q and -Q
    		    			 	 	 ||
				 	( isSame( cpl_propertylist_get_double( zplexp2->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ),
 					 			SPH_ZPL_KEYWORD_VALUE_HWP2_UPLUS ) &&
				 	  isSame( cpl_propertylist_get_double( zplexp1->properties, SPH_ZPL_KEYWORD_HWP2_SETTING),
   					     		 SPH_ZPL_KEYWORD_VALUE_HWP2_UMINUS) )  // +U and -U
    																	)
    	{
    		    	nset_first = 1; nset_step = -1;
    	} else {
        	SPH_ERROR_RAISE_WARNING( CPL_ERROR_ILLEGAL_INPUT,
        	                "The HWP2 offset pair-angles values are not as expected Q(0.0,45.0)/U(22.5,67.5)."
        	                "The actual values are: %5.2f, %5.2f\n",
        	                cpl_propertylist_get_double( zplexp1->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ),
        	                cpl_propertylist_get_double( zplexp2->properties, SPH_ZPL_KEYWORD_HWP2_SETTING ) );
        	SPH_INFO_MSG("The first set of the framlist will be considered as +Q(or +U), the second as -Q(or -U)");
        	nset_first = 0; nset_step = 1;
     	}
    	if (zplexp1) {
    		sph_zpl_exposure_delete( zplexp1 ); zplexp1 = NULL;
    	}
    	if (zplexp2) {
    		sph_zpl_exposure_delete( zplexp2 ); zplexp2 = NULL;
    	}
    }

	SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO, "nset_first = %d, nset_step = %d\n",  nset_first, nset_step)

    self->aoc_offset_di_plus = sph_zpl_aoc_offset_create_offset( self, framelist->framesets[nset_first],
    						master_bias_quadimage, master_dark_quadimage,
    						intensity_flat_masterframe,intensity_flat_quadimage,
    						polarization_flat_doubleimage,
    						modem_efficiency_masterframe );

    self->aoc_offset_di_minus =  sph_zpl_aoc_offset_create_offset( self, framelist->framesets[nset_first+nset_step],
							master_bias_quadimage, master_dark_quadimage,
							intensity_flat_masterframe, intensity_flat_quadimage,
							polarization_flat_doubleimage,
							modem_efficiency_masterframe );

    if ( !(self->aoc_offset_di_plus && self->aoc_offset_di_minus) ){
        SPH_ERROR_RAISE_ERR( SPH_ERROR_GENERAL,
                "aoc_offset_di_minus and/or aoc_offset_di_plus have(has) null pointer(s).\n");
        SPH_RAISE_CPL;
        __sph_zpl_aoc_offset_delete(
               		master_bias_quadimage,
               		master_dark_quadimage,
               		intensity_flat_masterframe,
               		intensity_flat_quadimage,
               		polarization_flat_doubleimage,
               		modem_efficiency_masterframe,
               		framelist);
        return (int)cpl_error_get_code();
    }

    //calculate quality parameters and save +U/+Q aoc offset as a double image
    recipe_error = sph_double_image_quality_check( self->aoc_offset_di_plus );
    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "sph_double_image_quality_check returns cpl error.\n"
                "cpl error code is: %d", recipe_error );
        __sph_zpl_aoc_offset_delete(
               		master_bias_quadimage,
               		master_dark_quadimage,
               		intensity_flat_masterframe,
               		intensity_flat_quadimage,
               		polarization_flat_doubleimage,
               		modem_efficiency_masterframe,
               		framelist);
        return recipe_error;
    }
    recipe_error = sph_double_image_save_dfs( self->aoc_offset_di_plus,
                                            self->aoc_offset_plus_outfilename,
                                            self->inframes,
                                            cpl_frameset_get_first( self->rawframes ),
                                            self->inparams,
                                            SPH_ZPL_TAG_AOC_OFFSET_PLUS_CALIB,
                                            SPH_RECIPE_NAME_ZPL_AOC_OFFSET,
                                            SPH_PIPELINE_NAME_ZIMPOL,
                                            pl);

    //calculate quality parameters and save -U/-Q aoc offset as a double image
    recipe_error = sph_double_image_quality_check( self->aoc_offset_di_minus );
    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "sph_double_image_quality_check returns cpl error.\n"
                "cpl error code is: %d", recipe_error );
        __sph_zpl_aoc_offset_delete(
               		master_bias_quadimage,
               		master_dark_quadimage,
               		intensity_flat_masterframe,
               		intensity_flat_quadimage,
               		polarization_flat_doubleimage,
               		modem_efficiency_masterframe,
               		framelist);
        return recipe_error;
    }
    recipe_error = sph_double_image_save_dfs( self->aoc_offset_di_minus,
                                            self->aoc_offset_minus_outfilename,
                                            self->inframes,
                                            cpl_frameset_get_first( self->rawframes ),
                                            self->inparams,
                                            SPH_ZPL_TAG_AOC_OFFSET_MINUS_CALIB,
                                            SPH_RECIPE_NAME_ZPL_AOC_OFFSET,
                                            SPH_PIPELINE_NAME_ZIMPOL,
                                            pl);

    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "sph_double_image_save_dfs for the aoc_offset_di_minus returns cpl error.\n"
                "cpl error code is: %d", recipe_error );
        __sph_zpl_aoc_offset_delete(
               		master_bias_quadimage,
               		master_dark_quadimage,
               		intensity_flat_masterframe,
               		intensity_flat_quadimage,
               		polarization_flat_doubleimage,
               		modem_efficiency_masterframe,
               		framelist);
        return recipe_error;
    }

    //calculate p/i for +Q/+U and -Q/-U-> then  add P/I(+Q/+U) to P/I(-U/-Q) and divide by 2.
    self->aoc_offset_master = sph_zpl_utils_calculate_polarization_offset(self->aoc_offset_di_plus, self->aoc_offset_di_minus);

    //quality check (calculate statistical moments for the aoc polarization master frame)
    recipe_error = sph_master_frame_quality_check( self->aoc_offset_master );
    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "sph_master_frame_quality_check returns cpl error.\n"
                "cpl error code is: %d", recipe_error );

        return recipe_error;
    }

    // save final aoc polarization offset as a master frame
    recipe_error = sph_master_frame_save_dfs( self->aoc_offset_master,
                                            self->aoc_offset_outfilename,
                                            self->inframes,
                                            cpl_frameset_get_first( self->rawframes ),
                                            self->inparams,
                                            SPH_ZPL_TAG_AOC_OFFSET_CALIB,
                                            SPH_RECIPE_NAME_ZPL_AOC_OFFSET,
                                            SPH_PIPELINE_NAME_ZIMPOL,
                                            pl);


    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "Could not create aoc polarization offset:\n"
                "sph_master_frame_save_dfs returns error code.\n"
                "cpl error code is: %d", recipe_error );
    }
    __sph_zpl_aoc_offset_delete(
           		master_bias_quadimage,
           		master_dark_quadimage,
           		intensity_flat_masterframe,
           		intensity_flat_quadimage,
           		polarization_flat_doubleimage,
           		modem_efficiency_masterframe,
           		framelist);

    SPH_INFO_MSG("sph_zpl_aoc_offset_run...End");
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}
/**@}*/
