/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <cpl.h>
#include <math.h>
#include "sph_filemanager.h"
#include "sph_keywordnode.h"
#include "sph_framecombination.h"
#include "sph_double_image.h"
#include "sph_error.h"
#include "sph_fits.h"
#include "sph_utils.h"
#include "sph_double_image.h"
#include "sph_error.h"
#include "sph_common_keywords.h"
#include "sph_zpl_basic_collapse_imaging.h"
#include "sph_zpl_exposure_imaging.h"
#include "sph_utils.h"
#include "sph_zpl_utils.h"
#include "sph_zpl_keywords.h"
#include "sph_zpl_tags.h"


/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_zpl_basic_collapse_imaging_run Basic Collapse Imaging Recipe
 *
 * This module provides the algorithm implementation for the basic (detector)
 * reduced image (imaging mode)
 *
 * @par Synopsis:
 * @code
 *   #include "sph_zpl_basic_collapse_imaging.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_zpl_basic_collapse_imaging recipe
 (imaging mode). The error  code returned is always a cpl error code
 (to allow maximal compatibility with  esorex, gasgano, etc.) even if during
 recipe execution an error in the SPHERE API is the cause. In this case
 (and if the underlying error is not a cpl error)  the cpl error code is set
 to the cpl_error_code that matches the failure reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/

static
cpl_boolean sph_zpl_basic_collapse_imaging_cfg_hasplane( sph_keywordnode* node, int fplane){
	sph_keywordnode*	nodermv		= NULL;
	int					cfg_plane   = -999;

	SPH_INFO_MSG("enter hasplane function");

	if ( (nodermv = sph_keywordnode_get_keywordnode_by_name( node, "REMOVE")) != NULL ){
		for ( nodermv = sph_keywordnode_GetFirstChild(nodermv); nodermv != NULL; nodermv = sph_keywordnode_GetNext(nodermv) ){
			if ( sph_keywordnode_GetInt(nodermv, "PLANE", &cfg_plane ) ){
				SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "cfg_plane = %d", cfg_plane);
				if ( fplane == cfg_plane ) {
					SPH_INFO_MSG("return TRUE!");
					return TRUE;
				}
			} else {
				SPH_ERR("Couldn't retrieve value of the PLANE leave for the current REMOVE plane");
			}
		}
	} else {
		SPH_ERR("Node doesn't contain REMOVE sub-node, please verify cfg file!");
	}

	SPH_INFO_MSG("return FALSE!");
	return FALSE;
}


static
cpl_frameset*	sph_zpl_basic_collapse_imaging_remove_planes( sph_zpl_basic_collapse_imaging* self, cpl_frameset* inframes )
{
	cpl_frameset* 				result 		= NULL;
	cpl_frame*					curframe	= NULL;
	cpl_frame*					newframe	= NULL;
	sph_keywordnode*			root		= NULL;
	sph_keywordnode*			node		= NULL;
	sph_zpl_exposure_imaging*	zplexp		= NULL;
	int                         nplanes     = 0;
	cpl_table*					ovsc_table	= NULL;
	cpl_boolean					rplane		= FALSE;
	sph_error_code				rerr		= CPL_ERROR_NONE;


	char string[256];
	int	  fplane = 0, opened = 0, append = 0, duplicate = 1;

	root = sph_keywordnode_load_configuration( self->basic_collapse_imaging_cfgfilename );
	if ( !root ) {
		return NULL;
	}

	result = cpl_frameset_new();

	curframe = cpl_frameset_get_first( inframes );
	while ( curframe ){


		if ( ( node = sph_keywordnode_get_keywordnode_by_name( root, "FILE")) == NULL ) {
			SPH_ERR("No FILE node are found in the cfg file");
			sph_keywordnode_destroy(root); //destroy cfg tree
			cpl_frameset_delete(result); result = NULL;
			return NULL;
		}

		duplicate = 1;
		for ( node = sph_keywordnode_GetFirstChild(node); node != NULL; node = sph_keywordnode_GetNext(node)) {
			if ( !sph_keywordnode_GetString( node, "NAME", string, sizeof(string)) ){
				SPH_INFO_MSG("Could not retrieve a leaf NAME(filename) for the current FILE-node")
			}
			SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "filename found from cfg tree = %s", string);

			if ( !strcmp( string, cpl_frame_get_filename( curframe ))) {

				duplicate = 0;

				fplane = 0; opened = 0; append = 0;
				nplanes = sph_fits_get_nplanes( cpl_frame_get_filename(curframe), 0 );
				SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "nplanes = %d", nplanes);
				while ( fplane < nplanes ) {
					SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "fplane = %d", fplane);
					zplexp = sph_zpl_exposure_imaging_load( curframe, fplane );
					//SPH_ERROR_ENSURE_GOTO_EXIT( zplexp, cpl_error_get_code());

					if (zplexp == NULL) {
						SPH_ERR("Problem to read zpl exposure: zplexp has null pointer");
						cpl_frameset_delete(result); result = NULL;
						sph_keywordnode_destroy(root); //destroy cfg tree
						return NULL;
					}

					SPH_INFO_MSG("Reset SPH ERROR to handle error of the hasplane function properly!");
					sph_error_reset();
					rplane = sph_zpl_basic_collapse_imaging_cfg_hasplane( node, fplane );
					rerr = sph_error_get_last_code();
					SPH_INFO_MSG("Reset SPH ERROR to handle error of the hasplane function properly!");
					sph_error_reset();
					if ( !rplane && rerr == CPL_ERROR_NONE )  {
						if ( opened == 0 ){
							newframe = sph_filemanager_create_temp_frame(cpl_frame_get_filename( curframe ), "NONE", CPL_FRAME_GROUP_NONE);
							ovsc_table = sph_zpl_exposure_imaging_ovsc_table_create_empty();
							sph_zpl_exposure_imaging_save_open( zplexp, newframe, NULL, ovsc_table);
							opened = 1;
						} else {
							sph_zpl_exposure_imaging_save_append( zplexp, newframe, ovsc_table );
							append = 1;
						}
					}
					sph_zpl_exposure_imaging_delete( zplexp );
					fplane = fplane + 1;
				} //end while TRUE loop

			}

		} //end for node loop


		if ( duplicate || rerr != CPL_ERROR_NONE )  {
			newframe = cpl_frame_duplicate( curframe );
			cpl_frameset_insert( result, newframe );
		} else {
			if ( append ) {
				sph_zpl_exposure_imaging_finalize_file( newframe, ovsc_table);
				cpl_frameset_insert( result, newframe );
			} else 	if ( opened  ) {
				sph_zpl_exposure_imaging_save_ovsc( newframe, ovsc_table );
				cpl_frameset_insert( result, newframe );
			}
			cpl_table_delete( ovsc_table ); ovsc_table = NULL;
		}

		curframe = cpl_frameset_get_next( inframes );
	} //end of while curframe loop

//EXIT:
	sph_keywordnode_destroy(root); //destroy cfg tree

	if ( cpl_frameset_is_empty( result) ) {
		SPH_INFO_MSG("No frames in the output! Return null...");
		cpl_frameset_delete( result );
		return NULL;
	}
	//sph_cube_finalise_all(result);

	SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

	return result;
}




cpl_error_code sph_zpl_basic_collapse_imaging_run( sph_zpl_basic_collapse_imaging* self )
{
    sph_double_image*				master_bias_doubleimage			= NULL;
    sph_double_image*				master_dark_doubleimage			= NULL;
    sph_double_image*				intensity_flat_doubleimage		= NULL;
    cpl_propertylist*				pl								= NULL;
    int								recipe_error					= CPL_ERROR_NONE;
    cpl_frameset*					rmvframes						= NULL;
    cpl_frameset*					ovsc_subtract_frames			= NULL;


    SPH_INFO_MSG("Starting sph_zpl_basic_collapse_imaging_run...");
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE


	rmvframes = sph_zpl_basic_collapse_imaging_remove_planes( self, self->rawframes );

    if ( rmvframes ){
    	//collapse/combine rmvframes
    	SPH_INFO_MSG("Combine rawframes rested after removing...");
		SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "frameset size = %d", (int)cpl_frameset_get_size(rmvframes) );
		//self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
        //                      rmvframes, self->coll_alg, self->framecomb_parameterlist);

		if ( self->subtract_overscan ) {
        	SPH_INFO_MSG("Subtract overscans from rawframes rested after removing...")
        	ovsc_subtract_frames = sph_zpl_utils_subtract_overscans( rmvframes );
        	if ( ovsc_subtract_frames ){
        		self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
        									ovsc_subtract_frames, self->coll_alg, self->framecomb_parameterlist);
        		cpl_frameset_delete( ovsc_subtract_frames );
        	} else {
        		SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Overscans could not be subtracted, trying to proceed further.")
               	self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
                                              rmvframes, self->coll_alg, self->framecomb_parameterlist);
        	}
        } else {
           	self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
                                          rmvframes, self->coll_alg, self->framecomb_parameterlist);
        }

    } else if ( self->rawframes ){
       	//collapse/combine frames
    	SPH_INFO_MSG("Combine initial raw frames...");
		SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO_MESSAGE, "frameset size = %d", (int)cpl_frameset_get_size(self->rawframes) );
    	//self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
    	//					self->rawframes, self->coll_alg, self->framecomb_parameterlist);

        if ( self->subtract_overscan ) {
        	SPH_INFO_MSG("Subtract overscans from rawframes...")
        	ovsc_subtract_frames = sph_zpl_utils_subtract_overscans( self->rawframes );
        	if ( ovsc_subtract_frames ){
        		self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
        									ovsc_subtract_frames, self->coll_alg, self->framecomb_parameterlist);
        		cpl_frameset_delete( ovsc_subtract_frames );
        	} else {
        		SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Overscans could not be subtracted, trying to proceed further.")
               	self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
                                              self->rawframes, self->coll_alg, self->framecomb_parameterlist);
        	}
        } else {
        	self->basic_collpase_imaging = sph_framecombination_double_image_from_frameset(
                                      self->rawframes, self->coll_alg, self->framecomb_parameterlist);
        }
    } else {
        SPH_ERR ("no frames to combine");
        return sph_error_get_last_code();
    }

    if ( !self->basic_collpase_imaging ){
    	SPH_ERR("Couldn't collapse input frames/rmvframes");
    	if( rmvframes ) {
    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
    	}
    	return sph_error_get_last_code();
    }


    //if master bias exists load it
    if ( self->master_bias_frame ){
    	SPH_INFO_MSG("Loading master bias...\n");
        master_bias_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->master_bias_frame ), 0);
		if ( master_bias_doubleimage ) {
		    SPH_INFO_MSG("Subtracting bias...");
			recipe_error = sph_double_image_subtract_double_image( self->basic_collpase_imaging, master_bias_doubleimage);
			if ( recipe_error != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting master bias:\n"
						"cpl error code is: %d", cpl_error_get_code() );
		    	if( rmvframes ) {
		    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
		    	}
		    	if ( self->basic_collpase_imaging ) {
		    		sph_double_image_delete( self->basic_collpase_imaging ); self->basic_collpase_imaging = NULL;
		    	}
				return recipe_error;
			}
		}
    }

    //if master dark exists load it
    if ( self->master_dark_frame){
       	SPH_INFO_MSG("Loading master dark...\n");
        master_dark_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->master_dark_frame ), 0);
		if ( master_dark_doubleimage ) {
		    SPH_INFO_MSG("Subtracting dark...");
			recipe_error = sph_double_image_subtract_double_image( self->basic_collpase_imaging, master_dark_doubleimage);
			if ( recipe_error != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting dark:\n"
						"cpl error code is: %d", cpl_error_get_code() );
		    	if( rmvframes ) {
		    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
		    	}
		    	if ( self->basic_collpase_imaging ) {
		    		sph_double_image_delete( self->basic_collpase_imaging ); self->basic_collpase_imaging = NULL;
		    	}

				return recipe_error;
			}
		}

    }

    //if intensity flat field exists load
    if ( self->intensity_flat_frame){
       	SPH_INFO_MSG("Loading intensity flat field...\n");
        intensity_flat_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->intensity_flat_frame ), 0);
	    SPH_INFO_MSG("Applying intensity flat field...");
		recipe_error = sph_double_image_divide_double_image( self->basic_collpase_imaging,intensity_flat_doubleimage);
		if ( recipe_error != CPL_ERROR_NONE ){
			sph_error_raise( SPH_ERROR_GENERAL,
					__FILE__, __func__, __LINE__,
					SPH_ERROR_ERROR,
					"Error is raised by applying (dividing) intensity flat:\n"
					"cpl error code is: %d", cpl_error_get_code() );
	    	if( rmvframes ) {
	    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
	    	}
	    	if ( self->basic_collpase_imaging ) {
	    		sph_double_image_delete( self->basic_collpase_imaging ); self->basic_collpase_imaging = NULL;
	    	}

			return recipe_error;
		}
	}



    //quality check (calculate statistical moments for the basic imaging reduced double image)
    recipe_error = sph_double_image_quality_check( self->basic_collpase_imaging );
    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "sph_double_image_quality_check returns cpl error.\n"
                "cpl error code is: %d", cpl_error_get_code() );
    	if( rmvframes ) {
    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
    	}
    	if ( self->basic_collpase_imaging ) {
    		sph_double_image_delete( self->basic_collpase_imaging ); self->basic_collpase_imaging = NULL;
    	}

        return recipe_error;
    }


    //save
    /* Attention:
     * The following function includes a step to update SPH_COMMON_KEYWORD_SPH_TYPE by
     * SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_DOUBLE_IMAGE. It is necessary because cpl_dsf_save_image
     * updates the double image property list by the property list from the first input frames.
     * In the case of some misbehavior exclude this step for checking.
     */

    SPH_INFO_MSG("Saving...");
    recipe_error = sph_double_image_save_dfs( self->basic_collpase_imaging,
	                                  self->basic_collapse_imaging_outfilename,
	                                  self->inframes,
	                                  cpl_frameset_get_first( self->rawframes ),
	                                  self->inparams,
	                                  SPH_ZPL_TAG_BASIC_COLLAPSE_IMAGING_CALIB,
	                                  SPH_RECIPE_NAME_ZPL_BASIC_COLL_IMAGING,
	                                  SPH_PIPELINE_NAME_ZIMPOL,
	                                  pl);

    if ( recipe_error ){
	       sph_error_raise( SPH_ERROR_GENERAL,
	              __FILE__, __func__, __LINE__,
	               SPH_ERROR_ERROR,
	               "Could not save basic imaging reduced double image:\n"
	               "sph_double_image_save_dfs returns error code.\n"
	               "cpl error code is: %d", cpl_error_get_code() );
	    	if( rmvframes ) {
	    		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
	    	}
	    	if ( self->basic_collpase_imaging ) {
	    		sph_double_image_delete( self->basic_collpase_imaging ); self->basic_collpase_imaging = NULL;
	    	}

	       return recipe_error;
	}


    sph_double_image_delete( master_bias_doubleimage );
    sph_double_image_delete( master_dark_doubleimage );
    sph_double_image_delete( intensity_flat_doubleimage );
	if( rmvframes ) {
		cpl_frameset_delete( rmvframes ); rmvframes = NULL;
	}

    SPH_INFO_MSG("sph_zpl_basic_collapse_imaging_run...End");
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}
/**@}*/
