/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <cpl.h>
#include <math.h>
#include "sph_error.h"
#include "sph_utils.h"
#include "sph_common_keywords.h"
#include "sph_double_image.h"
#include "sph_filemanager.h"
#include "sph_zpl_basic_imaging.h"
#include "sph_zpl_exposure_imaging.h"
#include "sph_zpl_utils.h"
#include "sph_zpl_keywords.h"
#include "sph_zpl_tags.h"


/*-----------------------------------------------------------------------------
  Error Codes
 -----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_zpl_basic_imaging_run Basic Calibration Imaging Recipe
 *
 * This module provides the algorithm implementation for the basic (detector)
 * corrected images (imaging mode)
 *
 * @par Synopsis:
 * @code
 *   #include "sph_zpl_basic_imaging.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Interpret the command line options and execute the data processing
 @param    frameset   the frames list
 @param    parlist    the parameters list

 @return   the cpl error code of the operation.

 This is the main recipe function for the sph_zpl_basic_imaging recipe
 (imaging mode). The error  code returned is always a cpl error code
 (to allow maximal compatibility with  esorex, gasgano, etc.) even if during
 recipe execution an error in the SPHERE API is the cause. In this case
 (and if the underlying error is not a cpl error)  the cpl error code is set
 to the cpl_error_code that matches the failure reason best.
 The error from the SPHERE API is still written in the log as usual
 with the more informative and accurate sph_error_code.

 */
/*----------------------------------------------------------------------------*/

static
cpl_frameset* sph_zpl_basic_imaging_create_basic( sph_zpl_basic_imaging* self, cpl_frame* inframe,
		sph_double_image* master_bias_doubleimage, sph_double_image* master_dark_doubleimage, sph_double_image* intensity_flat_doubleimage )
{
	sph_double_image*	curdoubleimage			= NULL;
	cpl_frameset*		inframeset				= NULL;
    cpl_frameset*		ovsc_subtract_frames	= NULL;
	cpl_frame*			outframe				= NULL;
	cpl_frameset*		result					= NULL;
	cpl_frameset*		diframeset				= NULL;
	cpl_frame*			diframe					= NULL;
	cpl_frame*			dupinframe				= NULL;
	cpl_propertylist*	pl						= NULL;
	int					iplane					= 0;
	int					rerr					= CPL_ERROR_NONE;
	char				outfilename[256];
	char				outfilename_tmp[256];


    SPH_INFO_MSG("Starting sph_zpl_basic_imaging_create_basic...");

	if (!inframe ){
        SPH_ERR("no inframes to basic calibrate")
        return NULL;
	}

	// create inframeset and put there only one inframe
	// we do so, because interface for the transformation zplexp frames(cubes) into double image frames(cubes)
	// uses cpl_frameset not cpl_frame
	dupinframe = cpl_frame_duplicate(inframe);
	inframeset = cpl_frameset_new();
	rerr = cpl_frameset_insert( inframeset, dupinframe );
	if ( rerr != CPL_ERROR_NONE ){
		SPH_ERR("Could not create inframeset");
		cpl_frameset_delete(inframeset); inframeset = NULL;
		return NULL;

	}

	//create double image frameset(cubes) from zplexp_imaging inframes (cubes)

    if ( self->subtract_overscan ) {
    	SPH_INFO_MSG("Subtract overscans from inframeset...")
    	ovsc_subtract_frames = sph_zpl_utils_subtract_overscans( inframeset );
    	if ( ovsc_subtract_frames ){
        	diframeset = sph_double_image_create_double_image_frameset_from_zplexp_imaging_cubes( ovsc_subtract_frames );
    		cpl_frameset_delete( ovsc_subtract_frames );
    	} else {
    		SPH_ERROR_RAISE_WARNING(SPH_ERROR_WARNING, "Overscans could not be subtracted, trying to proceed further.")
   	    	diframeset = sph_double_image_create_double_image_frameset_from_zplexp_imaging_cubes( inframeset );
    	}
    } else {
    	diframeset = sph_double_image_create_double_image_frameset_from_zplexp_imaging_cubes( inframeset );
    }

	if ( !diframeset ) {
		SPH_ERR("Failed creation of the double image frameset(cubes) from zplexp imaging inframes (cubes).");
		cpl_frameset_delete(inframeset); inframeset = NULL;
		return NULL;
	}
	diframe = cpl_frameset_get_first( diframeset );
	iplane = 0;
	result = cpl_frameset_new();

	while ( diframe ){
		curdoubleimage = sph_double_image_load( cpl_frame_get_filename( diframe), 0);
		if ( !curdoubleimage ) {
			//SPH_ERR("Could not load a double image from the current diframe");
			SPH_ERROR_RAISE_INFO(SPH_ERROR_INFO, "All planes from the current qiframe  calibrated or there was a problem by , "
					"the number of the iplane = %d", iplane)
			//return NULL;
		}

		if ( master_bias_doubleimage ) {
		    SPH_INFO_MSG("Subtracting bias...");
			rerr = sph_double_image_subtract_double_image( curdoubleimage, master_bias_doubleimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting master bias:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curdoubleimage) {
					sph_double_image_delete( curdoubleimage );
					curdoubleimage = NULL;
				}
				if ( diframeset ) {
					cpl_frameset_delete(diframeset); diframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}


		if ( master_dark_doubleimage ) {
		    SPH_INFO_MSG("Subtracting dark...");
			rerr = sph_double_image_subtract_double_image( curdoubleimage, master_dark_doubleimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by subtracting dark:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curdoubleimage) {
					sph_double_image_delete( curdoubleimage );
					curdoubleimage = NULL;
				}
				if ( diframeset ) {
					cpl_frameset_delete(diframeset); diframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}


		if ( intensity_flat_doubleimage ) {
		    SPH_INFO_MSG("Applying intensity flat field...");
			rerr = sph_double_image_divide_double_image( curdoubleimage,intensity_flat_doubleimage);
			if ( rerr != CPL_ERROR_NONE ){
				sph_error_raise( SPH_ERROR_GENERAL,
						__FILE__, __func__, __LINE__,
						SPH_ERROR_ERROR,
						"Error is raised by applying (dividing) intensity flat:\n"
						"cpl error code is: %d", cpl_error_get_code() );
				if (curdoubleimage) {
					sph_double_image_delete( curdoubleimage );
					curdoubleimage = NULL;
				}
				if ( diframeset ) {
					cpl_frameset_delete(diframeset); diframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
				return NULL;
			}
		}

	    //quality check (calculate statistical moments for the basic imaging)
	    rerr = sph_double_image_quality_check( curdoubleimage );
	    if ( rerr ){
	        sph_error_raise( SPH_ERROR_GENERAL,
	                __FILE__, __func__, __LINE__,
	                SPH_ERROR_ERROR,
	                "sph_double_image_quality_check returns cpl error.\n"
	                "cpl error code is: %d", rerr );
			if (curdoubleimage) {
				sph_double_image_delete( curdoubleimage );
				curdoubleimage = NULL;
			}
			if ( diframeset ) {
				cpl_frameset_delete(diframeset); diframeset = NULL;
			}
			cpl_frameset_delete(inframeset); inframeset = NULL;
	        return NULL;
	    }

	    //save
	    /* Attention:
	     * The following function includes a step to update SPH_COMMON_KEYWORD_SPH_TYPE by
	     * SPH_COMMON_KEYWORD_VALUE_SPH_TYPE_DOUBLE_IMAGE. It is necessary because cpl_dsf_save_image
	     * updates the double image property list by the property list from the first input frames.
	     * In the case of some misbehavior exclude this step for checking.
	     */
		//create the name of the calibrated plane
	    SPH_INFO_MSG("Creating output filename...");
	    sprintf( outfilename, "%s", self->basic_imaging_outfilename );
		sprintf( outfilename_tmp, "_%s_plane-", basename( (char*)cpl_frame_get_filename( inframe ) ) );
		strcat( outfilename, outfilename_tmp );
		sprintf( outfilename_tmp, "%d.fits", iplane);
		strcat( outfilename, outfilename_tmp );
		SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "outfilename = %s", outfilename)
	    SPH_INFO_MSG("Saving...");
	    rerr = sph_double_image_save_dfs( curdoubleimage,
		                                  outfilename,
		                                  inframeset,
		                                  inframe,
		                                  self->inparams,
		                                  SPH_ZPL_TAG_BASIC_IMAGING_CALIB,
		                                  SPH_RECIPE_NAME_ZPL_BASIC_IMAGING,
		                                  SPH_PIPELINE_NAME_ZIMPOL,
		                                  pl);

	    if ( rerr ){
		       sph_error_raise( SPH_ERROR_GENERAL,
		              __FILE__, __func__, __LINE__,
		               SPH_ERROR_ERROR,
		               "Could not save basic imaging:\n"
		               "sph_double_image_save_dfs returns error code.\n"
		               "cpl error code is: %d", rerr );
				if (curdoubleimage) {
					sph_double_image_delete( curdoubleimage );
					curdoubleimage = NULL;
				}
				if ( diframeset ) {
					cpl_frameset_delete(diframeset); diframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
		       return NULL;
		}

	    outframe = cpl_frame_new();
	    cpl_frame_set_filename( outframe, outfilename );
	    cpl_frame_set_group( outframe, CPL_FRAME_GROUP_PRODUCT );
	    cpl_frame_set_tag( outframe, SPH_ZPL_TAG_BASIC_IMAGING_CALIB );
	    rerr = cpl_frameset_insert( result, outframe);

	    if ( rerr ){
		       sph_error_raise( SPH_ERROR_GENERAL,
		              __FILE__, __func__, __LINE__,
		               SPH_ERROR_ERROR,
		               "Couldn't insert frame into the result frameset\n"
		               "cpl error code is: %d", rerr );
				if (curdoubleimage) {
					sph_double_image_delete( curdoubleimage );
					curdoubleimage = NULL;
				}
				if ( diframeset ) {
					cpl_frameset_delete(diframeset); diframeset = NULL;
				}
				cpl_frameset_delete(inframeset); inframeset = NULL;
		       return NULL;
		}


		sph_double_image_delete( curdoubleimage ); curdoubleimage = NULL;
		diframe = cpl_frameset_get_next( diframeset) ;
		iplane = iplane + 1;
	}

	if ( diframeset ) {
		cpl_frameset_delete(diframeset); diframeset = NULL;
	}
	if (inframeset) {
		cpl_frameset_delete(inframeset); inframeset = NULL;
	}
	if ( result && cpl_frameset_get_size(result) == 0) {
    	cpl_frameset_delete(result); result = NULL;
    }

	return result;

}




cpl_error_code sph_zpl_basic_imaging_run( sph_zpl_basic_imaging* self )
{
    sph_double_image*				master_bias_doubleimage			= NULL;
    sph_double_image*				master_dark_doubleimage			= NULL;
    sph_double_image*				intensity_flat_doubleimage		= NULL;


    cpl_frame*			curframe				= NULL;
    cpl_frameset*		basic_imaging_frames	= NULL;

    int					recipe_error			= CPL_ERROR_NONE;


    SPH_INFO_MSG("Starting sph_zpl_basic_imaging_run...");
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE


    //if master bias exists load it
    if ( self->master_bias_frame ){
    	SPH_INFO_MSG("Loading master bias...\n");
        master_bias_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->master_bias_frame ), 0);
    }

    //if master dark exists load it
    if ( self->master_dark_frame){
       	SPH_INFO_MSG("Loading master dark...\n");
        master_dark_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->master_dark_frame ), 0);
    }

    //if intensity flat field exists load
    if ( self->intensity_flat_frame){
       	SPH_INFO_MSG("Loading intensity flat field...\n");
        intensity_flat_doubleimage = sph_double_image_load( cpl_frame_get_filename( self->intensity_flat_frame ), 0);
    }

    self->basic_imaging_frameset = cpl_frameset_new();
    curframe = cpl_frameset_get_first( self->rawframes );
    while ( curframe ) {
        SPH_INFO_MSG("Invoke sph_zpl_basic_imaging_create_basic...");
        SPH_ERROR_RAISE_INFO( SPH_ERROR_INFO, "curframe name = %s", cpl_frame_get_filename( curframe ));
        basic_imaging_frames = sph_zpl_basic_imaging_create_basic( self, curframe,
        						master_bias_doubleimage, master_dark_doubleimage, intensity_flat_doubleimage );

       	recipe_error |= sph_utils_append_frames( self->basic_imaging_frameset, basic_imaging_frames );
       	cpl_frameset_delete(basic_imaging_frames); basic_imaging_frames = NULL;
        curframe = cpl_frameset_get_next( self->rawframes );
    }


    if ( recipe_error ){
        sph_error_raise( SPH_ERROR_GENERAL,
                __FILE__, __func__, __LINE__,
                SPH_ERROR_ERROR,
                "Could not create basic imaging reduced data:\n"
                "cpl error code is: %d", recipe_error );
    }

    if (master_bias_doubleimage) {
    	sph_double_image_delete( master_bias_doubleimage ); master_bias_doubleimage = NULL;
    }
    if (master_dark_doubleimage) {
    	sph_double_image_delete( master_dark_doubleimage ); master_dark_doubleimage = NULL;
    }
    if (intensity_flat_doubleimage) {
    	sph_double_image_delete( intensity_flat_doubleimage ); intensity_flat_doubleimage = NULL;
    }

    SPH_INFO_MSG("Remove temporarily files: sph_filemanager_clean...");
    sph_filemanager_clean();

    SPH_INFO_MSG("sph_zpl_basic_imaging_run...End");
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}
/**@}*/
