/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_filemanager.h"
#include "sph_pixel_polyfit_table.h"
#include "sph_cube.h"
#include "sph_common_keywords.h"
#include "sph_create_super_flat.h"

/*----------------------------------------------------------------------------*/
/**
 @brief Create flat field polynomial pixel fit.

 @param shortdff        a frame of the short flat field to use for
                        normalisation (can be NULL)
 @param longdffs        a frameset containing long DFF frames
 @param order            the order for the fit
 @param    dit_id            the DIT id of the input flat cubes to use
 @param    czFrameOut        name of the output frame to create

 @return error code of the operaiton.

 Description: Create a polynomial fit to the wavelength dependence of the
 flat field for all pixels. The resulting frame is a sph_pixel_polyfit_table.
 The function creates the fit only for one specific DIT setting as given
 in the function call -- if not all input flat frames contain flats with the
 given DIT an error is returned and the fit table is not created.
 The short DFF is an optional input frame and can be NULL. If it is set,
 the broad band long flat (which has to be included in the longdffs
 frameset) is divided by the short DFF before the caluclated flat field fit
 is normalised (divided by the long braod band flat).

 */
/*----------------------------------------------------------------------------*/
static
sph_pixel_polyfit_table*
sph_create_flat_fit(         sph_master_frame* longdff1,
        sph_master_frame* longdff2,
        sph_master_frame* longdff3,
        sph_master_frame* longdff4,
        int order )
{
    sph_pixel_polyfit_table*        fittab        = NULL;
    cpl_vector*              lambdas = NULL;
    cpl_imagelist*              imlist = NULL;
    cpl_imagelist*              imcoeffs = NULL;
    cpl_image*              errimage = NULL;

    cpl_ensure( longdff1 && longdff2 && longdff3,
            CPL_ERROR_NULL_INPUT, NULL);

    cpl_ensure( cpl_error_get_code() == CPL_ERROR_NONE,
                    cpl_error_get_code(), NULL);

    cpl_ensure(cpl_propertylist_has(longdff1->properties,SPH_COMMON_KEYWORD_CALIB_LAMBDA),CPL_ERROR_ILLEGAL_INPUT,NULL);
    cpl_ensure(cpl_propertylist_has(longdff2->properties,SPH_COMMON_KEYWORD_CALIB_LAMBDA),CPL_ERROR_ILLEGAL_INPUT,NULL);
    cpl_ensure(cpl_propertylist_has(longdff3->properties,SPH_COMMON_KEYWORD_CALIB_LAMBDA),CPL_ERROR_ILLEGAL_INPUT,NULL);
    if ( longdff4 ) {
        cpl_ensure(cpl_propertylist_has(longdff4->properties,SPH_COMMON_KEYWORD_CALIB_LAMBDA),CPL_ERROR_ILLEGAL_INPUT,NULL);
        lambdas = cpl_vector_new(4);
    }
    else lambdas = cpl_vector_new(3);
    cpl_vector_set(lambdas,0,
            cpl_propertylist_get_double( longdff1->properties, SPH_COMMON_KEYWORD_CALIB_LAMBDA ));
    cpl_vector_set(lambdas,1,
            cpl_propertylist_get_double( longdff2->properties, SPH_COMMON_KEYWORD_CALIB_LAMBDA ));
    cpl_vector_set(lambdas,2,
            cpl_propertylist_get_double( longdff3->properties, SPH_COMMON_KEYWORD_CALIB_LAMBDA ));
    if ( longdff4 ) {
        cpl_vector_set(lambdas,3,
                cpl_propertylist_get_double( longdff4->properties, SPH_COMMON_KEYWORD_CALIB_LAMBDA ));
    }
    imlist = cpl_imagelist_new();
    cpl_imagelist_set(imlist,sph_master_frame_get_image(longdff1,1),0);
    cpl_imagelist_set(imlist,sph_master_frame_get_image(longdff2,1),1);
    cpl_imagelist_set(imlist,sph_master_frame_get_image(longdff3,1),2);
    if ( longdff4 ) {
        cpl_imagelist_set(imlist,sph_master_frame_get_image(longdff4,1),3);
    }


    errimage = cpl_image_new(cpl_image_get_size_x(cpl_imagelist_get(imlist,0)),
            cpl_image_get_size_y(cpl_imagelist_get(imlist,0)),CPL_TYPE_DOUBLE);
    imcoeffs = cpl_fit_imagelist_polynomial(lambdas,imlist,0,order,
            CPL_FALSE,CPL_TYPE_DOUBLE,errimage);
    cpl_ensure(cpl_error_get_code()== CPL_ERROR_NONE,cpl_error_get_code(),NULL);
    fittab = sph_pixel_polyfit_table_new_from_imlist( imcoeffs, errimage );
    sph_pixel_polyfit_table_save_cube(fittab,"fittab.fits");
    cpl_image_delete(errimage);
    cpl_imagelist_delete(imcoeffs);
    cpl_imagelist_delete(imlist);
    cpl_vector_delete(lambdas);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return fittab;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Create wavelength dependent flat field.

 @param pdtframe        a frame of the PDT to use for wavelength associaitons
 @param pftframe        a frame containing the flat polynomial fit
 @param czFrameOut        filename of the output frame
 @return

 Description: Create a wavelength dependent flat field from a PDT, a set of
 long DFF and a short DFF. Each pixel value is assigned a flat field value
 that is calculated from the wavelength associated with that pixel from the PDT
 and the value calculated from the pixel fit.

 */
/*----------------------------------------------------------------------------*/
sph_master_frame*
sph_create_super_flat( sph_pixel_description_table* pdt,
        sph_master_frame* longdff1,
        sph_master_frame* longdff2,
        sph_master_frame* longdff3,
        sph_master_frame* longdff4)
{
    sph_pixel_polyfit_table*        pft            = NULL;
    cpl_polynomial*                    poly        = NULL;
    int                                xx            = 0;
    int                                yy            = 0;
    double                            mse            = 0.0;
    double                            val            = 0.0;
    sph_pixel_descriptor*            pdescr        = NULL;
    sph_master_frame*                flat        = NULL;

    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
        return NULL;
    }

    if ( !pdt ||
            !longdff1 ||
            !longdff2 ||
            !longdff3 ) {
        return NULL;
    }

    pft = sph_create_flat_fit( longdff1, longdff2, longdff3,
            longdff4, 2 );
    if ( !pft ) {
        return NULL;
    }

    flat = sph_master_frame_new( pdt->nx, pdt->ny );
    if ( !flat ) {
        sph_pixel_polyfit_table_delete( pft );
        return NULL;
    }
    for (xx = 0; xx < pdt->nx; ++xx) {
        for (yy = 0; yy < pdt->ny; ++yy) {
            pdescr = sph_pixel_description_table_get_descriptor( pdt, xx, yy );
            if ( pdescr ) {
                if ( pdescr->specid ) {
                    poly = sph_pixel_polyfit_table_get( pft, xx, yy, &mse);
                    if ( poly ) {
                        val = cpl_polynomial_eval_1d( poly, pdescr->wavelength, NULL );
                        if ( cpl_error_get_code() == CPL_ERROR_NONE ) {
                            cpl_image_set( flat->image, xx + 1, yy + 1, val );
                            cpl_image_set( flat->rmsmap, xx + 1, yy + 1, mse );
                            cpl_image_set( flat->badpixelmap, xx + 1, yy + 1, 0 );
                            cpl_image_set( flat->ncombmap, xx + 1, yy + 1, 1 );
                        }
                        else {
                            SPH_RAISE_CPL_RESET
                            cpl_image_set( flat->image, xx + 1, yy + 1, 0.0 );
                            cpl_image_set( flat->rmsmap, xx + 1, yy + 1, 0.0 );
                            cpl_image_set( flat->badpixelmap, xx + 1, yy + 1, 4 );
                            cpl_image_set( flat->ncombmap, xx + 1, yy + 1, 0 );
                        }
                        cpl_polynomial_delete( poly ); poly = NULL;
                    }
                    else {
                        cpl_error_reset();
                        cpl_image_set( flat->image, xx + 1, yy + 1, 0.0 );
                        cpl_image_set( flat->rmsmap, xx + 1, yy + 1, 0.0 );
                        cpl_image_set( flat->badpixelmap, xx + 1, yy + 1, 3 );
                        cpl_image_set( flat->ncombmap, xx + 1, yy + 1, 0 );
                    }
                }
                else {
                    cpl_error_reset();
                    cpl_image_set( flat->image, xx + 1, yy + 1, 0.0 );
                    cpl_image_set( flat->rmsmap, xx + 1, yy + 1, 0.0 );
                    cpl_image_set( flat->badpixelmap, xx + 1, yy + 1, 2 );
                    cpl_image_set( flat->ncombmap, xx + 1, yy + 1, 0 );
                }
            }
        }
    }

    sph_pixel_polyfit_table_delete( pft ); pft = NULL;
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
        if ( flat ) sph_master_frame_delete(flat);
        flat = NULL;
        return NULL;
    }
    return flat;
}
