/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#include <cpl.h>
#include <fitsio.h>
#include <longnam.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "sph_master_frame.h"
#include "sph_fits.h"
#include "sph_cube.h"
#include "sph_error.h"
#include "sph_common_keywords.h"
#include "sph_pixel_descriptor.h"
#include "sph_pixel_description_table.h"
#include "sph_framelist.h"
#include "sph_pixel_polyfit_table.h"
#include "sph_smart_imagelist.h"
#include "sph_dataset.h"
#include "sph_dataset_stats.h"
#include "sph_keyword_manager.h"
#include "sph_utils.h"
#include "sph_filemanager.h"
#define SPH_CUBE_UNLINK_EXTS

/*----------------------------------------------------------------------------*/
/**
 * @defgroup sph_cube Master cube object
 * @par Synopsis:
 * @code
 * typedef struct _sph_cube_
 * {
 *    cpl_imagelist       *   imlist;
 *    cpl_table           *   assoc_table;
 *    cpl_type                type;
 *    sph_interpol_alg        interpol_alg;
 *    cpl_propertylist    *   proplist;
 *    const char          *   filename;
 *    short int               file_exists;
 *    short int               image_extno;
 *    short int               badpixel_extno;
 *    short int               rmsmap_extno;
 *    short int               ncombmap_extno;
 *    short int               tbl_extno;
 *    short int               nx;
 *    short int               ny;
 * } sph_cube;
 * @endcode
 * @par Desciption:
 * This module provides functions to define and operate on a cube with
 * three dimensions. The sph_cube holds information of a datacube with
 * a third axis (image planes) where the image planes can be separated
 * a non-constant step-size apart. In this way it differs from the
 * Euro3D cube defined in other ESO reduction software.
 * The information about the location of the third axis is saved in a
 * FITS table extension, which contains a list of the planes and the
 * associated z-value.
 * To retrieve information from the cube, some function methods exist that
 * use interpolation between planes if necessary.
 *
 */
/*----------------------------------------------------------------------------*/
/**@{*/
sph_error_code SPH_CUBE_GENERAL             = SPH_CUBE_ERR_START + 0;
sph_error_code SPH_CUBE_NO_IMAGE_EXT        = SPH_CUBE_ERR_START + 1;
sph_error_code SPH_CUBE_NO_BADPIXEL_EXT     = SPH_CUBE_ERR_START + 2;
sph_error_code SPH_CUBE_ERR_INVALID_FITMODE = SPH_CUBE_ERR_START + 3;
sph_error_code SPH_CUBE_FAILED_PIXHIST      = SPH_CUBE_ERR_START + 4;


/*-----------------------------------------------------------------------------
 Function definitions
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new sphere master cube object
 * @param filename      the filename to be used to store the cube
 *                      If the file already exists it is overwritten
 *                      as soon as the master cube is filled with data
 *
 * This creates a new empty sphere master cube associated to a given filename.
 * Note that the file is not created immediately after this function is called,
 * but rather only when the cube is filled with actual data, by appending images
 * or master frames for example.
 *
 * @return      pointer to the newly created sph_cube or null if an
 *              error occured
 */
/*----------------------------------------------------------------------------*/
sph_cube* sph_cube_new(const char* filename)
{
    sph_cube*           cube        = NULL;

    unlink(filename);
    cpl_ensure(filename, CPL_ERROR_NULL_INPUT, NULL);
    cube = cpl_calloc( 1, sizeof(sph_cube));
    cpl_ensure(cube != NULL, SPH_ERROR_MEMORY_FAIL, NULL);

    cube->assoc_table = cpl_table_new( 0);
    SPH_ERROR_ENSURE_GOTO_EXIT(cube->assoc_table , SPH_ERROR_MEMORY_FAIL);
    cube->proplist = cpl_propertylist_new();
    SPH_ERROR_ENSURE_GOTO_EXIT(cube->proplist, SPH_ERROR_MEMORY_FAIL);
    cube->type = CPL_TYPE_UNSPECIFIED;

    cube->filename = cpl_calloc(256,sizeof(char));
    sprintf(cube->filename,"%s",filename);

    cube->file_exists = 0;
    cube->nx = 0;
    cube->ny = 0;

    return cube;

EXIT:
	sph_cube_delete(cube); cube = NULL;
	return NULL;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Create a new sphere master cube object
 * @param filename      the filename to be used to store the cube
 *                      If the file already exists it is overwritten
 *                      as soon as the master cube is filled with data
 *
 * This creates a new empty sphere master cube associated to a given filename.
 * Note that the file is not created immediately after this function is called,
 * but rather only when the cube is filled with actual data, by appending images
 * or master frames for example.
 *
 * @return      pointer to the newly created sph_cube or null if an
 *              error occured
 */
/*----------------------------------------------------------------------------*/
sph_cube* sph_cube_new_dfs(const char* filename,
        cpl_frameset* allframes,
        cpl_frame* template_frame,
        cpl_parameterlist* params,
        const char* tag,
        const char* recipe,
        const char* pipename,
        cpl_propertylist* plist)
{
    sph_cube*           cube        = NULL;

    cpl_ensure(filename && pipename && recipe && tag, CPL_ERROR_NULL_INPUT, NULL);
    cube = sph_cube_new(filename);
    cpl_ensure(cube != NULL, SPH_ERROR_MEMORY_FAIL, NULL);
    cube->dfs_allframes = allframes;
    cube->dfs_template_frame = template_frame;
    cube->dfs_paramlist = params;
    cube->dfs_pipename = cpl_calloc(strlen(pipename)+1,sizeof(char));
    cube->dfs_recipe = cpl_calloc(strlen(recipe)+1,sizeof(char));
    cube->dfs_tag = cpl_calloc(strlen(tag)+1,sizeof(char));
    sprintf(cube->dfs_recipe,"%s",recipe);
    sprintf(cube->dfs_tag,"%s",tag);
    sprintf(cube->dfs_pipename,"%s",pipename);
    if ( plist ) {
        cpl_propertylist_append(cube->proplist,plist);
    }
    return cube;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Append an master frame to the cube
 * @param self      The cube
 * @param mframe    The master frame to append
 * @param zvalue    The zvalue of the plane where the image is located on
 *
 * This function appends a master frame to the data cube. It adds the image,
 * badpixelmap, ncombmap and the RMSMAP to the cube. Any extension that
 * does not exist yet is created.
 *
 * @return Error code of the operation
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_cube_append_master(sph_cube* self,
                           sph_master_frame* mframe,
                           float zvalue)
{
    int nx = 0;
    int ny = 0;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    cpl_ensure_code(self, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(mframe, CPL_ERROR_NULL_INPUT);

    nx = cpl_image_get_size_x(mframe->image);
    ny = cpl_image_get_size_y(mframe->image);

    if ( (self->nx != nx || self->ny != ny) && self->file_exists )
    {
    	 SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
    	        		"Dont have the same size of cube "
    	                "and master frame!");
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    if ( !self->file_exists )
    {
    	self->nx = nx;
    	self->ny = ny;
    	self->file_exists = 1;
    }
    sph_cube_append_image( self->filename,
            mframe->image,NULL,
            SPH_CUBE_IMAGE_EXT );
    sph_cube_append_image( self->filename,
            mframe->badpixelmap,NULL,
            SPH_CUBE_BADPIX_EXT );
    sph_cube_append_image( self->filename,
            mframe->rmsmap,NULL,
            SPH_CUBE_RMSMAP_EXT );
    sph_cube_append_image( self->filename,
            mframe->ncombmap,NULL,
            SPH_CUBE_NCOMBMAP_EXT );
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Append an image to the cube
 * @param fname      The filename of the file to append to
 * @param image    The image to append
 * @param pl        Propertylist to write (see note below!!)
 * @param ext    The extension of where the append the image to
 * This function appends an image to the data cube. Any extension that
 * does not exist yet is created.
 *
 * @return Error code of the operation
 * @note: the propertylist (if passed as non NULL) is written
 * directly without passing it through the keyword manager.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_cube_append_image(const char* fname,
                           cpl_image* image,
                           cpl_propertylist* pl,
                           int extend)
{
    int rerr = CPL_ERROR_NONE;
    int nx = 0;
    int ny = 0;
    char ftmpname[1024];
    char basename[1024];
    char tail[256];
    FILE* fp = NULL;
    int file_exists = 0;
    cpl_imagelist* imlist = NULL;
    cpl_ensure_code(fname, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(image, CPL_ERROR_NULL_INPUT);

    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);

    sph_filemanager_split(fname,basename,tail);
    sprintf(ftmpname,"%s_e%02d.%s",basename,extend,tail);
    fp = fopen(ftmpname,"r");
    if (fp) {
        fclose(fp); fp = NULL;
        file_exists = 1;
    }

    imlist =  cpl_imagelist_new();
    cpl_imagelist_set(imlist,image,0);

    if ( file_exists )
    {
        cpl_imagelist_save(imlist,ftmpname,CPL_TYPE_UNSPECIFIED,pl,CPL_IO_APPEND);
    }
    else {
        cpl_imagelist_save(imlist,ftmpname,CPL_TYPE_UNSPECIFIED,pl,CPL_IO_CREATE);
    }

    cpl_imagelist_unset(imlist,0);
    cpl_imagelist_delete(imlist); imlist = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Finalise all extensions saved to a cube
 * @param fname      The filename base of the saved imagelists
 *
 * Create a new multi-extension file from all files that have
 * the same basename (given as the fname parameter) but end with
 *  "_eXX.fits" where X is the extension number.
 *
 * @return Error code of the operation
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_cube_finalise_file(const char* fname)
{
    char* ftmpname = NULL;
    char* basename = NULL;
    char* tail = NULL;
    FILE* fp = NULL;
    int file_exists = 1;
    cpl_imagelist* imlist = NULL;
    cpl_imagelist* imlist2 = NULL;
    cpl_image*      im = NULL;
    int     extend = 0;
    int     nextend = 0;
    int     pp = 0;
    int     p0 = 0;
    cpl_propertylist* pl = NULL;
    cpl_propertylist* pl2 = NULL;
    cpl_ensure_code(fname, CPL_ERROR_NULL_INPUT);
    basename = cpl_calloc(strlen(fname) + 1,sizeof(char));
    tail = cpl_calloc(strlen(fname) + 1,sizeof(char));
    ftmpname = cpl_calloc(strlen(fname) + 40,sizeof(char));

    sph_filemanager_split(fname,basename,tail);
    while ( file_exists ) {
        file_exists = 0;
        sprintf(ftmpname,"%s_e%02d.%s",basename,nextend,tail);
        fp = fopen(ftmpname,"r");
        if (fp) {
            fclose(fp); fp = NULL;
            file_exists = 1;
            nextend++;
        }
    }
    fp = fopen(fname,"r");
    file_exists = 0;
    if (fp && nextend > 0 ) {
        fclose(fp); fp = NULL;
        if ( cpl_fits_count_extensions(fname) != nextend - 1) {
            SPH_ERROR_RAISE_ERR(
                    CPL_ERROR_INCOMPATIBLE_INPUT,
                    "A file %s already exists but has "
                    "different number of extensions(%d).", fname,
                    (int)cpl_fits_count_extensions(fname));
            goto EXIT;
        }
        file_exists = 1;
    }
    if ( file_exists ) {
        for (extend = 0; extend < nextend; ++extend) {
            sprintf(ftmpname,"%s_e%02d.%s",basename,extend,tail);
            imlist = cpl_imagelist_load(fname,CPL_TYPE_UNSPECIFIED,extend);
            pl = cpl_propertylist_load(fname,extend);
            imlist2 = cpl_imagelist_load(ftmpname,CPL_TYPE_UNSPECIFIED,0);
            pl2 = cpl_propertylist_load(ftmpname,0);
            SPH_ERROR_ENSURE_GOTO_EXIT(imlist2,cpl_error_get_code());
            SPH_ERROR_ENSURE_GOTO_EXIT(imlist,cpl_error_get_code());
            p0 = cpl_imagelist_get_size(imlist);
            for (pp = 0; pp < cpl_imagelist_get_size(imlist2); ++pp) {
                im = cpl_imagelist_get(imlist2,pp);
                cpl_imagelist_set(imlist,cpl_image_duplicate(im),pp + p0);
                SPH_ERROR_CHECK_STATE_ONERR_GOTO_EXIT;
            }
            cpl_imagelist_delete(imlist2);imlist2 = NULL;

            cpl_propertylist_append( pl, pl2 );
            cpl_imagelist_save(imlist,ftmpname,CPL_TYPE_UNSPECIFIED,pl,CPL_IO_CREATE);
            cpl_propertylist_delete(pl);pl = NULL;
            cpl_propertylist_delete(pl2);pl2 = NULL;
            cpl_imagelist_delete(imlist); imlist = NULL;
        }
    }
    for (extend = 0; extend < nextend; ++extend) {
        sprintf(ftmpname,"%s_e%02d.%s",basename,extend,tail);
        imlist = cpl_imagelist_load(ftmpname,CPL_TYPE_UNSPECIFIED,0);
        SPH_ERROR_ENSURE_GOTO_EXIT(imlist,cpl_error_get_code());

        pl = cpl_propertylist_load(ftmpname,0);
        if ( extend == 0 ) {
            cpl_imagelist_save(imlist,fname,CPL_TYPE_UNSPECIFIED,pl,CPL_IO_CREATE);
        }
        else cpl_imagelist_save(imlist,fname,CPL_TYPE_UNSPECIFIED,pl,CPL_IO_EXTEND);
        cpl_propertylist_delete(pl);pl = NULL;
        cpl_imagelist_delete(imlist); imlist = NULL;
#ifdef SPH_CUBE_UNLINK_EXTS
#pragma omp critical
        {
        	unlink(ftmpname);
        }
#endif
    }
EXIT:
    cpl_free(basename); basename = NULL;
    cpl_free(ftmpname); ftmpname = NULL;
    cpl_free(tail); tail = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Finalise all files in frameset
 * @param frameset
 *
 * @return error code of the oepration
 *
 * This function will finalise all filles that are inside the
 * given frameset and are valid for finalisation. The function
 * does this by just looking at all the filenames and
 * checking if corresponding "eXX" files exist. These
 * are then finalised.
 *
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_cube_finalise_all(cpl_frameset* frameset) {
    int         ff = 0;
    cpl_ensure_code(frameset,CPL_ERROR_NULL_INPUT);

    for (ff = 0; ff < cpl_frameset_get_size(frameset); ++ff) {
        sph_cube_finalise_file(
                cpl_frame_get_filename(
                        cpl_frameset_get_frame_const(
                                frameset,ff)
                                )
                         );
    }

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Append an master frame to the cube
 * @param self      The cube
 * @param dimage    The double image to append
 * @param zvalue    The zvalue of the plane where the image is located on
 *
 * This function appends a double_image to the data cube. It adds the image,
 * badpixelmap, ncombmap and the RMSMAP to the cube. Any extension that
 * does not exist yet is created.
 *
 * @return Error code of the operation
 *
 */
/*----------------------------------------------------------------------------*/
int sph_cube_append_double_image(sph_cube* self,
                                  sph_double_image* dimage,
                                  float zvalue)
{
    int nx = 0;
    int ny = 0;

    cpl_ensure_code(self != NULL && dimage != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(dimage->iframe != NULL && dimage->pframe != NULL, CPL_ERROR_NULL_INPUT);

    nx = cpl_image_get_size_x(dimage->iframe->image);
    ny = cpl_image_get_size_y(dimage->iframe->image);

    if ( (self->nx != nx || self->ny != ny) && self->file_exists )
    {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
        		"Dont have the same size of cube "
                "and double image!");
        return CPL_ERROR_ILLEGAL_INPUT;
    }
    if ( !self->file_exists )
    {
        self->nx = nx;
        self->ny = ny;
        self->file_exists = 1;

    }
    sph_cube_append_image( self->filename,
            dimage->iframe->image,NULL,
            SPH_CUBE_IMAGE_EXT );

    sph_cube_append_image( self->filename,
            dimage->iframe->badpixelmap,NULL,
            SPH_CUBE_BADPIX_EXT );

    sph_cube_append_image( self->filename,
            dimage->iframe->rmsmap, NULL,
            SPH_CUBE_RMSMAP_EXT );

    sph_cube_append_image( self->filename,
            dimage->iframe->ncombmap, NULL,
            SPH_CUBE_NCOMBMAP_EXT );


    sph_cube_append_image( self->filename,
            dimage->pframe->image, NULL,
            SPH_CUBE_IMAGE2_EXT );

    sph_cube_append_image( self->filename,
            dimage->pframe->badpixelmap, NULL,
            SPH_CUBE_BADPIX2_EXT );

    sph_cube_append_image( self->filename,
            dimage->pframe->rmsmap, NULL,
            SPH_CUBE_RMSMAP2_EXT );

    sph_cube_append_image( self->filename,
            dimage->pframe->ncombmap, NULL,
            SPH_CUBE_NCOMBMAP2_EXT );

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Delete the master cube.
  @param    self        the master cube to delete
  @return   the error code

  @note     A CPL error is raised in case there was a problem.

  Deallocate the master cube. The associated memory stored table, imagelist and
  propertylist are destroyed. The file or any extensions in it are NOT deleted.
 */
/*----------------------------------------------------------------------------*/
int sph_cube_delete(sph_cube* self)
{
    int rerr = CPL_ERROR_NONE;
    if (self) {
        if (self->assoc_table)
        {
            cpl_table_delete(self->assoc_table);
        }
        if ( self->filename) {
            cpl_free(self->filename);
        }
        if (self->imlist)
        {
            cpl_imagelist_delete(self->imlist);
        }

        if (self->proplist)
        {
            cpl_propertylist_delete(self->proplist);
        }
        if ( self->dfs_tag ) {
            cpl_free(self->dfs_tag); self->dfs_tag = NULL;
        }
        if ( self->dfs_pipename ) {
            cpl_free(self->dfs_pipename); self->dfs_pipename = NULL;
        }
        if ( self->dfs_recipe ) {
            cpl_free(self->dfs_recipe); self->dfs_recipe = NULL;
        }
        cpl_free( self );
    }
    return rerr;
}


/**@}*/
