/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#include <complex.h>
#include "sph_fft.h"
#include "sph_error.h"
#include "sph_common_keywords.h"
#include <string.h>
#include <fftw3.h>

static
sph_error_code
sph_fft_forward_row( sph_fft* self, int row ) {
    int    nx = self->nx;
    if ( self->method == SPH_FFT_GSL_RADIX2 ) {
        gsl_fft_complex_radix2_forward( &self->complexdata[ 2 * row * nx ], 1, nx );
    }
    else if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) {
        gsl_fft_complex_forward( &self->complexdata[ 2 * row * nx ], 1, nx, self->wavetab, self->workspace );
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_inverse_row( sph_fft* self, int row ) {
    int    nx = self->nx;
    if ( self->method == SPH_FFT_GSL_RADIX2 ) {
        gsl_fft_complex_radix2_inverse( &self->complexdata[ 2 * row * nx ], 1, nx );
    }
    else if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) {
        gsl_fft_complex_inverse( &self->complexdata[ 2 * row * nx ], 1, nx, self->wavetab, self->workspace );
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_forward_col( sph_fft* self, int col ) {
    int    nx = self->nx;
    int ny = self->ny;
    int    errcode = 0;
    if ( self->method == SPH_FFT_GSL_RADIX2 ) {
        gsl_fft_complex_radix2_forward( &self->complexdata[ 2 * col ], nx, ny );
    }
    else if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) {
        if ( ( errcode = gsl_fft_complex_forward( &self->complexdata[ 2 * col ], nx, ny, self->wavetab, self->workspace )) != 0 ) {
            sph_error_raise(SPH_ERROR_GENERAL,__FILE__,__func__,__LINE__,SPH_ERROR_ERROR,"Error from GSL: %s", gsl_strerror(errcode));
        }
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_inverse_col( sph_fft* self, int col ) {
    int    nx = self->nx;
    int ny = self->ny;
    if ( self->method == SPH_FFT_GSL_RADIX2 ) {
        gsl_fft_complex_radix2_inverse( &self->complexdata[ 2 * col ], nx, ny );
    }
    else if (self->method == SPH_FFT_GSL_MIXEDRADIX ) {
        gsl_fft_complex_inverse( &self->complexdata[ 2 * col ], nx, ny, self->wavetab, self->workspace );
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_shift_row( sph_fft* self, int row, double delta ) {
    int    nx = self->nx;
    int     xx    = 0;
    double    y0 =0.0;
    double    x0    = 0.0;
    double kval = 0.0;
    double    shift    = 0;
    double dshift = 0.0;
    gsl_complex            mulfac;
    gsl_complex            mulresult;
    gsl_complex*        valp    = NULL;

    y0 = ( self->ny - 1 ) / 2;
    x0 = ( self->nx - 1 ) / 2;
    for (xx = 0; xx < nx; ++xx) {
        shift = delta * ( row - y0 - 0.5 ) + self->shiftx;
        if ( xx <= x0 ) {
            kval = xx;
        }
        else {
            kval = xx - nx;
        }
        dshift = (double)shift * ( kval / (double)nx );
        mulfac = gsl_complex_polar( 1.0, 2.0 * SPH_PI * dshift  );
        valp = (gsl_complex*)&self->complexdata[ 2 * ( row * nx + xx ) ];
        mulresult = gsl_complex_mul( *valp, mulfac );
        valp->dat[0] = mulresult.dat[0];
        valp->dat[1] = mulresult.dat[1];
    }
    return CPL_ERROR_NONE;
}

static
sph_error_code
sph_fft_filter_row( sph_fft* self, int row, sph_fft_filter_method method, double delta, double param1, double param2 ) {
    int    nx = self->nx;
    int    ny = self->ny;
    int     xx    = 0;
    double    y0 =0.0;
    double    x0    = 0.0;
    double kvaly = 0.0;
    double kvalx = 0.0;
    double dshift = 0.0;
    double  r = 0.0;
    double factor = 0.0;
    double order = 0.0;
    gsl_complex*        valp    = NULL;

    y0 = ( self->ny - 1 ) / 2;
    x0 = ( self->nx - 1 ) / 2;
    if ( row <= y0 ) {
        kvaly = row;
    }
    else {
        kvaly = row - ny;
    }
    for (xx = 0; xx < nx; ++xx) {
        if ( xx <= x0 ) {
            kvalx = xx;
        }
        else {
            kvalx = xx - nx;
        }
        dshift =  kvalx / (double)nx * kvalx / (double)nx +
                    kvaly / (double)ny * kvaly / (double)ny;
        valp = (gsl_complex*)&self->complexdata[ 2 * ( row * nx + xx ) ];

        if ( method == SPH_FFT_FILTER_METHOD_TH ) {
            if ( 4.0 * dshift > 1.0 - delta * delta ) {
                valp->dat[0] = 0.0;//exp( -dshift/(delta*delta));
                valp->dat[1] = 0.0;//exp( -dshift/(delta*delta));
            }
        }
        if ( method == SPH_FFT_FILTER_METHOD_FERMI ) {
            r = 1.0 - delta * delta;
            factor = 1.0 / ( 1 + exp( (4 * dshift - r) / param1 ) );
            valp->dat[0] = valp->dat[0] * factor;
            valp->dat[1] = valp->dat[1] * factor;
        }
        if ( method == SPH_FFT_FILTER_METHOD_BUTTER ) {
            order = 2 * log( 0.882 / sqrt(10.624 * 10.624 - 1.0) ) / log( param1 / param2 );
            r = param1 / pow(0.882, 2.0 / order);
            factor = 1.0 / sqrt( 1 + pow(4 * dshift / r, order ) );
            valp->dat[0] = valp->dat[0] * factor;
            valp->dat[1] = valp->dat[1] * factor;
        }
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_shift_col( sph_fft* self, int col, double delta ) {
    int    nx = self->nx;
    int ny = self->ny;
    int     yy    = 0;
    gsl_complex            mulfac;
    gsl_complex            mulresult;
    gsl_complex*        valp    = NULL;
    double    x0;
    double y0;
    double    shift    = 0;
    double kval = 0.0;
    double dshift = 0.0;

    x0 = ( self->nx - 1 ) / 2;
    y0 = ( self->ny - 1 ) / 2;
    for (yy = 0; yy < ny; ++yy) {
        shift = delta * ( col - x0 - 0.5 ) + self->shifty;
        if ( yy <= y0 ) {
            kval = yy;
        }
        else {
            kval = yy - ny;
        }
        dshift = (double)shift * ( kval / (double)ny );
        mulfac = gsl_complex_polar( 1.0, 2.0 * SPH_PI * dshift  );
        valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * nx + col ) ];
        mulresult = gsl_complex_mul( *valp, mulfac );
        valp->dat[0] = mulresult.dat[0];
        valp->dat[1] = mulresult.dat[1];
    }
    return CPL_ERROR_NONE;
}
static
sph_error_code
sph_fft_prepare_gsl_radix2( sph_fft* self,cpl_image* im, int newx, int newy) {
    cpl_image*            workim = NULL;
    int                    bpix = 0;
    int                    N        = 0;
    int                    xx = 0;
    int                    yy = 0;
    sph_error_code        rerr = CPL_ERROR_NONE;

    N = cpl_image_get_size_x(im);

    workim = cpl_image_new( newx, newy, cpl_image_get_type(im) );
    if ( !workim ) return SPH_ERROR_GENERAL;
    if ( self->complexdata ) cpl_free(self->complexdata);
    self->complexdata = NULL;
    self->complexdata = cpl_calloc(2 * newx * newy, sizeof(double) );

    if ( !self->complexdata ) return SPH_ERROR_GENERAL;

    SPH_RAISE_CPL_RESET;
    rerr = cpl_image_copy( workim, im, ( newx - N ) / 2 + 1, ( newy - N ) / 2 + 1 );
    SPH_RAISE_CPL_RESET;

    for (xx = 0; xx < newx; ++xx) {
        for (yy = 0; yy < newy; ++yy) {
            self->complexdata[2 * (yy * newx + xx) ] = cpl_image_get( workim, xx + 1, yy + 1, &bpix );
        }
    }
    if ( workim ) {
      cpl_image_delete( workim );
      workim = NULL;
    }
    self->nx = newx;
    self->ny = newy;
    return CPL_ERROR_NONE;
}
static
void
sph_fft_alloc_mixedradix(sph_fft* self) {
    if ( self->wavetab ) gsl_fft_complex_wavetable_free(self->wavetab);
    if ( self->workspace ) gsl_fft_complex_workspace_free(self->workspace);
    self->wavetab = gsl_fft_complex_wavetable_alloc( self->nx );
    self->workspace = gsl_fft_complex_workspace_alloc( self->nx );
}

static
sph_error_code
sph_fft_prepare_gsl_mixed_radix( sph_fft* self,cpl_image* im ) {
    int                    bpix = 0;
    int                    N        = 0;
    int                    xx = 0;
    int                    yy = 0;

    N = cpl_image_get_size_x(im);
    if ( cpl_image_get_size_y(im) != N ) {
        SPH_ERR("Not the same size in x and y of image in FFT.");
        return CPL_ERROR_ILLEGAL_INPUT;
    }

    if ( self->complexdata ) {
        cpl_free(self->complexdata);self->complexdata = NULL;
    }
    self->complexdata = cpl_calloc(2 * N * N, sizeof(double) );

    if ( !self->complexdata ) return SPH_ERROR_GENERAL;
    for (xx = 0; xx < N; ++xx) {
        for (yy = 0; yy < N; ++yy) {
            self->complexdata[2 * (yy * N + xx) ] = cpl_image_get( im, xx + 1, yy + 1, &bpix );
        }
    }
    self->nx = N;
    self->ny = N;
    sph_fft_alloc_mixedradix(self);
    return CPL_ERROR_NONE;
}
sph_fft*
sph_fft_new( sph_fft_method method ) {
    sph_fft*     result = NULL;
    result = cpl_calloc( 1, sizeof(sph_fft) );
    result->method = method;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Smooth the image
 * @param image
 * @param fwhm
 * @return error code
 *
 * This smoothes the image with a gauss of the
 * given width.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_smooth_image( cpl_image** pimage, double fwhm) {
    cpl_image*  im = NULL;
    cpl_image*  image = NULL;
    cpl_image*  imc = NULL;
    double      x0 = 0.0;
    double      y0 = 0.0;
    double      sigma = 0.0;
    sph_fft*        fft = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    image = *pimage;
    cpl_ensure_code(image,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(fwhm > 0,
            CPL_ERROR_ILLEGAL_INPUT);
    sigma = fwhm / 2.3548;
    x0 = (double)cpl_image_get_size_x(image) / 2.0;
    y0 = (double)cpl_image_get_size_y(image) / 2.0;
    fft = sph_fft_new(SPH_FFT_FFTW3_DP);
    im = cpl_image_new(cpl_image_get_size_x(image),
    		cpl_image_get_size_y(image),CPL_TYPE_DOUBLE);
    cpl_image_fill_gaussian(im,x0 + 1,y0 + 1,1.0,sigma,sigma);
    cpl_image_multiply_scalar(im,1.0/cpl_image_get_absflux(im));
    imc = sph_fftw_convolve(fft,image,im);
    if (imc) {
        cpl_image_delete(image);
        *pimage = imc;
    }
    else {
        SPH_ERROR_RAISE_ERR(SPH_ERROR_GENERAL,"Could not"
                "smooth image");
    }
    cpl_image_delete(im); im = NULL;
    sph_fft_delete(fft);
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Perform a skew only
 * @param self
 * @param im the input image
 * @param degree    the angle of the skew in degrees
 *
 * @return error code of the operation
 *
 * This performs a skew in x ( rows only shifted). The operation
 * corresponds to
 * [ 1    - tan (angle / 2 ) ]     *  [ x ]
 * [ 0    1                  ]     *  [ y ]
 *
 * Use the function sph_fft_rotate_complex2image to return the image.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_skewx( sph_fft* self, cpl_image* im, double degrees ) {
    int                    N        = 0;
    double                angle    = degrees / 180.0 * SPH_PI;
    double                tantananglehalf    = tan( angle / 2.0 ) * tan( angle / 2.0 );
    int                    newx    = 2;
    int                    newy    = 2;
    int                    xx        = 0;
    int                    yy        = 0;
    int                    bpix    = 0;
    double                delta    = 0.0;
    cpl_image*            workim    = NULL;
    sph_error_code        rerr    = CPL_ERROR_NONE;


    if ( !im ) return SPH_ERROR_GENERAL;
    if ( degrees > 90.0 || degrees < -90.0 ) {
        return SPH_ERROR_GENERAL;
    }
    if ( cpl_image_get_size_x(im) != cpl_image_get_size_y(im) ) {
        return SPH_ERROR_GENERAL;
    }
    N = cpl_image_get_size_x( im );
    self->angle    = angle;
    while ( newx <= N + N * tantananglehalf ) {
        newx = newx * 2;
    }
    newy = newx;
    if ( newx > 2048 * 4 || newy > 2048 * 4 ) {
        return SPH_ERROR_GENERAL; // this is too big to handle
    }

    workim = cpl_image_new( newx, newy, CPL_TYPE_DOUBLE );
    if ( !workim ) return SPH_ERROR_GENERAL;
    self->complexdata = cpl_calloc(2 * newx * newy, sizeof(double) );
    if ( !self->complexdata ) return SPH_ERROR_GENERAL;

    rerr = cpl_image_copy( workim, im, ( newx - N ) / 2 + 1, ( newy - N ) / 2 + 1);

    for (xx = 0; xx < newx; ++xx) {
        for (yy = 0; yy < newy; ++yy) {
            self->complexdata[2 * (yy * newx + xx) ] = cpl_image_get( workim, xx + 1, yy + 1, &bpix );
        }
    }
    self->Npp = newy * 2;
    self->Np = newx;
    self->nx = newx;
    self->ny = newy;

    delta = -tan(angle/2.0);
    for (yy = 0; yy < newy; ++yy) {
        sph_fft_forward_row( self, yy );
        sph_fft_shift_row( self, yy, delta );
        sph_fft_inverse_row( self, yy );
    }
    cpl_image_delete(workim); workim = NULL;
    return rerr;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Perform a skew only
 * @param self
 * @param im the input image
 * @param degree    the angle of the skew in degrees
 *
 * @return error code of the operation
 *
 * This performs a skew in y ( columns only shifted). The operation
 * corresponds to
 * [ 1                 0 ]     *  [ x ]
 * [ -tan(angle/2)    1 ]     *  [ y ]
 *
 * Use the function sph_fft_rotate_complex2image to return the image.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_skewy( sph_fft* self, cpl_image* im, double degrees ) {
    int                    N        = 0;
    double                angle    = degrees / 180.0 * SPH_PI;
    double                tantananglehalf    = tan( angle / 2.0 ) * tan( angle / 2.0 );
    int                    newx    = 2;
    int                    newy    = 2;
    int                    xx        = 0;
    int                    yy        = 0;
    int                    bpix    = 0;
    double                delta    = 0.0;
    cpl_image*            workim    = NULL;
    sph_error_code        rerr    = CPL_ERROR_NONE;


    if ( !im ) return SPH_ERROR_GENERAL;
    if ( degrees > 90.0 || degrees < -90.0 ) {
        return SPH_ERROR_GENERAL;
    }
    if ( cpl_image_get_size_x(im) != cpl_image_get_size_y(im) ) {
        return SPH_ERROR_GENERAL;
    }
    N = cpl_image_get_size_x( im );
    self->angle    = angle;
    while ( newx <= N + N * tantananglehalf ) {
        newx = newx * 2;
    }
    newy = newx;
    if ( newx > 2048 * 4 || newy > 2048 * 4 ) {
        return SPH_ERROR_GENERAL; // this is too big to handle
    }

    workim = cpl_image_new( newx, newy, CPL_TYPE_DOUBLE );
    if ( !workim ) return SPH_ERROR_GENERAL;
    self->complexdata = cpl_calloc(2 * newx * newy, sizeof(double) );
    if ( !self->complexdata ) return SPH_ERROR_GENERAL;

    rerr = cpl_image_copy( workim, im, ( newx - N ) / 2 + 1, ( newy - N ) / 2 + 1 );
    //rerr = cpl_image_copy( workim, im, 1, 1 );

    for (xx = 0; xx < newx; ++xx) {
        for (yy = 0; yy < newy; ++yy) {
            self->complexdata[2 * (yy * newx + xx) ] = cpl_image_get( workim, xx + 1, yy + 1, &bpix );
        }
    }
    self->Npp = newy * 2;
    self->Np = newx;
    self->nx = newx;
    self->ny = newy;

    delta = -tan(angle/2.0);
    for (xx = 0; xx < newx; ++xx) {
        sph_fft_forward_col( self, xx );
        sph_fft_shift_col( self, xx, delta );
        sph_fft_inverse_col( self, xx );
    }
    cpl_image_delete(workim); workim = NULL;
    return rerr;
}
sph_error_code
sph_fft_set_complex(sph_fft* self, int kx ,int ky, double real, double imag) {
    int xx = 0;
    int yy = 0;
    double x0 = 0.0;
    double y0 = 0.0;
    gsl_complex*        valp    = NULL;

    y0 = ( self->ny - 1 ) / 2;
    x0 = ( self->nx - 1 ) / 2;

    if ( kx < 0 ) {
        xx = self->nx + kx;
    }
    else {
        xx = kx;
    }
    if ( ky < 0 ) {
        yy = self->ny + ky;
    }
    else {
        yy = ky;
    }
    if ( xx >= 0 && yy >=0 && yy < self->ny && xx < self->nx ) {
        valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * self->nx + xx ) ];
        valp->dat[0] = real;
        valp->dat[1] = imag;
    }
    else return CPL_ERROR_ILLEGAL_INPUT;

    return CPL_ERROR_NONE;

}
gsl_complex
sph_fft_get_complex(sph_fft* self, int kx ,int ky ) {
    int xx = 0;
    int yy = 0;
    double x0 = 0.0;
    double y0 = 0.0;
    gsl_complex*        valp    = NULL;
    gsl_complex result;
    y0 = ( self->ny - 1 ) / 2;
    x0 = ( self->nx - 1 ) / 2;

    result.dat[0]=0.0;
    result.dat[1]=0.0;

    if ( kx < 0 ) {
        xx = self->nx + kx;
    }
    else {
        xx = kx;
    }
    if ( ky < 0 ) {
        yy = self->ny + ky;
    }
    else {
        yy = ky;
    }
    if ( xx >= 0 && yy >=0 && yy < self->ny && xx < self->nx ) {
        valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * self->nx + xx ) ];
        result.dat[0] = valp->dat[0];
        result.dat[1] = valp->dat[1];
    }
    else {
        SPH_ERR("Access out of bounds.");
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Perform a forward FFT
 * @param self
 *
 * @return error code of the operation
 *
 * This performs a forward FFT.
 * The sph_fft object must have been prepared using
 * one of the the sph_fft_prepare function. The resulting image
 * can be obtained with the sph_fft_complex2image functions.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_forward( sph_fft* self ) {
    int        kk        = 0;


    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_forward_row( self, kk );
    }
    for (kk = 0; kk < self->nx; ++kk) {
        sph_fft_forward_col( self, kk );
    }
    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Perform an inverse FFT
 * @param self
 *
 * @return error code of the operation
 *
 * This performs an inverse FFT.
 * The sph_fft object must have been prepared using
 * one of the the sph_fft_prepare function. The resulting image
 * can be obtained with the sph_fft_complex2image functions.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_inverse( sph_fft* self ) {
    int        kk        = 0;


    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_inverse_col( self, kk );
    }
    for (kk = 0; kk < self->nx; ++kk) {
        sph_fft_inverse_row( self, kk );
    }
    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Perform a shifting using FFT of a cpl_image
 * @param self
 * @param dx
 * @param dy
 *
 * @return shifted images or NULL
 *
 * This creates a new shifted version of the input image.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_shift_image( sph_fft* self, cpl_image* im, double dx, double dy ) {
    cpl_image*    tim = NULL;
    fftw_complex*           in = NULL;
    fftw_plan               p;
    int                     N = 0;
    int                     xx = 0;
    int                     yy = 0;

    if ( self->method == SPH_FFT_GSL_RADIX2 ||
            self->method == SPH_FFT_GSL_MIXEDRADIX )
    {
        sph_fft_prepare(self,im);
        sph_fft_forward(self);
        sph_fft_shift(self,dx,dy);
        sph_fft_inverse(self);
        if ( self->method == SPH_FFT_GSL_RADIX2 ) {
            sph_fft_crop_complex_array(self,self->nx/2,0);
        }
        tim = sph_fft_complex2image_real(self);
    }
    else if ( self->method == SPH_FFT_FFTW3_DP ) {
        N = cpl_image_get_size_x(im);
        cpl_ensure(cpl_image_get_size_y(im) == N,CPL_ERROR_ILLEGAL_INPUT,NULL);
        in = sph_fft_create_fftw_complex(self,im,N,N);
        p = fftw_plan_dft_2d(N,N,in,in,
                FFTW_FORWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        sph_fft_shift_fftw_array(self,in,dx,dy);
        p = fftw_plan_dft_2d(N,N,in,in,
                FFTW_BACKWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        tim = cpl_image_duplicate(im);
        for (xx = 0; xx < N; ++xx) {
            for (yy = 0; yy < N; ++yy) {
                cpl_image_set(tim,xx+1,yy+1,creal(in[yy * N + xx])/ ( N * N ) );
            }
        }
        fftw_free(in); in = NULL;
    }
    else {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_UNSUPPORTED_MODE,
                "This FFT mode is not supported for shifting");
    }
    return tim;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Perform a shifting using FFT of a cpl_image
 * @param self
 * @param dx
 * @param dy
 *
 * @return shifted images or NULL
 *
 * This creates a new shifted version of the input image.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_freq_filter( sph_fft* self, cpl_image* im,
        sph_fft_filter_method method, double radius,
        double param1, double param2 )
{
    cpl_image*    tim = NULL;
    fftw_complex*           in = NULL;
    fftw_plan               p;
    int                     N = 0;
    int                     xx = 0;
    int                     yy = 0;
    int                     kk = 0;

    if ( self->method == SPH_FFT_GSL_RADIX2 ||
            self->method == SPH_FFT_GSL_MIXEDRADIX )
    {
        sph_fft_prepare(self,im);
        for (kk = 0; kk < self->nx; ++kk) {
            sph_fft_forward_col( self, kk );
        }
        for (kk = 0; kk < self->ny; ++kk) {
            sph_fft_forward_row( self, kk );
            sph_fft_filter_row( self, kk, method, radius, param1, param2 );
            sph_fft_inverse_row( self, kk );
        }
        for (kk = 0; kk < self->nx; ++kk) {
            sph_fft_inverse_col( self, kk );
        }
        if ( self->method == SPH_FFT_GSL_RADIX2 ) {
            sph_fft_crop_complex_array(self,self->nx/2,0);
        }
        tim = sph_fft_complex2image_real(self);
    }
    else if ( self->method == SPH_FFT_FFTW3_DP ) {
        N = cpl_image_get_size_x(im);
        cpl_ensure(cpl_image_get_size_y(im) == N,CPL_ERROR_ILLEGAL_INPUT,NULL);
        in = sph_fft_create_fftw_complex(self,im,N,N);
        p = fftw_plan_dft_2d(N,N,in,in,
                FFTW_FORWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        sph_fft_filter_fftw_array(self,in,method,radius,param1,param2);
        p = fftw_plan_dft_2d(N,N,in,in,
                FFTW_BACKWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        tim = cpl_image_duplicate(im);
        for (xx = 0; xx < N; ++xx) {
            for (yy = 0; yy < N; ++yy) {
                cpl_image_set(tim,xx+1,yy+1,creal(in[yy * N + xx])/ ( N * N ) );
            }
        }
        fftw_free(in); in = NULL;
    }
    else {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_UNSUPPORTED_MODE,
                "This FFT mode is not supported for shifting");
    }
    return tim;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Perform the scaling algorithm
 * @param self
 *
 * @return error code of the operation
 *
 * This performs the actual scaling.
 * The sph_fft object must have been prepared using
 * the sph_fft_scale_prepare function. The resulting image
 * can be obtained with the sph_rotate_complex2image function.
 *
 * See sph_fft_scale_prepare for more info.
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_scale( sph_fft* self, cpl_image* im, double scale ) {
    int nxreal        = 0;
    int nxfreq        = 0;
    int    xx            = 0;
    int    yy            = 0;
    int    nx            = 0;
    int    ny            = 0;
    double fact     = 0.0;
    double    dshift     = 0.0;
    cpl_image*        imout = NULL;
    sph_error_code    rerr = CPL_ERROR_NONE;
    fftw_complex*           in = NULL;
    fftw_complex*           arr2 = NULL;
    fftw_plan               p;
    double                  acc = 0.0;
    double              minacc  = 1000000.0;
    if ( !im ) {
        SPH_NO_SELF;
        return NULL;
    }
    nx = cpl_image_get_size_x(im);
    ny = cpl_image_get_size_y(im);
    acc = 0.5 / ( scale * nx );
    if ( nx != ny) {
        SPH_ERR("X and Y dimensions of image for scale operations must be same.");
        return NULL;
    }

    for (xx = nx; xx < 3 * nx; xx+=2) {
        for (yy = nx; yy < 3*nx; yy+=2) {
            fact = (double)xx/(double)yy;
            if ( fabs( fact - scale) < minacc ) {
                nxreal = yy;
                nxfreq = xx;
                minacc = fabs( fact-scale);
            }
        }
    }
    if ( nxreal < 1 || nxfreq < 1 ) {
        SPH_ERROR_RAISE_ERR(SPH_ERROR_GENERAL,
                "Could not obtain good padding sizes for accuracy %f.",
                acc);
        return NULL;
    }
    else {
        sph_error_raise(SPH_ERROR_GENERAL,__FILE__,__func__,
                __LINE__,SPH_ERROR_INFO,
                "Am scaling from %d to %d pixels in real "
                "space and %d pixels in freqency space, acc was %f, scaling ac: %f.",nx,nxreal,nxfreq,acc,minacc);
    }
    if ( self->method == SPH_FFT_GSL_MIXEDRADIX ||
            self->method == SPH_FFT_GSL_RADIX2 )
    {
        rerr = sph_fft_prepare(self,im);
        rerr |= sph_fft_pad_complex_array_real_space(self,nxreal);
        rerr |= sph_fft_forward(self);
        if ( nxfreq < nxreal ) {
            rerr |= sph_fft_crop_complex_array(self,nxfreq,1);
        }
        else {
            rerr |= sph_fft_pad_complex_array_freq_space(self,nxfreq);
        }
        dshift = ( 1.0 - (double)nxfreq/(double)nxreal ) * -0.5;
        sph_error_raise(SPH_ERROR_GENERAL,__FILE__,__func__,
                __LINE__,SPH_ERROR_INFO,
                "Am using a phase shift of %f pixels.",dshift);
        rerr |= sph_fft_shift(self,dshift,dshift);
        rerr |= sph_fft_inverse(self);
        rerr |= sph_fft_crop_complex_array(self,nx,0);
        if ( rerr == CPL_ERROR_NONE ) {
            imout = sph_fft_complex2image_real(self);
        }
        else {
            SPH_ERR("Problem when performing scale operations.");
            return NULL;
        }
    }
    else if ( self->method == SPH_FFT_FFTW3_DP ) {
        in = sph_fft_create_fftw_complex(self,im,nxreal,nxreal);
        p = fftw_plan_dft_2d(nxreal,nxreal,in,in,
                FFTW_FORWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        dshift = ( 1.0 - (double)nxfreq/(double)nxreal ) * -0.5;
        if ( nxfreq < nxreal ) {
            arr2 = sph_fft_crop_fftw_array(self,in,nxfreq,1);
            fftw_free(in);
            in = arr2;arr2 = NULL;
        }
        else {
            arr2 = sph_fft_pad_fftw_array_freq_space(self,in,nxfreq);
            fftw_free(in);
            in = arr2;arr2 = NULL;
        }
        sph_fft_shift_fftw_array(self,in,dshift,dshift);
        p = fftw_plan_dft_2d(self->nx,self->ny,in,in,
                FFTW_BACKWARD,FFTW_ESTIMATE);
        fftw_execute(p);
        fftw_destroy_plan(p);
        arr2 = sph_fft_crop_fftw_array(self,in,nx,0);
        fftw_free(in);
        in = arr2;arr2 = NULL;
        imout = cpl_image_new(self->nx, self->ny, cpl_image_get_type(im));
        for (xx = 0; xx < self->nx; ++xx) {
            for (yy = 0; yy < self->ny; ++yy) {
                cpl_image_set(imout,xx+1,yy+1,(double)self->nx / (double)nxreal * (double)self->nx / (double)nxreal * creal(in[yy * self->nx + xx])/(scale * scale * self->nx * self->nx));
            }
        }
        fftw_free(in);
    }
    return imout;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Rotate an image
 * @param self      sph_fft workspace
 * @param image     input image (unchanged)
 * @param angle     rotation angle in degrees
 *
 * @return rotated image or NULL on error
 *
 * @see sph_fft_rotate_prepare
 *
 * This function is a top level wrapper for the other rotation
 * fft functions. It rotates a copy of an input image
 * using FFT and returns it.
 * @Note: this function returns and error message if the sph_fft
 * was not initialised with SPH_FFT_GSL_RADIX2.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_rotate_image( sph_fft* self, cpl_image* image, double degrees ) {
    cpl_image*            result    = NULL;
    sph_error_code        rerr    = CPL_ERROR_NONE;
    cpl_image*              im = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    if ( self->method == SPH_FFT_FFTW3_DP ) {
        return sph_fftw_rotate_image(self,image,degrees);
    }

    cpl_ensure( self->method == SPH_FFT_GSL_RADIX2, CPL_ERROR_ILLEGAL_INPUT, NULL);

    im = cpl_image_duplicate(image);
    if ( degrees > 270.0 ) {
        cpl_image_turn( im, -3 );
        degrees = degrees - 270.0;
    }
    if ( degrees > 180.0 ) {
        cpl_image_turn( im, -2 );
        degrees = degrees - 180.0;
    }
    if ( degrees > 90.0 ) {
        cpl_image_turn( im, -1 );
        degrees = degrees - 90.0;
    }
    if ( degrees < -270.0 ) {
        cpl_image_turn( im, 3 );
        degrees = degrees + 270.0;
    }
    if ( degrees < -180.0 ) {
        cpl_image_turn( im, 2 );
        degrees = degrees + 180.0;
    }
    if ( degrees < -90.0 ) {
        cpl_image_turn( im, 1 );
        degrees = degrees + 90.0;
    }
    rerr = sph_fft_rotate_prepare( self, im, -degrees, 0.0, 0.0 );
    if ( rerr != CPL_ERROR_NONE ) {
        cpl_image_delete(im);
        return NULL;
    }
    rerr = sph_fft_rotate(self);
    if ( rerr != CPL_ERROR_NONE ) {
        cpl_image_delete(im);
        return NULL;
    }
    result = sph_fft_rotate_complex2image(self);
    cpl_image_delete(im);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}


cpl_image*
sph_fftw_convolve( sph_fft* self, cpl_image* image, cpl_image* conv ) {
    cpl_image*            result    = NULL;
    fftw_complex*           out = NULL;
    fftw_complex*           outconv = NULL;
    fftw_plan               p;
    int                     xx  = 0;
    int                     yy  = 0;
    double*                 inarr0 = NULL;
    double*                 inarrconv0 = NULL;
    double*                 inarr = NULL;
    double*                 inarrconv = NULL;
    int                     nx       = 0;
    int                     ny       = 0;
    int                     nxconv       = 0;
    int                     nyconv       = 0;

    cpl_ensure( self,CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure( self->method == SPH_FFT_FFTW3_DP, CPL_ERROR_ILLEGAL_INPUT, NULL);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    cpl_ensure(image,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    cpl_ensure(conv,CPL_ERROR_ILLEGAL_OUTPUT,NULL);

    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);

    nxconv = cpl_image_get_size_x(conv);
    nyconv = cpl_image_get_size_y(conv);

    cpl_ensure(nxconv == nx, CPL_ERROR_ILLEGAL_INPUT, NULL);
    cpl_ensure(nyconv == ny, CPL_ERROR_ILLEGAL_INPUT, NULL);

    inarr0 = cpl_image_get_data_double(image);
    inarrconv0 = cpl_image_get_data_double(conv);

    inarr = cpl_calloc(nx * ny, sizeof(double));
    inarrconv = cpl_calloc(nx * ny, sizeof(double));

    out = (fftw_complex*)
            fftw_malloc(sizeof(fftw_complex) * nx * ny);
    outconv = (fftw_complex*)
            fftw_malloc(sizeof(fftw_complex) * nx * ny);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            outconv[yy * nx + xx] = out[yy * nx + xx] = 0.0;
        }
    }

    p = fftw_plan_dft_r2c_2d(ny,nx,inarr,out,FFTW_ESTIMATE);
    for (xx = nx / 2; xx < nx; ++xx) {
        for (yy = ny / 2; yy < ny; ++yy) {
            inarr[yy * nx + xx] = inarrconv0[(yy - ny / 2) * nx + xx - nx / 2]
                                         ;
        }
        for (yy = 0; yy < ny / 2; ++yy) {
            inarr[yy * nx + xx] = inarrconv0[(yy + ny / 2) * nx + xx - nx / 2]
                                         ;
        }
    }
    for (xx = 0; xx < nx / 2; ++xx) {
        for (yy = 0; yy < ny / 2; ++yy) {
            inarr[yy * nx + xx] = inarrconv0[(yy + ny / 2) * nx + xx + nx / 2]
                                         ;
        }
        for (yy = ny / 2; yy < ny; ++yy) {
            inarr[yy * nx + xx] = inarrconv0[(yy - ny / 2) * nx + xx + nx / 2]
                                         ;
        }
    }
    fftw_execute(p);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            outconv[yy * nx + xx] = out[yy * nx + xx];
            inarr[yy * nx + xx] = inarr0[yy * nx + xx];
        }
    }
    fftw_execute(p);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            out[yy * nx + xx] *= outconv[yy * nx + xx];
            inarr[yy * nx + xx] = 0.0;
        }
    }
    fftw_destroy_plan(p);
    p = fftw_plan_dft_c2r_2d(ny,nx,outconv,inarr,FFTW_ESTIMATE);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            outconv[yy * nx + xx] = out[yy * nx + xx];
        }
    }
    fftw_execute(p);

    result = cpl_image_duplicate(image);
    inarr0 = cpl_image_get_data_double(result);
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            inarr0[yy * nx + xx] = inarr[yy * nx + xx] / ( nx * ny);
        }
    }
    fftw_destroy_plan(p);
    cpl_free(inarr); inarr = NULL;
    cpl_free(inarrconv); inarrconv = NULL;
    fftw_free(out); out = NULL;
    fftw_free(outconv); outconv = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

cpl_image*
sph_fftw_rotate_image( sph_fft* self, cpl_image* image, double degrees ) {
    cpl_image*            result    = NULL;
    cpl_image*              im = NULL;
    fftw_complex*           in = NULL;
    fftw_plan               p;
    int                     N = 0;
    int                     embed[1];
    int                     xx  = 0;
    int                     yy  = 0;
    double                  shift = 0.0;
    double                  delta = 0.1;
    double                  k = 0.0;

    cpl_ensure( self->method == SPH_FFT_FFTW3_DP, CPL_ERROR_ILLEGAL_INPUT, NULL);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    N = cpl_image_get_size_x(image);
    cpl_ensure(cpl_image_get_size_y(image) == N,CPL_ERROR_ILLEGAL_INPUT,NULL);
    im = cpl_image_duplicate(image);
    if ( degrees > 270.0 ) {
        cpl_image_turn( im, -3 );
        degrees = degrees - 270.0;
    }
    if ( degrees > 180.0 ) {
        cpl_image_turn( im, -2 );
        degrees = degrees - 180.0;
    }
    if ( degrees > 90.0 ) {
        cpl_image_turn( im, -1 );
        degrees = degrees - 90.0;
    }
    if ( degrees < -270.0 ) {
        cpl_image_turn( im, 3 );
        degrees = degrees + 270.0;
    }
    if ( degrees < -180.0 ) {
        cpl_image_turn( im, 2 );
        degrees = degrees + 180.0;
    }
    if ( degrees < -90.0 ) {
        cpl_image_turn( im, 1 );
        degrees = degrees + 90.0;
    }

    self->angle = -1.0 * degrees / 180.0 * CPL_MATH_PI;
    delta = -tan(self->angle/2.0);
    in = sph_fft_create_fftw_complex(self,im,N,N);
    cpl_ensure(in,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    embed[0] = N* N;
    // Forward rows
    p = fftw_plan_many_dft(1,&N,N,in,embed,1,
            N,in,embed,1,N,
            FFTW_FORWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);
    // Shift rows
    for (xx = 0; xx < N; ++xx) {
        for (yy = 0; yy < N; ++yy) {
            shift = delta * ( yy - ((double)N-1.0)/2.0 ) + self->shiftx;
            if ( xx < ((double)N)/2.0 ) {
                 k = (double)xx;
             }
             else {
                 k = (double)xx - (double)N;
             }

            shift = shift * (k/(double)N);

            in[yy * N + xx] = in[yy * N + xx] * cexp(2.0 * shift * CPL_MATH_PI * I);
        }
    }

    // Forward cols
    p = fftw_plan_many_dft(1,&N,N,in,embed,N,
            1,in,embed,N,1,
            FFTW_FORWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);

    delta = 1.0 * sin(self->angle);

    // Backward rows
    p = fftw_plan_many_dft(1,&N,N,in,embed,1,
            N,in,embed,1,N,
            FFTW_BACKWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);

    // Shift cols
    for (xx = 0; xx < N; ++xx) {
        for (yy = 0; yy < N; ++yy) {
            shift = delta * ( xx - ((double)N-1.0)/2.0 ) + self->shifty;
            if ( yy < ((double)N)/2.0 ) {
                 k = (double)yy;
             }
             else {
                 k = (double)yy - (double)N;
             }

            shift = shift * (k/(double)N);

            in[yy * N + xx] = in[yy * N + xx] * cexp(2.0 * shift * CPL_MATH_PI * I);
        }
    }

    // Forward rows
    p = fftw_plan_many_dft(1,&N,N,in,embed,1,
            N,in,embed,1,N,
            FFTW_FORWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);

    delta = -tan(self->angle/2.0);
    // Backward cols
    p = fftw_plan_many_dft(1,&N,N,in,embed,N,
            1,in,embed,N,1,
            FFTW_BACKWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);

    // Shift rows
    for (xx = 0; xx < N; ++xx) {
        for (yy = 0; yy < N; ++yy) {
            shift = delta * ( yy - ((double)N-1.0)/2.0  );// + self->shiftx;
            if ( xx < ((double)N)/2.0 ) {
                 k = (double)xx;
             }
             else {
                 k = (double)xx - (double)N;
             }

            shift = shift * (k/(double)N);

            in[yy * N + xx] = in[yy * N + xx] * cexp(2.0 * shift * CPL_MATH_PI * I);
        }
    }

    // Backward rows
    p = fftw_plan_many_dft(1,&N,N,in,embed,1,
            N,in,embed,1,N,
            FFTW_BACKWARD,FFTW_ESTIMATE);
    fftw_execute(p);
    fftw_destroy_plan(p);


    result = cpl_image_duplicate(im);
    for (xx = 0; xx < N; ++xx) {
        for (yy = 0; yy < N; ++yy) {
            cpl_image_set(result,xx+1,yy+1,creal(in[yy * N + xx])/(N * N * N));
        }
    }
    cpl_image_delete(im); im = NULL;
    fftw_free(in);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Perform the 3-step skew rotation algorithm
 * @param self
 *
 * @return error code of the operation
 *
 * This performs the actual rotation according to the 3-step skew
 * algorithm. The sph_fft object must have been prepared using
 * the sph_fft_rotate_prepare function. The resulting image
 * can be obtained with the sph_rotate_complex2image function.
 *
 * See sph_fft_rotate_prepare for more info.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_rotate( sph_fft* self ) {
    int        kk        = 0;
    double                delta    = 0.0;
    double                dx        = self->shiftx;
    double                dy        = self->shifty;
    double                fac = 1.0; //should become 2 when image doubling used

    delta = -tan(self->angle/2.0);
    //self->shiftx = self->shiftx + 0.707;
    //self->shifty = self->shifty - 0.5;
    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_forward_row( self, kk );
        sph_fft_shift_row( self, kk, delta );
    }
    for (kk = 0; kk < self->nx; ++kk) {
        sph_fft_forward_col( self, kk );
    }
    //sph_fft_rotate_double_sizey( self );
    delta = fac * sin(self->angle);
    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_inverse_row( self, kk );
    }
    for (kk = 0; kk < self->nx; ++kk) {
        sph_fft_shift_col( self, kk, delta );
    }
    self->shiftx = 0.0;
    self->shifty= 0.0;
    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_forward_row( self, kk );
    }
    delta = -tan(self->angle/2.0) / fac;
    for (kk = 0; kk < self->nx; ++kk) {
        sph_fft_inverse_col( self, kk );
    }
    for (kk = 0; kk < self->ny; ++kk) {
        sph_fft_shift_row( self, kk, delta );
        sph_fft_inverse_row( self, kk );
    }
    self->shiftx = dx;
    self->shifty = dy;
    return CPL_ERROR_NONE;
}

/* static  */
/* sph_error_code */
/* sph_fft_rotate_double_sizey( sph_fft* self ) { */
/*     gsl_complex_packed_array newarr; */

/*     newarr = cpl_calloc( 4 * self->Np * self->Npp, sizeof(double) ); */
/*     memcpy( newarr, self->complexdata, self->Npp * self->Np * sizeof(double)); */
/*     memcpy( &newarr[self->Np * self->Npp ], self->complexdata, self->Npp * self->Np * sizeof(double) ); */
/*     //memcpy( &newarr[self->Np * self->Npp / 2], self->complexdata, self->Npp * self->Np * sizeof(double) ); */
/*     cpl_free(    self->complexdata); */
/*     self->complexdata = newarr; */
/*     self->ny = self->Npp; */
/*     self->nx = self->Np; */
/*     return CPL_ERROR_NONE; */
/* } */
/*----------------------------------------------------------------------------*/
/**
 * @brief Create an image from the complex array
 * @param self
 *
 * @return the new image
 *
 * Ths converts the real part of the complex array
 * to a cpl_image without croppinh and
 * returns it.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_complex2image_real( sph_fft* self ) {
    int    nx = self->nx;
    int ny = self->ny;
    int    xx    = 0;
    int    yy    = 0;
    cpl_image*    result    = NULL;
    result = cpl_image_new( nx, ny, CPL_TYPE_DOUBLE );
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            cpl_image_set( result, xx + 1, yy + 1, self->complexdata[ 2* ( yy * nx + xx ) ] );
        }
    }
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Create an image from the complex array
 * @param self
 *
 * @return the new image
 *
 * Ths converts the imaginary part of the complex array
 * to a cpl_image without croppinh and
 * returns it.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_complex2image_imag( sph_fft* self ) {
    int    nx = self->nx;
    int ny = self->ny;
    int    xx    = 0;
    int    yy    = 0;
    cpl_image*    result    = NULL;

    result = cpl_image_new( nx, ny, CPL_TYPE_DOUBLE );
    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            cpl_image_set( result, xx + 1, yy + 1, self->complexdata[ 2* ( yy * nx + xx) + 1 ] );
        }
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Create an image from the complex array
 * @param self
 *
 * @return the new image
 *
 * Ths converts the real part of the complex array prepared
 * for in sph_fft_rotate_prepare to a cpl_image, crops it
 * to the original size and returns it.
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_fft_rotate_complex2image( sph_fft* self ) {
    int    nx = self->nx;
    int ny = self->ny;
    int    xx    = 0;
    int    yy    = 0;
    cpl_image*    result    = NULL;

    result = cpl_image_new( nx / 2, ny / 2, CPL_TYPE_DOUBLE );
    for (xx = 0; xx < nx / 2; ++xx) {
        for (yy = 0; yy < ny / 2; ++yy) {
            cpl_image_set( result, xx + 1, yy + 1, self->complexdata[ 2* ( ( yy + ny / 4 ) * nx + xx + nx / 4)] );
        }
    }
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Prepare FFT module for FFT operations
 * @param self
 * @param im        the image to FFT (unmodified by sph_fft)
 *
 * @return error code of the operation
 *
 * This function prepares the sph_fft module for a FFT
 * operation. Space for a complex array of at least double the size of the image
 * is allocated.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_prepare( sph_fft* self, cpl_image* im ) {
    int                    newx    = 2;
    int                    newy    = 2;
    sph_error_code        rerr    = CPL_ERROR_NONE;
    int                    N = 0;

    if ( !im ) return SPH_ERROR_GENERAL;
    if ( cpl_image_get_size_x(im) != cpl_image_get_size_y(im) ) {
        return SPH_ERROR_GENERAL;
    }
    if ( self->method == SPH_FFT_GSL_RADIX2 ) {
        N = cpl_image_get_size_x( im );
        while ( newx <= N ) {
            newx = newx * 2;
        }
        newy = newx;
        if ( newx > 2048 * 4 || newy > 2048 * 4 ) {
            return SPH_ERROR_GENERAL; // this is too big to handle
        }
        sph_fft_prepare_gsl_radix2(self,im,newx,newy);
        self->Npp = newy * 2;
        self->Np = newx;
    }
    else if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) {
        sph_fft_prepare_gsl_mixed_radix(self,im);
    }
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Prepare FFT module for rotation
 * @param self
 * @param im        the image to rotate (unmodified by sph_fft)
 * @param degree    the angle in degrees
 * @param dx        additional constant shift in x
 * @param dy        additional constant shift in y
 *
 * @return error code of the operation
 *
 * This function prepares the sph_fft module for a rotation/shift
 * operation. Space for a complex array of at least double the size of the image
 * is allocated.
 * The rotation algorithm is the 3-phase skew algorithm as described
 * in Owen et al. 1996 and implemented in other languages like IDL or MATLAB
 * ( see http://eeg.sourceforge.net/doc_m2html/bioelectromagnetism/RotateImage.html)
 * Currently the implementation here does leaves out a step that is important
 * in many other applications: the doubling of the image size and masking to avoid
 * aliasing. This may be implemented in future if this is found to be a problem.
 * Note that you may have to exclude a border of width equal to the pixel
 * shift when using the created images. Also, there is a "bug" which invalidates
 * the top pixel row even when dx and dy are zero.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_fft_rotate_prepare( sph_fft* self, cpl_image* im, double degrees, double dx, double dy ) {
    int                    N        = 0;
    double                angle    = degrees / 180.0 * SPH_PI;
    double                tantananglehalf    = tan( angle / 2.0 ) * tan( angle / 2.0 );
    int                    newx    = 2;
    int                    newy    = 2;
    sph_error_code        rerr    = CPL_ERROR_NONE;

    if ( !im ) return SPH_ERROR_GENERAL;
    if ( degrees > 90.0 || degrees < -90.0 ) {
        return SPH_ERROR_GENERAL;
    }
    if ( cpl_image_get_size_x(im) != cpl_image_get_size_y(im) ) {
        return SPH_ERROR_GENERAL;
    }
    N = cpl_image_get_size_x( im );
    self->angle    = angle;
    while ( newx <= N + N * tantananglehalf ) {
        newx = newx * 2;
    }
    while ( newy <= 2 * N ) {
        newy = newy * 2;
    }
    newy = newy / 2;
    if ( newx > 2048 * 4 || newy > 2048 * 4 ) {
        return SPH_ERROR_GENERAL; // this is too big to handle
    }

    self->shiftx = -dx;
    self->shifty = -dy;

    sph_fft_prepare_gsl_radix2(self,im,newx,newy);

    self->Npp = newy * 2;
    self->Np = newx;

    return rerr;
}


sph_error_code
sph_fft_filter_fftw_array( sph_fft* self, fftw_complex* in,sph_fft_filter_method method, double delta, double param1, double param2 ) {
    int         xx = 0;
    int         yy = 0;
    int         nx = 0;
    int         ny = 0;
    double      kx = 0.0;
    double      ky = 0.0;
    double      dshift = 0.0;
    double  r = 0.0;
    double factor = 0.0;
    double order = 0.0;

    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(in,CPL_ERROR_NULL_INPUT);
    nx = self->nx;
    ny = self->ny;

    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            if ( xx < ((double)nx)/2.0 ) {
                kx = (double)xx;
            }
            else {
                kx = (double)xx - (double)nx;
            }
            if ( yy < ((double)ny)/2.0 ) {
                ky = (double)yy;
            }
            else {
                ky = (double)yy - (double)ny;
            }

            dshift = (kx/(double)nx) * (kx/(double)nx) +
                    (ky/(double)ny) * (ky/(double)ny);
            if ( method == SPH_FFT_FILTER_METHOD_TH ) {
                if ( 4.0 * dshift > 1.0 - delta * delta ) {
                    in[yy * nx + xx] = 0.0;//exp( -dshift/(delta*delta));
                    in[yy * nx + xx] = 0.0;//exp( -dshift/(delta*delta));
                }
            }
            if ( method == SPH_FFT_FILTER_METHOD_FERMI ) {
                r = 1.0 - delta * delta;
                factor = 1.0 / ( 1 + exp( (4 * dshift - r) / param1 ) );
                in[yy * nx + xx] *=  factor;
                in[yy * nx + xx] *=  factor;
            }
            if ( method == SPH_FFT_FILTER_METHOD_BUTTER ) {
                order = 2 * log( 0.882 / sqrt(10.624 * 10.624 - 1.0) ) / log( param1 / param2 );
                r = param1 / pow(0.882, 2.0 / order);
                factor = 1.0 / sqrt( 1 + pow(4 * dshift / r, order ) );
                in[yy * nx + xx] *= factor;
                in[yy * nx + xx] *= factor;
            }

            if ( 4.0 * dshift > 1.0 - delta * delta ) {
                in[yy * nx + xx] = 0.0;
                in[yy * nx + xx] = 0.0;
            }
        }
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
sph_error_code
sph_fft_shift_fftw_array( sph_fft* self, fftw_complex* in,double dx, double dy ) {
    int         xx = 0;
    int         yy = 0;
    double      shiftx = 0.0;
    double      shifty = 0.0;
    int         nx = 0;
    int         ny = 0;
    double      kx = 0.0;
    double      ky = 0.0;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(in,CPL_ERROR_NULL_INPUT);
    nx = self->nx;
    ny = self->ny;

    for (xx = 0; xx < nx; ++xx) {
        for (yy = 0; yy < ny; ++yy) {
            shiftx = -1.0 * dx;
            shifty = -1.0 * dy;
            if ( xx < ((double)nx)/2.0 ) {
                kx = (double)xx;
            }
            else {
                kx = (double)xx - (double)nx;
            }
            if ( yy < ((double)ny)/2.0 ) {
                ky = (double)yy;
            }
            else {
                ky = (double)yy - (double)ny;
            }

            shiftx = shiftx * (kx/(double)nx);
            shifty = shifty * (ky/(double)ny);

            in[yy * nx + xx] = in[yy * nx + xx] * cexp(2.0 * shiftx * CPL_MATH_PI * I);
            in[yy * nx + xx] = in[yy * nx + xx] * cexp(2.0 * shifty * CPL_MATH_PI * I);
        }
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

fftw_complex*
sph_fft_crop_fftw_array( sph_fft* self, fftw_complex* in,int newx, int shiftflag )
{
    fftw_complex*            newdata = NULL;
    int xx;
    int yy;
    int xxold;
    int yyold;
    int newy = newx;

    cpl_ensure(in,CPL_ERROR_ILLEGAL_INPUT,NULL);
    cpl_ensure(newx <= self->nx, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    cpl_ensure(newy <= self->ny, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    newdata = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * newx * newy);
    cpl_ensure(newdata,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    if ( shiftflag == 0 ) {
        for (xx = 0; xx < newx; ++xx) {
            for (yy = 0; yy < newy; ++yy) {
                xxold = xx + ( self->nx - newx ) / 2;
                yyold = yy + ( self->ny - newy ) / 2;
                newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
            }
        }
    }
    else {
        for (xx = 0; xx < newx/2; ++xx) {
            for (yy = 0; yy < newy/2; ++yy) {
                xxold = xx;
                yyold = yy;
                newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
            }
            for (yy = newy/2; yy < newy; ++yy) {
                xxold = xx;
                yyold = yy + self->ny - newy;
                newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
            }
        }
        for (xx = newx/2; xx < newx; ++xx) {
            for (yy = 0; yy < newy/2; ++yy) {
                xxold = xx + self->nx - newx;
                yyold = yy;
                newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
            }
            for (yy = newy/2; yy < newy; ++yy) {
                xxold = xx + self->nx - newx;
                yyold = yy + self->ny - newy;
                newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
            }
        }
    }
    self->nx = newx;
    self->ny = newy;
    return newdata;
}
fftw_complex*
sph_fft_pad_fftw_array_freq_space( sph_fft* self, fftw_complex* in, int newn )
{
    fftw_complex*            newdata = NULL;
    int newx = newn;
    int newy = newn;
    int xx;
    int yy;
    int xxold;
    int yyold;
    int shiftx = newx-self->nx;
    int shifty = newy-self->ny;

    cpl_ensure(in,CPL_ERROR_ILLEGAL_INPUT,NULL);
    cpl_ensure(newx >= self->nx, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    cpl_ensure(newy >= self->ny, CPL_ERROR_INCOMPATIBLE_INPUT, NULL);
    newdata = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * newx * newy);
    cpl_ensure(newdata,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    for (xx = 0; xx < newx; ++xx) {
        for (yy = 0; yy < newy; ++yy) {
            newdata[yy * newx + xx ] = 0;
        }
    }
    for (xx = 0; xx < ( newx - shiftx )/ 2 ; ++xx) {
        xxold = xx;
        for (yy = 0; yy < ( newy - shifty ) / 2; ++yy) {
            yyold = yy;
            newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
        }
        for (yy = ( newy + shifty ) / 2; yy < newy; ++yy) {
            yyold = self->ny/2 + yy - (newy+shifty)/2;
            newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
        }
    }
    for (xx = ( newx + shiftx )/ 2; xx < newx ; ++xx) {
        xxold = self->nx/2 + xx - (newx+shiftx)/2;
        for (yy = 0; yy < ( newy - shifty ) / 2; ++yy) {
            yyold = yy;
            newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
        }
        for (yy = ( newy + shifty ) / 2; yy < newy; ++yy) {
            yyold = self->ny/2 + yy - (newy+shifty)/2;
            newdata[yy * newx + xx] = in[yyold * self->nx + xxold];
        }
    }
    self->nx = newx;
    self->ny = newy;
    return newdata;
}

sph_error_code
sph_fft_crop_complex_array( sph_fft* self, int newx, int shiftflag )
{
    double*            newdata = NULL;
    int xx;
    int yy;
    int xxold;
    int yyold;
    int newy = newx;

    newdata = cpl_calloc(2 * newx * newy, sizeof(double) );
    if ( shiftflag == 0 ) {
        for (xx = 0; xx < newx; ++xx) {
            for (yy = 0; yy < newy; ++yy) {
                xxold = xx + ( self->nx - newx ) / 2;
                yyold = yy + ( self->ny - newy ) / 2;
                newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
                newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
            }
        }
    }
    else {
        for (xx = 0; xx < newx/2; ++xx) {
            for (yy = 0; yy < newy/2; ++yy) {
                xxold = xx;
                yyold = yy;
                newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
                newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
            }
            for (yy = newy/2; yy < newy; ++yy) {
                xxold = xx;
                yyold = yy + self->ny - newy;
                newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
                newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
            }
        }
        for (xx = newx/2; xx < newx; ++xx) {
            for (yy = 0; yy < newy/2; ++yy) {
                xxold = xx + self->nx - newx;
                yyold = yy;
                newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
                newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
            }
            for (yy = newy/2; yy < newy; ++yy) {
                xxold = xx + self->nx - newx;
                yyold = yy + self->ny - newy;
                newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
                newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
            }
        }
    }
    cpl_free(self->complexdata);
    self->complexdata = newdata;
    self->nx = newx;
    self->ny = newy;
    if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) sph_fft_alloc_mixedradix(self);

    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Obtain scaling information using a row collapse of image
 * @param self      ftt module
 * @param im1       the image to obtain scaling info from
 *
 * @return period along the rows of the image.
 *
 * This routine measures the period along the x-axis (along the rows) of
 * the input image. It returns the period in pixels along with an estimate of
 * the error in the variable sigma.
 *
 *
 */
/*----------------------------------------------------------------------------*/
double
sph_fft_get_scale_vector(sph_fft* self, cpl_image* im1, double *sigma )
{
    cpl_image*		powerim = NULL;
    cpl_image*		imcut = NULL;
    int				xx = 0;
    int				yy = 0;
    double    y0 	= 0.0;
    double    x0    = 0.0;
    double kvalx = 0.0;
    double kvaly = 0.0;
    cpl_vector*	fftv = NULL;
    cpl_vector*	fftvx = NULL;
    int		nlx = 0;
    int		nly = 0;
    gsl_complex*	valp = NULL;
    double  xm = 0.0;
    double	area = 0.0;
    double	offset = 0.0;

    nlx = cpl_image_get_size_x(im1);
    nly = cpl_image_get_size_x(im1);

    sph_fft_prepare(self,im1);
    sph_fft_forward(self);

    x0 = ( self->nx - 1 ) / 2;
    y0 = ( self->ny - 1 ) / 2;
    powerim = cpl_image_new( self->nx, self->ny, CPL_TYPE_DOUBLE );
    for (xx = 0; xx < self->nx ; ++xx) {
        for (yy = 0; yy < self->ny; ++yy) {
            if ( yy <= y0 ) {
                kvaly = yy;
            }
            else {
                kvaly = yy - self->ny;
            }
            if ( xx <= x0 ) {
                kvalx = xx;
            }
            else {
                kvalx = xx - self->nx;
            }
            valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * self->nx + xx ) ];
            cpl_image_set(powerim,xx+1,yy+1,gsl_complex_abs2(*valp));
        }
    }
    imcut = cpl_image_extract( powerim, 2,  1, nlx/2, 1);
    fftv = cpl_vector_new_from_image_row(imcut,1);
    fftvx = cpl_vector_duplicate(fftv);
    for (xx = 0; xx < cpl_vector_get_size(fftvx); ++xx) {
		cpl_vector_set(fftvx,xx,xx+1.0);
	}
    cpl_vector_fit_gaussian(fftvx,NULL,fftv,NULL,CPL_FIT_ALL,&xm,sigma,&area,&offset,NULL,NULL,NULL);
    *sigma = *sigma / xm;
    xm = cpl_image_get_size_x(im1)/xm;
    *sigma = *sigma * xm;
    cpl_image_delete(imcut);
    cpl_image_delete(powerim);
    cpl_vector_delete(fftv);
    cpl_vector_delete(fftvx);
    return xm;
}
sph_error_code
sph_fft_set_shift(sph_fft* self, double dx, double dy) {
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    self->shiftx = dx;
    self->shifty = dy;
    return CPL_ERROR_NONE;
}

fftw_complex*
sph_fft_create_fftw_complex( sph_fft* self, cpl_image* im, int newx, int newy )
{
    int xx;
    int yy;
    int xxold;
    int yyold;
    int nx = 0;
    int ny = 0;
    int shiftx = 0;
    int shifty = 0;
    fftw_complex*           in = NULL;
    int                     bpix = 0;

    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(im,CPL_ERROR_NULL_INPUT,NULL);
    nx = cpl_image_get_size_x(im);
    ny = cpl_image_get_size_y(im);
    shiftx = newx - nx;
    shifty = newy - ny;
    cpl_ensure(shiftx >= 0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    cpl_ensure(shifty >= 0,CPL_ERROR_ILLEGAL_INPUT,NULL);
    in = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * newx * newy);
    cpl_ensure(in,CPL_ERROR_ILLEGAL_OUTPUT,NULL);
    if ( shiftx || shifty ) {
        for (xx = 0; xx < newx; ++xx) {
            for (yy = 0; yy < newy; ++yy) {
                in[yy * newx + xx ] = 0;
            }
        }
    }
    for (xx = shiftx/2; xx < newx - shiftx/2; ++xx)
    {
        for (yy = shifty/2; yy < newy - shifty/2; ++yy)
        {
            xxold = xx - shiftx/2;
            yyold = yy - shifty/2;
            in[yy * newx + xx] = cpl_image_get(im,xxold+1,yyold+1,&bpix);
        }
    }
    self->nx = newx;
    self->ny = newy;
    return in;
}
sph_error_code
sph_fft_pad_complex_array_real_space( sph_fft* self, int newn )
{
    double*            newdata = NULL;
    int newx = newn;
    int newy = newn;
    int xx;
    int yy;
    int xxold;
    int yyold;
    int shiftx = newx-self->nx;
    int shifty = newx-self->ny;

    if ( self->nx >= newx || self->ny >= newy ) return CPL_ERROR_NONE;

    newdata = cpl_calloc(2 * newx * newy, sizeof(double) );
    for (xx = shiftx/2; xx < newx - shiftx/2; ++xx)
    {
        for (yy = shifty/2; yy < newy - shifty/2; ++yy)
        {
            xxold = xx - shiftx/2;
            yyold = yy - shifty/2;
            newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
            newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
        }
    }
    cpl_free(self->complexdata);
    self->complexdata = newdata;
    self->nx = newx;
    self->ny = newy;
    if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) sph_fft_alloc_mixedradix(self);
    return CPL_ERROR_NONE;
}

sph_error_code
sph_fft_shift( sph_fft* self, double dx, double dy )
{
    int xx;
    int yy;
    double    y0 =0.0;
    double    x0    = 0.0;
    double kvalx = 0.0;
    double kvaly = 0.0;
    double dshiftx = 0.0;
    double dshifty = 0.0;
    gsl_complex            mulfac;
    gsl_complex            mulresult;
    gsl_complex*        valp    = NULL;

    x0 = ( self->nx - 1 ) / 2;
    y0 = ( self->ny - 1 ) / 2;
    for (xx = 0; xx < self->nx ; ++xx) {
        for (yy = 0; yy < self->ny; ++yy) {
            if ( yy <= y0 ) {
                kvaly = yy;
            }
            else {
                kvaly = yy - self->ny;
            }
            dshiftx = -1.0 * (double)dy * ( kvaly / (double)self->ny );
            mulfac = gsl_complex_polar( 1.0, 2.0 * SPH_PI * dshiftx  );
            valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * self->nx + xx ) ];
            mulresult = gsl_complex_mul( *valp, mulfac );
            valp->dat[0] = mulresult.dat[0];
            valp->dat[1] = mulresult.dat[1];
            if ( xx <= x0 ) {
                kvalx = xx;
            }
            else {
                kvalx = xx - self->nx;
            }
            dshifty = -1.0 * (double)dx * ( kvalx / (double)self->nx );
            mulfac = gsl_complex_polar( 1.0, 2.0 * SPH_PI * dshifty  );
            valp = (gsl_complex*)&self->complexdata[ 2 * ( yy * self->nx + xx ) ];
            mulresult = gsl_complex_mul( *valp, mulfac );
            valp->dat[0] = mulresult.dat[0];
            valp->dat[1] = mulresult.dat[1];
        }
    }
    return CPL_ERROR_NONE;
}

sph_error_code
sph_fft_pad_complex_array_freq_space( sph_fft* self, int newn )
{
    double*            newdata = NULL;
    int newx = newn;
    int newy = newn;
    int xx;
    int yy;
    int xxold;
    int yyold;
    int shiftx = newx-self->nx;
    int shifty = newy-self->ny;

    if ( self->nx >= newx || self->ny >= newy ) return CPL_ERROR_NONE;

    newdata = cpl_calloc(2 * newx * newy, sizeof(double) );
    for (xx = 0; xx < ( newx - shiftx )/ 2 ; ++xx) {
        xxold = xx;
        for (yy = 0; yy < ( newy - shifty ) / 2; ++yy) {
            yyold = yy;
            newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
            newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
        }
        for (yy = ( newy + shifty ) / 2; yy < newy; ++yy) {
            yyold = self->ny/2 + yy - (newy+shifty)/2;
            newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
            newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
        }
    }
    for (xx = ( newx + shiftx )/ 2; xx < newx ; ++xx) {
        xxold = self->nx/2 + xx - (newx+shiftx)/2;
        for (yy = 0; yy < ( newy - shifty ) / 2; ++yy) {
            yyold = yy;
            newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
            newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
        }
        for (yy = ( newy + shifty ) / 2; yy < newy; ++yy) {
            yyold = self->ny/2 + yy - (newy+shifty)/2;
            newdata[2 * ( yy * newx + xx)] = self->complexdata[2*(yyold * self->nx + xxold)];
            newdata[2 * ( yy * newx + xx) + 1] = self->complexdata[2*(yyold * self->nx + xxold) + 1];
        }
    }
    cpl_free(self->complexdata);
    self->complexdata = newdata;
    self->nx = newx;
    self->ny = newy;
    if ( self->method == SPH_FFT_GSL_MIXEDRADIX ) sph_fft_alloc_mixedradix(self);
    return CPL_ERROR_NONE;
}

void sph_fft_delete( sph_fft* self ) {
    if ( self ) {

        if ( self->complexdata) cpl_free( self->complexdata );
        if ( self->wavetab ) gsl_fft_complex_wavetable_free(self->wavetab);
        if ( self->workspace ) gsl_fft_complex_workspace_free(self->workspace);
        cpl_free(self);
    }
}

