/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */
#define _SPH_POLYGON_C_
#include "sph_error.h"
#include "cpl.h"
#include "sph_polygon.h"
#include "sph_point.h"
#include <string.h>
#include <math.h>
#include <limits.h>

/*----------------------------------------------------------------------------*/
/**
 @brief Create a new empty polygon

 @param nothing

 @return the pointer to new polygon.

 Description: create a new empty polygon.

 */
/*----------------------------------------------------------------------------*/
sph_polygon* sph_polygon_new( void ) {
    sph_polygon*        result      = NULL;
    result = cpl_calloc( sizeof( sph_polygon ), 1 );
    result->npoints = 0;
    result->points = NULL;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Create a new duplicate polygon

 @param the polygon to copy

 @return the pointer to new polygon.

 Description: create a new duplicate polygon.

 */
/*----------------------------------------------------------------------------*/
sph_polygon* sph_polygon_duplicate( sph_polygon* poly ) {
    sph_polygon*        result      = NULL;
    int                    ii            = 0;
    if ( !poly ) {
        SPH_ERROR_RAISE_ERR(CPL_ERROR_NULL_INPUT,"Null input poly");
        return NULL;
    }
    result = cpl_calloc( sizeof( sph_polygon ), 1 );
    for (ii = 0; ii < poly->npoints; ++ii) {
        sph_polygon_add_point(result,poly->points[ii]->x, poly->points[ii]->y);
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Frees the singleton polygon

 @param nothing

 This function frees the memory of the static polygon.

 */
/*----------------------------------------------------------------------------*/
void sph_polygon_free_static( void ) {
}


/*----------------------------------------------------------------------------*/
/**
 @brief Add a new point to an existing polygon

 @param self        the polygon to add the point to
 @param x            the x-coordinate of point to add
 @param y            the y-coordinate of point to add

 @return error code of the operation.

 Description: Adds a new point to the polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_add_point( sph_polygon* self, double x, double y ) {
    sph_error_code      rerr        = CPL_ERROR_NONE;
    sph_point**         temp        = NULL;

    if ( self->points ) {
        temp = cpl_calloc( sizeof(sph_point*), self->npoints );
        memcpy(temp, self->points, self->npoints * sizeof(sph_point*) );
        cpl_free(self->points);
    }
    self->points = cpl_calloc( sizeof(sph_point*), self->npoints + 1 );
    if ( !self->points ) {
        return SPH_ERROR_MEMORY_FAIL;
    }
    if ( temp ) {
        memcpy(self->points, temp, self->npoints * sizeof(sph_point*) );
        cpl_free(temp);
    }
    self->points[self->npoints] = sph_point_new( x, y );
    self->npoints++;
    return rerr;
}

double sph_polygon_get_top(sph_polygon* self) {
    double result            = -DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->y > result) result = self->points[ii]->y;
    }
    return result;
}
double sph_polygon_get_bottom(sph_polygon* self) {
    double result            = DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->y < result) result = self->points[ii]->y;
    }
    return result;
}
double sph_polygon_get_left(sph_polygon* self) {
    double result            = DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->x < result) result = self->points[ii]->x;
    }
    return result;
}
double sph_polygon_get_right(sph_polygon* self) {
    double result            = -DBL_MAX;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        if ( self->points[ii]->x > result) result = self->points[ii]->x;
    }
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return the centre x coordinate

 @param self        the polygon

 @return the x coordinate of the geometric centre

 Description: returns the x coordinate of the geometric centre.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_get_midx( sph_polygon* self ) {
    double result            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        result = result + self->points[ii]->x;
    }
    if ( self->npoints > 0 )
        result = result / self->npoints;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief enlarge the polygon by a factor

 @param self        the polygon
 @param    factor        the factor to enlarge by (lower 1 means shrink)

 @return error code

 Description: enlarge the polygon by a factor. If the factor is less than 1, it
 corresponds to a shrink. The factor must be larger 0.
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_enlarge( sph_polygon* self, double fac ) {
    double x            = 0.0;
    double y            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    if ( fac <= DBL_MIN ) {
        SPH_NULL_ERROR
        return CPL_ERROR_NULL_INPUT;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        x += self->points[ii]->x;
        y += self->points[ii]->y;
    }
    if ( self->npoints > 0 ) {
        x /= self->npoints;
        y /= self->npoints;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        self->points[ii]->x = x + fac * ( self->points[ii]->x - x );
        self->points[ii]->y = y + fac * ( self->points[ii]->y - y );
    }
    return CPL_ERROR_NONE;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Return the centre y coordinate

 @param self        the polygon

 @return the y coordinate of the geometric centre

 Description: returns the y coordinate of the geometric centre.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_get_midy( sph_polygon* self ) {
    double result            = 0.0;
    int        ii                = 0;
    if ( !self ) {
        SPH_NULL_ERROR;
        return 0.0;
    }
    for ( ii = 0;  ii < self->npoints; ++ ii) {
        result = result + self->points[ii]->y;
    }
    if ( self->npoints > 0 )
        result = result / self->npoints;
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return the edge of a polygon as sph_line

 @param self        the polygon the edge belongs to
 @param edge        the id of the edge (see below)

 @return the edge as a new sph_line

 Description: returns the edge of a polygon as a new sph_line
 (the pointer to the sph_line is returned). The edge is identified
 by an integer id, between 1 and the number of edges of the polygon.
 The first edge is the one between points 0 and point 1. The last
 edge is the one between the past point and point 0.

 */
/*----------------------------------------------------------------------------*/
sph_line*   sph_polygon_get_edge( sph_polygon* self, int edge ) {
    sph_line*       result      = sph_line_new(0.0,0.0,1.0,1.0);
    if ( __sph_polygon_get_edge_inplace( self, edge, result ) ) {
        sph_line_delete(result);result = NULL;
    }
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Clip a polygon to a target polygon.

 @param self        the polygon to clip
 @param target        the polygon to clip to

 @return error code of the operation

 Description: This clips the polygon self, so that it afterward
 corresponds to an area that is the unio between the area of self and
 the target polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_clip_to_target(
        sph_polygon* self,
        sph_polygon* target,
        sph_polygon_workspace* workspace)
{
    sph_polygon*        newpol      = NULL;
    int                 ii          = 0;
    int                 nextp       = 0;
    int                    current_npoints = 0;
    sph_point*          intersect   = NULL;
    sph_line*           targetedge  = NULL;
    sph_line*           sourceedge  = NULL;
    int                 ee          = 0;
    int                 kk          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;
    sph_line*           workline   = NULL;

    if ( !self || !target ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }

    workline = workspace->line1;

    intersect = workspace->point1;
    newpol = workspace->poly1;
    sourceedge = workspace->line2;
    targetedge = workspace->line3;

    for ( ee = 1; ee <= target->npoints; ++ee ) {
        current_npoints = 0;
        __sph_polygon_get_edge_inplace( target, ee, targetedge );
        for ( ii = 0; ii < self->npoints; ++ii ) {
            nextp = ii + 1;
            if ( nextp == self->npoints ) nextp = 0;
            if ( sph_polygon_test_same_side(
                    target, ee, self->points[ii], workline
                    )
                )
            {
                if ( sph_polygon_test_same_side(
                        target, ee, self->points[nextp], workline
                        )
                    )
                {
                    __sph_poly_add_point_static(
                            newpol,
                            self->points[nextp],
                            current_npoints );
                    current_npoints++;
                }
                else {
                    sourceedge->p1->x = self->points[ii]->x;
                    sourceedge->p1->y = self->points[ii]->y;
                    sourceedge->p2->x = self->points[nextp]->x;
                    sourceedge->p2->y = self->points[nextp]->y;
                    intersect->x = __sph_line_intersect_point_x(
                            targetedge,
                            sourceedge );
                    intersect->y = __sph_line_intersect_point_y(
                            targetedge,
                            sourceedge );
                    __sph_poly_add_point_static(
                            newpol,
                            intersect,
                            current_npoints );
                    current_npoints++;
                }
            }
            else {
                if ( sph_polygon_test_same_side(
                        target, ee, self->points[nextp], workline
                    ) )
                {
                    sourceedge->p1->x = self->points[ii]->x;
                    sourceedge->p1->y = self->points[ii]->y;
                    sourceedge->p2->x = self->points[nextp]->x;
                    sourceedge->p2->y = self->points[nextp]->y;
                    intersect->x = __sph_line_intersect_point_x(
                            targetedge, sourceedge );
                    intersect->y = __sph_line_intersect_point_y(
                            targetedge, sourceedge );
                    __sph_poly_add_point_static( newpol,
                            intersect, current_npoints );
                    current_npoints++;
                    __sph_poly_add_point_static( newpol,
                            self->points[nextp], current_npoints );
                    current_npoints++;
                }
            }
        }
        if ( self->npoints <= current_npoints ) {
            for ( kk = 0; kk < self->npoints; ++kk ) {
                self->points[kk]->x = newpol->points[kk]->x;
                self->points[kk]->y = newpol->points[kk]->y;
            }
            for (kk = self->npoints; kk < current_npoints; ++kk) {
                sph_polygon_add_point(self,
                        newpol->points[kk]->x,
                        newpol->points[kk]->y );
            }
        }
        else {
            for (kk = current_npoints; kk < self->npoints; ++kk) {
                sph_point_delete(self->points[kk]);
                self->points[kk] = NULL;
            }
            self->npoints = current_npoints;
            for ( kk = 0; kk < current_npoints; ++kk ) {
                self->points[kk]->x = newpol->points[kk]->x;
                self->points[kk]->y = newpol->points[kk]->y;
            }
        }
    }
    return rerr;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Calculate the overlap are of the polygon with another.

 @param self        the first polygon
 @param target        the second polygon to overlap.

 @return area of overlap.

 Description: The calculates the area of overlap.

 */
/*----------------------------------------------------------------------------*/
double
sph_polygon_calculate_overlap(
        sph_polygon* self,
        sph_polygon* target,
        sph_polygon_workspace* ws )
{
    sph_polygon*        copy    = NULL;
    double              result  = 0.0;
    sph_error_code      rerr    = CPL_ERROR_NONE;

    copy = sph_polygon_duplicate(self);
    if ( !copy ) {
        return -1.0;
    }
    rerr |= sph_polygon_clip_to_target( copy, target, ws );
    if ( rerr != CPL_ERROR_NONE ) {
        return -1.0;
    }
    result = sph_polygon_area( copy );
    sph_polygon_delete( copy );
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief NOT IMPLEMENTED YET!

 @param

 @return

 Description:

 */
/*----------------------------------------------------------------------------*/
sph_error_code sph_polygon_sort( sph_polygon* self ) {
    if ( !self ) {
        SPH_NULL_ERROR;
        return CPL_ERROR_NULL_INPUT;
    }
    return CPL_ERROR_UNSPECIFIED;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Test if a point is on the same side of edge as all polygon points

 @param self        the polygon the edge belongs to
 @param ee            the id of the edge (see sph_polygon_get_edge)
 @param x            the x-coord. of point to test.
 @param y            the y-coord. of point to test.

 @return 1 if point is on same side, 0 if not and -1 on error.

 Description: This function tests of a point is on the same side of an edge
 as all polygon points of the polygon self.

 */
/*----------------------------------------------------------------------------*/
short sph_polygon_test_same_side(
        sph_polygon* self,
        int ee,
        sph_point* p,
        sph_line* work_line
        )
{
    int             pt      = ee + 1;
    sph_line*       edge    = NULL;
    short           result  = 0;

    if ( pt >= self->npoints ) {
        pt = pt - self->npoints;
    }

    edge = (sph_line*)work_line;
    __sph_polygon_get_edge_inplace( self, ee, edge );
    result = sph_point_ccw_test( *edge->p1, *self->points[pt], *edge->p2 ) == sph_point_ccw_test( *edge->p1, *p, *edge->p2 );
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Test if a point is inside a polygon

 @param self         the polygon
 @param    x            the x-coordinate of point to test
 @param y            the y-coordinate of point to test

 @return 0 if not inside, 1 if inside, -1 on error.

 Description: This function tests if a point is inside the
 given polygon. If an error occures (e.g. the polygon has 2 or less points)
 -1 is returned.

 */
/*----------------------------------------------------------------------------*/
short sph_polygon_test_inside( sph_polygon* self, double x, double y ) {
    double              minx        = 0.0;
    double              miny        = 0.0;
    sph_line*           edge        = NULL;
    sph_line*           toout       = NULL;
    int                 ii          = 0;
    int                 intersectc  = 0;

    if ( !self ) {
        return -1;
    }
    if ( self->npoints < 3 ) {
        return -1;
    }
    minx         = self->points[ 0 ]->x;
    miny         = self->points[ 0 ]->y;
    for ( ii = 1; ii < self->npoints; ++ii ) {
        if ( self->points[ii]->x < minx ) {
            minx = self->points[ii]->x;
        }
        if ( self->points[ii]->y < miny ) {
            miny = self->points[ii]->y;
        }
    }

    toout = sph_line_new( x , y , minx - 0.1, miny - 0.1 );
    for ( ii = 1; ii <= self->npoints; ++ii ) {
        edge = sph_polygon_get_edge( self, ii );
        if ( sph_line_intersect_test( edge, toout) ) {
            intersectc++;
        }
        sph_line_delete( edge );
    }
    sph_line_delete( toout );
    if ( intersectc % 2 ) {
        return 1;
    }
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Shift a polygon.

 @param self        the polygon to shift
 @param dx            the x amount to shift by
 @param dy            the y amount to shift by

 @return error code of the operation

 Description: Shift a polygon.

 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_shift( sph_polygon* self,
                   double dx, double dy )
{
    int                 ii          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;

    if ( !self ) {
        return CPL_ERROR_NULL_INPUT;
    }

    for ( ii = 0; ii < self->npoints; ++ii ) {
        rerr |= sph_point_shift( self->points[ii],
                                 dx, dy );
    }
    return rerr;
}
/*----------------------------------------------------------------------------*/
/**
 @brief Rotate a polygon around a point and shift afterwards.

 @param self        the polygon to rotate and shift
 @param cosangle    the cosine of the angle to rotate by
 @param    sinangle    the sine of the angle to rotate by
 @param dx            the x amount to shift by
 @param dy            the y amount to shift by

 @return error code of the operation

 Description: Rotate and afterwards shift a polygon. The rotation is
 specified by the sine and cosine of the rotation angle (measured counter
 clockwise). This input format is for performance reasons.
 The shift is performed after the rotation.

 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_polygon_rotate_around( sph_polygon* self,
                                          double x, double y,
                                          double cosangle,
                                          double sinangle,
                                          double dx, double dy )
{
    int                 ii          = 0;
    sph_error_code      rerr        = CPL_ERROR_NONE;

    if ( !self ) {
        return CPL_ERROR_NULL_INPUT;
    }

    for ( ii = 0; ii < self->npoints; ++ii ) {
        rerr |= sph_point_rotate_around( self->points[ii],
                                 x, y,
                                 cosangle,
                                 sinangle,
                                 dx, dy );
    }
    return rerr;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Return a point(vertex) of the polygon

 @param  self     the polygon
 @param  point  the point (vertext number) to return

 @return pointer to the vertex or NULL if vertex does not exist.

 Description: This returns a pointer to a vertex.
 */
/*----------------------------------------------------------------------------*/
sph_point*
sph_polygon_get_point( sph_polygon* self, int p ) {
    if ( !self ) return NULL;
    if ( p<0 || p>=self->npoints ) return NULL;
    return self->points[p];
}

/*----------------------------------------------------------------------------*/
/**
 @brief Calculate the are of the polygon

 @param  self     the polygon

 @return area of the polygon

 Description: This calculates the are of all clokwise or
 counter clockwise sorted polygons.

 */
/*----------------------------------------------------------------------------*/
double sph_polygon_area( sph_polygon* self) {

      double  area = 0.0;
      int     ii, jj =0;

      for ( ii = 0; ii < self->npoints; ii++) {
        jj++; if ( jj == self->npoints) jj=0;
        area += (self->points[ii]->x + self->points[jj]->x) *
        ( self->points[ii]->y - self->points[jj]->y );
      }

      return fabs(area)*.5;
}

/*----------------------------------------------------------------------------*/
/**
 @brief Delete the polygon and free space

 @param self        the polygon to delete

 @return nothing

 Description: Does just that.

 */
/*----------------------------------------------------------------------------*/
void sph_polygon_delete( sph_polygon* self ) {
    int         ii  = 0;

    if ( self ) {
        if ( self->points ) {
            for ( ii = 0; ii < self->npoints; ++ii ) {
                sph_point_delete(self->points[ii]);
            }
            cpl_free(self->points);
        }
        cpl_free(self);
    }
}

/*----------------------------------------------------------------------------*/
/**
 @brief Free all space allocated for this module
 */
/*----------------------------------------------------------------------------*/
void sph_polygon_free_all( void ) {
}
/*----------------------------------------------------------------------------*/
/**
 @brief Free the workspace
 */
/*----------------------------------------------------------------------------*/
void sph_polygon_workspace_delete( sph_polygon_workspace* self ) {
    if ( self ) {
        sph_point_delete(self->point1);
        sph_point_delete(self->point2);
        sph_point_delete(self->point3);
        sph_line_delete(self->line1);
        sph_line_delete(self->line2);
        sph_line_delete(self->line3);
        sph_polygon_delete(self->poly1);
        cpl_free(self);
    }
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Allocate a new polygon workspace for use in the clip function
 * @return  pointer to new sph_polygon_workspace or NULL
 */
/*----------------------------------------------------------------------------*/
sph_polygon_workspace*
sph_polygon_workspace_new( void ) {
    sph_polygon_workspace*      self = NULL;

    self = cpl_calloc( 1, sizeof( sph_polygon_workspace ) );
    self->line1 = sph_line_new(0.0,0.0,0.0,0.0);
    self->line2 = sph_line_new(0.0,0.0,0.0,0.0);
    self->line3 = sph_line_new(0.0,0.0,0.0,0.0);
    self->point1 = sph_point_new(0.0,0.0);
    self->point2 = sph_point_new(0.0,0.0);
    self->point3 = sph_point_new(0.0,0.0);
    self->poly1 = sph_polygon_new();
    return self;
}

/*----------------------------------------------------------------------------*/
//          STATIC FUNCTIONS
/*----------------------------------------------------------------------------*/
static
void
__sph_poly_add_point_static(
        sph_polygon* newpol,
        sph_point* p,
        int pointc )
{
    if ( pointc < newpol->npoints ) {
        newpol->points[pointc]->x = p->x;
        newpol->points[pointc]->y = p->y;
    }
    else {
        sph_polygon_add_point( newpol, p->x, p->y );
    }
}

static
double __sph_line_intersect_point_y( sph_line* lineA, sph_line* lineB ) {
    double          x           = 0;
    double          y           = 0;
    double          x1          = 0;
    double          x2          = 0;
    double          x3          = 0;
    double          x4          = 0;
    double          y1          = 0;
    double          y2          = 0;
    double          y3          = 0;
    double          y4          = 0;
    double          denom       = 0;

    x1 = lineA->p1->x;
    x2 = lineA->p2->x;
    y1 = lineA->p1->y;
    y2 = lineA->p2->y;
    x3 = lineB->p1->x;
    x4 = lineB->p2->x;
    y3 = lineB->p1->y;
    y4 = lineB->p2->y;

    denom = ( ( x1 - x2 ) * ( y3 - y4 ) - ( y1 - y2 ) * ( x3 - x4 ) );

    if ( fabs( denom ) < DBL_MIN ) {
        return 0;
    }
    x = ( x1 * y2 - y1 * x2 ) * (x3 - x4) -
           ( x1 - x2 ) * ( x3 * y4 - y3 * x4 );

    y = ( x1 * y2 - y1 * x2 ) * (y3 - y4) -
           ( y1 - y2 ) * ( x3 * y4 - y3 * x4 );

    return y / denom;
}

static
double __sph_line_intersect_point_x(
        sph_line* lineA,
        sph_line* lineB )
{
    double          x           = 0;
    double          y           = 0;
    double          x1          = 0;
    double          x2          = 0;
    double          x3          = 0;
    double          x4          = 0;
    double          y1          = 0;
    double          y2          = 0;
    double          y3          = 0;
    double          y4          = 0;
    double          denom       = 0;

    x1 = lineA->p1->x;
    x2 = lineA->p2->x;
    y1 = lineA->p1->y;
    y2 = lineA->p2->y;
    x3 = lineB->p1->x;
    x4 = lineB->p2->x;
    y3 = lineB->p1->y;
    y4 = lineB->p2->y;

    denom = ( ( x1 - x2 ) * ( y3 - y4 ) - ( y1 - y2 ) * ( x3 - x4 ) );

    if ( fabs( denom ) < DBL_MIN ) {
        return 0;
    }
    x = ( x1 * y2 - y1 * x2 ) * (x3 - x4) -
           ( x1 - x2 ) * ( x3 * y4 - y3 * x4 );

    y = ( x1 * y2 - y1 * x2 ) * (y3 - y4) -
           ( y1 - y2 ) * ( x3 * y4 - y3 * x4 );

    return x / denom;
}

static
int __sph_polygon_get_edge_inplace(
        sph_polygon* self,
        int edge,
        sph_line* inplace )
{
    int             p1          = 0;
    int             p2          = 0;

    if ( edge < 1 || !self ) {
        return -1;
    }

    if ( edge < self->npoints ) {
        p1 = edge - 1;
        p2 = edge;
    }
    else if ( edge == self->npoints ) {
        p1 = edge -1;
        p2 = 0;
    }
    else {
        return -1;
    }
    inplace->p1->x = self->points[ p1 ]->x;
    inplace->p1->y = self->points[ p1 ]->y;
    inplace->p2->x = self->points[ p2 ]->x;
    inplace->p2->y = self->points[ p2 ]->y;
    return 0;
}

