/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <cpl.h>
#include "sph_error.h"
#include "sph_transform.h"
#include <math.h>

#define SPH_TRANSFORM_VAL_BADISBAD 0.000001
#define SPH_TRANSFORM_MAX_BADPIX 10000000
#define SPH_TRANSFORM_OK_BADPIX 100000


/*----------------------------------------------------------------------------*/
/**
 * @defgroup A SPHERE API Module
 * @par Synopsis:
 * @code
 * typedef _module_ {
 * } module
 * @endcode
 * @par Desciption:
 *
 * This module provides functionality for apertures, extending the functionality
 * as it exists for cpl_apertures.
 */
/*----------------------------------------------------------------------------*/
/**@{*/


static
sph_transform* __sph_transform_new(void);
static
sph_error_code __sph_transform_apply_filter( sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap);
static
sph_error_code
__sph_transform_distort( sph_polynomial_distortion_model* distortionmap,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap);
static
sph_error_code
__sph_transform_rotate(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double angle);
static
sph_error_code
__sph_transform_scale(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double scale);
static
sph_error_code
__sph_transform_shift(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double dx,
        double dy);
static
sph_error_code
__sph_transform_apply_to_images_fft(
        sph_transform* self,
        cpl_image** image,
        cpl_image** badpix,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel);
static
sph_error_code
__sph_transform_apply_to_images_cpl_warp(
        sph_transform* self,
        cpl_image** image,
        cpl_image** badpix,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel);
static
sph_error_code
__sph_transform_apply_displacements_cpl(
        sph_transform* self,
        cpl_image** image,
        cpl_image** rmsmap,
        cpl_image** ncombmap,
        cpl_image* displacement_x,
        cpl_image* displacement_y) ;
/*----------------------------------------------------------------------------*/
/**
 * @brief Create new default transformation
 *
 * @return new default transformation structure
 *
 *----------------------------------------------------------------------------*/
sph_transform* sph_transform_new_default(void) {
    sph_transform*      result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    result = __sph_transform_new();
    result->method = SPH_TRANSFORM_METHOD_FFT;
    result->filt_method = SPH_FFT_FILTER_METHOD_NONE;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Create new cpl_warp transformation
 *
 * @return new transformation structure
 *
 *----------------------------------------------------------------------------*/
sph_transform* sph_transform_new_cpl_warp(void) {
    sph_transform*      result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    result = __sph_transform_new();
    result->method = SPH_TRANSFORM_METHOD_CPL_WARP;
    result->kerneltype = CPL_KERNEL_DEFAULT;
    result->radius = CPL_KERNEL_DEF_WIDTH;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Create new FFT transformation
 * @param filtmehthod       the FFT filter method to use
 * @param param1            first parameter of filter
 * @param param2            second parameter of filter
 *
 * @return new transformation structure
 *
 * The parameters are interpreted depending on the FFT filter method. For
 * the top hat filter, the first parameter is the freq. radius,
 * and the second is ignored.
 * For the Fermi filter, the first is the freq. radius and the second the
 * temperature.
 * For the Butterworth filter, the first is the pass frequency and the second
 * the stop frequency.
 * All frequencies should be expressed as fractions of the total frequency
 * domain, and so should all be ranging for 0 (no frequencies filtered) to
 * 1 (all frequencies fitlered out).
 *
 *----------------------------------------------------------------------------*/
sph_transform* sph_transform_new_fft(sph_fft_filter_method filtmethod,
        double param1,
        double param2) {
    sph_transform*      result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    result = __sph_transform_new();
    result->method = SPH_TRANSFORM_METHOD_FFT;
    result->filt_method = filtmethod;
    if ( filtmethod == SPH_FFT_FILTER_METHOD_TH ) {
        result->fft_filter_radius = param1;
    }
    if ( filtmethod == SPH_FFT_FILTER_METHOD_FERMI ) {
        result->fft_filter_radius = param1;
        result->fft_fermi_temp = param2;
    }
    if ( filtmethod == SPH_FFT_FILTER_METHOD_BUTTER ) {
        result->fft_butter_pass= param1;
        result->fft_butter_stop = param2;
    }
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Center (shift), dedistort and rotate master frame
 * @param mframe        input master frame (change on output)
 * @param distortionmap the distortionmap to apply (maybe NULL)
 * @param centx         the centre in x (pixel coords)
 * @param centy         the centre in y (pixel coords)
 * @param angle         the rotation angle to derotate from (degrees ccw x-axis)
 * @param scale         the scaling (1.0 no change,<1 shrink,>1 enlarge)
 * @param filter_radius frequency filter radius (see sph_fft_freq_filter)
 * @param irdis_model    IRDIS model, set to NULL for all other instruments.
 * @return error code
 *
 * This routine will first recenter the master frame so that the
 * image centre corresponds to the given centre coordinates. For
 * even sized images, the image centre is between the 4 central pixels
 * for odd sized images it is the centre of the central pixel.
 * For example, if the image size is 20 x 20 pixels, giving centx and centy
 * as 10.0 and 10.0 will not result in a shift, but giving 8.5 and 11.5
 * will result in a shift of 1.5 pixels in x and -1.5 pixels in y.
 *
 * After centering, the distortion map is applied. This is skipped if
 * the distortionmap pointer is NULL.
 *
 * If the angle provided is not 0 the master frame is now rotated around the
 * image centre. If the scaling is not 1 the image is now scaled.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_transform_apply(
        sph_transform* self,
        sph_master_frame* mframe,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel)
{
    cpl_image*      bpix = NULL;
    cpl_ensure_code(mframe,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    bpix = sph_master_frame_get_badpixelmap(mframe);
    sph_master_frame_interpolate_bpix(mframe);
    cpl_ensure_code( sph_transform_apply_to_images(
            self,
            &mframe->image,
            &bpix,
            &mframe->ncombmap,
            &mframe->rmsmap,
            distortionmap,
            centx,centy,angle,scale,irdmodel) == CPL_ERROR_NONE,
            cpl_error_get_code() );
    sph_master_frame_set_bads_from_image(mframe,bpix);
    cpl_image_delete(bpix); bpix = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}
/*----------------------------------------------------------------------------*/
/**
 * @brief Center (shift), dedistort and rotate double image
 * @param doubleimage   input double image (change on output)
 * @param distortionmap the distortionmap to apply (maybe NULL)
 * @param centx         the centre in x (pixel coords)
 * @param centy         the centre in y (pixel coords)
 * @param angle         the rotation angle to derotate from (degrees ccw x-axis)
 * @param scale         the scaling (1.0 no change,<1 shrink,>1 enlarge)
 * @param filter_radius frequency filter radius (see sph_fft_freq_filter)
  * @return error code
 *
 * This routine is a wrapper for  sph_common_science_apply_transforms functions.
 * It will first be applied to the intensity master frame (iframe) of the given
 * double image and then to the polarization master frame (pframe) of the given
 * double image. See sph_common_science_apply_transforms for the detailed
 * transformation description.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_transform_apply_to_doubleimage(
        sph_transform* self,
        sph_double_image* doubleimage,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale)
{
    cpl_ensure_code(doubleimage->iframe,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(doubleimage->pframe,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    sph_transform_apply(self,doubleimage->iframe, distortionmap,
            centx, centy, angle, scale, NULL );
    sph_transform_apply(self,doubleimage->pframe, distortionmap,
            centx, centy, angle, scale, NULL );

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


/*----------------------------------------------------------------------------*/
/**
 * @brief Center (shift), dedistort and rotate master frame
 * @param image         input image (change on output, maybe NULL)
 * @param badpixel      input image (change on output, maybe NULL)
 * @param weights       input image (change on output, maybe NULL)
 * @param rms           input image (change on output, maybe NULL)
 * @param distortionmap the distortionmap to apply (maybe NULL)
 * @param centx         the centre in x (pixel coords)
 * @param centy         the centre in y (pixel coords)
 * @param angle         the rotation angle (degrees ccw x-axis)
 * @param scale         the scaling (1.0 no change,<1 shrink,>1 enlarge)
 * @param irdis_model    IRDIS model, set to NULL for all other instruments.
 * @return error code
 *
 * This routine will first recenter the images so that the
 * image centre corresponds to the given centre coordinates. For
 * even sized images, the image centre is between the 4 central pixels
 * for odd sized images it is the centre of the central pixel.
 * For example, if the image size is 20 x 20 pixels, giving centx and centy
 * as 10.0 and 10.0 will not result in a shift, but giving 8.5 and 11.5
 * will result in a shift of 1.5 pixels in x and -1.5 pixels in y.
 *
 * After centering, the distortion map is applied. This is skipped if
 * the distortionmap pointer is NULL.
 *
 * If the angle provided is not 0 the images are now rotated around the
 * image centre. If the scaling is not 1 the images is now scaled.
 *
 */
/*----------------------------------------------------------------------------*/
sph_error_code
sph_transform_apply_to_images(
        sph_transform* self,
        cpl_image** image,
        cpl_image** badpix,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel)
{

    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(image || badpix || ncombmap || rmsmap,CPL_ERROR_NULL_INPUT);

    if ( self->method == SPH_TRANSFORM_METHOD_FFT ) {
        __sph_transform_apply_to_images_fft(self,image,badpix,ncombmap,rmsmap,distortionmap,centx,centy,angle,scale,irdmodel);
    }
    else if ( self->method == SPH_TRANSFORM_METHOD_CPL_WARP ) {
        __sph_transform_apply_to_images_cpl_warp(self,image,badpix,ncombmap,rmsmap,distortionmap,centx,centy,angle,scale,irdmodel);
    }
    else {
        SPH_ERROR_RAISE_ERR(
                CPL_ERROR_UNSUPPORTED_MODE,
                "The transform method is not supported");
    }

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Remap the badpixels to rotated, shifted and distortion coords.
 * @param inbadbpix  the original badpixelmap
 * @param angle     the rotation angle
 * @param centx     rotation centre in pixel coordinates
 * @param centy     rotation centre in pixel coordinates
 * @param dx        shift in x
 * @param dy        shift in y
 * @param distmap   the polynomial distortion model to apply (maybe NULL)
 * @param irdis_model    IRDIS model, set to NULL for all other instruments.
 *
 * @return the new badpixel image or NULL on error.
 *
 * This function creates a new badpixel image of the same type as the input
 * image (usually CPL_TYPE_INT). It maps all pixels with values >= 1 to
 * a new image using the transformation specified (shift, rotation and,
 * if pointer to a model is given, distortion).
 * The transformation is purely geometric -- the pixel is transformed to
 * a polygon (rectangle) that is rotated and mapped onto the output image.
 * In the output image, all pixels that overlap the original bad pixel are
 * set to the value 1.
 *
 * The special input parameter "irdis_model" is should be set
 * if this function is to rotate a badpixel map from IRDIS in DBI or DPI mode,
 * as it correctly handles the badpixels at the edge of the images due to
 * the circular aperture. Set to NULL for any other instrument.
 *
 * Note that the function will not work correctly for
 * extreme distortions (with maximal distortions of more than a few pixels).
 *
 */
/*----------------------------------------------------------------------------*/
cpl_image*
sph_transform_remap_badpix(
        sph_transform* self,
        cpl_image* inbadpix,
        double angle, double centx,
        double centy,double dx0, double dy0,
        sph_polynomial_distortion_model* distmap,
        double      scale,
        sph_ird_instrument_model* irdmodel)
{
    cpl_vector*         badpixpos_x = NULL;
    cpl_vector*         badpixpos_y = NULL;
    cpl_image*          imout = NULL;
    cpl_image*          tmpim = NULL;
    int                 xx       = 0;
    int                 yy      = 0;
    int                 bpcount = 0;
    int                 bpix = 0;
    double              midx = 0.0;
    double              midy = 0.0;
    sph_polygon*        poly = NULL;
    sph_polygon*        poly2 = NULL;
    double              cosangle = cos(angle/180.0*CPL_MATH_PI);
    double              sinangle = sin(angle/180.0*CPL_MATH_PI);
    double              overlap = 0.0;
    int                 ok = 1;
    double              dx = dx0;
    double              dy = dy0;
    double              centimx = 0.0;
    double              centimy = 0.0;
    static int          __already_warned = 0;
    int                 bpactual = 0;
    sph_polygon_workspace*        ws   = NULL;

    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    cpl_ensure(inbadpix,CPL_ERROR_NULL_INPUT,NULL);
    cpl_ensure(self,CPL_ERROR_NULL_INPUT,NULL);
    imout = cpl_image_new(cpl_image_get_size_x(inbadpix),cpl_image_get_size_y(inbadpix),cpl_image_get_type(inbadpix));
    badpixpos_x = cpl_vector_new(SPH_TRANSFORM_MAX_BADPIX);
    badpixpos_y = cpl_vector_new(SPH_TRANSFORM_MAX_BADPIX);
    cpl_ensure(imout,CPL_ERROR_ILLEGAL_INPUT,NULL);

    centimx = cpl_image_get_size_x(inbadpix) / 2.0;
    centimy = cpl_image_get_size_y(inbadpix) / 2.0;

    for (xx = 0; xx < cpl_image_get_size_x(inbadpix); ++xx) {
        for (yy = 0; yy < cpl_image_get_size_y(inbadpix); ++yy) {
            ok = 1;
            if ( irdmodel ) {
                ok = sph_ird_instrument_model_check_in_dbi_fov(irdmodel, xx, yy, ok, 10.0);
            }
            if ( ok )
            {
                if ( cpl_image_get(inbadpix,xx+1,yy+1,&bpix) >= 1 ) {
                    cpl_vector_set(badpixpos_x,bpcount,xx * 1.0);
                    cpl_vector_set(badpixpos_y,bpcount,yy * 1.0);
                    bpcount++;
                    if ( bpcount >= SPH_TRANSFORM_MAX_BADPIX ) {
                        SPH_ERROR_RAISE_ERR(CPL_ERROR_ILLEGAL_INPUT,
                                "There are in total %d badpixels. "
                                "This recipe can not cope with such bad quality data."
                                "Please check your dark and flat quality.",
                                bpcount);
                        return NULL;
                    }
                }
            }
            else {
                // Dont do anything here -- pixels that are outside the
                // IRDIS FOV are not changed in their badpixel properties
                // and not rotated.
                cpl_image_set(imout,xx + 1, yy + 1, cpl_image_get(inbadpix,xx+1,yy+1,&bpix));
            }
        }
    }
    if ( bpcount >= SPH_TRANSFORM_OK_BADPIX && __already_warned == 0) {
        SPH_ERROR_RAISE_WARNING(SPH_ERROR_GENERAL,
                "There are in total %d badpixels. "
                "This is so large that the performance is suffering. "
        		"I will continue, but please check your dark and flat quality.",bpcount);
#pragma omp critical
        {
        	__already_warned = 1;
        }
    }
    if ( bpcount > 0 ) {
        cpl_vector_set_size(badpixpos_x,bpcount);
        cpl_vector_set_size(badpixpos_y,bpcount);
        poly = sph_polygon_new();
        sph_polygon_add_point(poly,0.0,0.0);
        sph_polygon_add_point(poly,0.0,1.0);
        sph_polygon_add_point(poly,1.0,1.0);
        sph_polygon_add_point(poly,1.0,0.0);
        poly2 = sph_polygon_new();
        sph_polygon_add_point(poly2,0.0,0.0);
        sph_polygon_add_point(poly2,0.0,1.0);
        sph_polygon_add_point(poly2,1.0,1.0);
        sph_polygon_add_point(poly2,1.0,0.0);

        ws = sph_polygon_workspace_new();

        for (bpix = 0; bpix < cpl_vector_get_size(badpixpos_x); ++bpix) {
            poly->points[0]->x = cpl_vector_get(badpixpos_x,bpix);
            poly->points[0]->y = cpl_vector_get(badpixpos_y,bpix);
            poly->points[1]->x = cpl_vector_get(badpixpos_x,bpix);
            poly->points[1]->y = cpl_vector_get(badpixpos_y,bpix) + 1 - 0.00000000001;
            poly->points[2]->x = cpl_vector_get(badpixpos_x,bpix) + 1 - 0.00000000001;
            poly->points[2]->y = cpl_vector_get(badpixpos_y,bpix) + 1 - 0.00000000001;
            poly->points[3]->x = cpl_vector_get(badpixpos_x,bpix) + 1 - 0.00000000001;
            poly->points[3]->y = cpl_vector_get(badpixpos_y,bpix);
            midx = sph_polygon_get_midx(poly);
            midy = sph_polygon_get_midy(poly);
            if ( distmap ) {
                sph_polygon_shift(poly,-centx,-centy);
                sph_polynomial_distortion_model_apply_poly(distmap,poly);
                sph_polygon_shift(poly,centx,centy);
            }
            if ( scale != 1.0 && scale != 0.0 ) {
                dx = dx0 + (midx-centx) * ( scale - 1.0);
                dy = dy0 + (midy-centy) * ( scale - 1.0);
            }
            sph_polygon_rotate_around(poly,centx,centy,cosangle,sinangle,dx,dy);
            midx = sph_polygon_get_midx(poly);
            midy = sph_polygon_get_midy(poly);
            for (xx = (int)midx-1; xx <= (int)midx+1; ++xx) {
                for (yy = (int)midy-1; yy <= (int)midy+1; ++yy) {
                    if (xx >= 0 &&
                            xx < cpl_image_get_size_x(imout) &&
                            yy >= 0 &&
                            yy < cpl_image_get_size_y(imout) ) {
                        poly2->points[0]->x = xx;
                        poly2->points[0]->y = yy;
                        poly2->points[1]->x = xx;
                        poly2->points[1]->y = yy + 1.0;
                        poly2->points[2]->x = xx + 1.0;
                        poly2->points[2]->y = yy + 1.0;
                        poly2->points[3]->x = xx + 1.0;
                        poly2->points[3]->y = yy;
                        overlap = sph_polygon_calculate_overlap(poly,poly2,ws);
                        if ( overlap > SPH_TRANSFORM_VAL_BADISBAD ) {
                            bpactual++;
                            cpl_ensure(cpl_image_set(imout,xx+1,yy+1,1.0)==CPL_ERROR_NONE,cpl_error_get_code(),NULL);
                        }
                    }
                }
            }
        }
        sph_polygon_workspace_delete(ws); ws = NULL;
    }
    cpl_vector_delete(badpixpos_x); badpixpos_x = NULL;
    cpl_vector_delete(badpixpos_y); badpixpos_y = NULL;
    sph_polygon_delete(poly); poly = NULL;
    sph_polygon_delete(poly2); poly2 = NULL;
    if (tmpim ) {
        cpl_image_delete(tmpim);
    }
    sph_polygon_free_all();
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return imout;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Delete the transform object and free up space.
 * @param self      the transform structure to delete
 */
/*----------------------------------------------------------------------------*/
void
sph_transform_delete(sph_transform* self) {
    if ( self ) cpl_free(self);
}

/*----------------------------------------------------------------------------*/
/**
 *  Static functons
 */
/*----------------------------------------------------------------------------*/
static
sph_transform* __sph_transform_new(void) {
    sph_transform*      result = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;

    result = cpl_calloc(1,sizeof(sph_transform));
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_NULL;
    return result;
}

static
sph_error_code
__sph_transform_apply_to_images_cpl_warp(
        sph_transform* self,
        cpl_image** image,
        cpl_image** badpix,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel)
{
    cpl_image*          displacement_x = NULL;
    cpl_image*          displacement_y = NULL;
    double              imcentx = 0.0;
    double              imcenty = 0.0;
    double              dx = 0.0;
    double              dy = 0.0;
    cpl_image*          tmpim = NULL;
    int                 nx = 0;
    int                 ny = 0;
    int                 xx  = 0;
    int                 yy  = 0;
    sph_point*           pix_cent = NULL;
    sph_point*           new_cent = NULL;
    double              cosangle = 0.0;
    double              sinangle = 0.0;
    double              v = 0.0;

    cpl_ensure_code(image || badpix || ncombmap || rmsmap,CPL_ERROR_NULL_INPUT);

    if ( image ) {
        nx = cpl_image_get_size_x(*image);
        ny = cpl_image_get_size_y(*image);
    }
    else if ( badpix ) {
        nx = cpl_image_get_size_x(*badpix);
        ny = cpl_image_get_size_y(*badpix);
    }
    else if ( ncombmap ) {
        nx = cpl_image_get_size_x(*ncombmap);
        ny = cpl_image_get_size_y(*ncombmap);
    }
    else if ( rmsmap ) {
        nx = cpl_image_get_size_x(*rmsmap);
        ny = cpl_image_get_size_y(*rmsmap);
    }

    imcentx = (double)nx;
    imcenty = (double)ny;
    imcentx /= 2.0;
    imcenty /= 2.0;
    dx = imcentx - centx;
    dy = imcenty - centy;


    if ( distortionmap ) {
        __sph_transform_distort(distortionmap,image,ncombmap,rmsmap);
    }
    displacement_x = cpl_image_new( nx, ny, CPL_TYPE_DOUBLE);
    displacement_y = cpl_image_new( nx, ny, CPL_TYPE_DOUBLE);
    if ( dx != 0.0 || dy != 0.0 ) {
        cpl_image_add_scalar(displacement_x,dx);
        cpl_image_add_scalar(displacement_y,dy);
        __sph_transform_apply_displacements_cpl(self,image,rmsmap,ncombmap,displacement_x,displacement_y);
        cpl_image_multiply_scalar(displacement_x,0.0);
        cpl_image_multiply_scalar(displacement_y,0.0);
    }

    if ( angle != 0.0 || scale != 1.0 ) {
        pix_cent = sph_point_new(0.0,0.0);
        new_cent = sph_point_new(0.0,0.0);
        cosangle = cos(1.0 * angle * CPL_MATH_RAD_DEG);
        sinangle = sin(1.0 * angle * CPL_MATH_RAD_DEG);
        for (xx = 0; xx < nx; ++xx) {
            for (yy = 0; yy < ny; ++yy) {
                new_cent->x = pix_cent->x = (double)xx + 0.5;
                new_cent->y = pix_cent->y = (double)yy + 0.5;
                sph_point_rotate_around(new_cent,imcentx,imcenty,cosangle,sinangle,0.0,0.0);
                new_cent->x = (new_cent->x - imcentx) / scale + imcentx;
                new_cent->y = (new_cent->y - imcenty) / scale + imcenty;
                v =  pix_cent->x - new_cent->x;
                cpl_image_set(displacement_x, xx + 1, yy + 1, v);
                v = pix_cent->y - new_cent->y;
                cpl_image_set(displacement_y, xx + 1, yy + 1, v);
            }
        }
        sph_point_delete(pix_cent);
        sph_point_delete(new_cent);
        __sph_transform_apply_displacements_cpl(self,image,rmsmap,ncombmap,displacement_x,displacement_y);
        cpl_image_multiply_scalar(displacement_x,0.0);
        cpl_image_multiply_scalar(displacement_y,0.0);
    }


    if ( badpix ) {

        tmpim = sph_transform_remap_badpix(self,*badpix,
                -1.0*angle,centx,centy,imcentx-centx,imcenty-centy,
                distortionmap,scale,irdmodel);

        cpl_image_delete(*badpix); *badpix = NULL;

        *badpix = tmpim;  tmpim = NULL;
    }

    cpl_image_delete(displacement_x); displacement_x = NULL;
    cpl_image_delete(displacement_y); displacement_y = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}



static
sph_error_code
__sph_transform_apply_to_images_fft(
        sph_transform* self,
        cpl_image** image,
        cpl_image** badpix,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        sph_polynomial_distortion_model* distortionmap,
        double centx,
        double centy,
        double angle,
        double scale,
        sph_ird_instrument_model* irdmodel)
{
    cpl_image*          tmpim = NULL;
    double              imcentx = 0.0;
    double              imcenty = 0.0;
    sph_fft*            fft = NULL;

    cpl_ensure_code(image || badpix || ncombmap || rmsmap,CPL_ERROR_NULL_INPUT);

    if ( image ) {
        imcentx = (double)cpl_image_get_size_x(*image);
        imcenty = (double)cpl_image_get_size_y(*image);
    }
    else if ( badpix ) {
        imcentx = (double)cpl_image_get_size_x(*badpix);
        imcenty = (double)cpl_image_get_size_y(*badpix);
    }
    else if ( ncombmap ) {
        imcentx = (double)cpl_image_get_size_x(*ncombmap);
        imcenty = (double)cpl_image_get_size_y(*ncombmap);
    }
    else if ( rmsmap ) {
        imcentx = (double)cpl_image_get_size_x(*rmsmap);
        imcenty = (double)cpl_image_get_size_y(*rmsmap);
    }

    imcentx /= 2.0;
    imcenty /= 2.0;

    fft = sph_fft_new(SPH_FFT_FFTW3_DP);

    __sph_transform_apply_filter(self,fft,image,ncombmap,rmsmap);

    if ( fabs(imcentx - centx) > 0.000001 ||
            fabs(imcenty - centy) > 0.000001 ) {
        __sph_transform_shift(self,
                fft,image,
                ncombmap,rmsmap,
                imcentx-centx,imcenty-centy);
    }
    if ( distortionmap ) {
        __sph_transform_distort(distortionmap,
                image,
                ncombmap,rmsmap);
    }

    if ( angle != 0.0 ) {
        __sph_transform_rotate(self,fft,image,ncombmap,rmsmap,angle);
    }
    if (scale != 1.0 ) {
        __sph_transform_scale(self,fft,image,ncombmap,rmsmap,scale);
    }

    if ( badpix ) {

        tmpim = sph_transform_remap_badpix(self,*badpix,
                -1.0*angle,centx,centy,imcentx-centx,imcenty-centy,
                distortionmap,scale,irdmodel);

        cpl_image_delete(*badpix); *badpix = NULL;

        *badpix = tmpim;  tmpim = NULL;
    }

    sph_fft_delete(fft); fft = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_transform_apply_displacements_cpl(
        sph_transform* self,
        cpl_image** image,
        cpl_image** rmsmap,
        cpl_image** ncombmap,
        cpl_image* displacement_x,
        cpl_image* displacement_y)
{
    cpl_vector*         profile = NULL;
    cpl_image*          tmpim = NULL;
    cpl_image*          correction_map = NULL;
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    profile = cpl_vector_new( 2 * (int)self->radius);
    cpl_vector_fill_kernel_profile(profile,self->kerneltype,self->radius);

    if ( image ) {
        tmpim = cpl_image_new( cpl_image_get_size_x( *image ),
                cpl_image_get_size_y(*image), cpl_image_get_type(*image));
        cpl_image_warp(tmpim,*image,displacement_x, displacement_y, profile, self->radius, profile, self->radius);
        cpl_image_delete(*image);*image = NULL;
        correction_map = cpl_image_new(cpl_image_get_size_x(tmpim),
                                        cpl_image_get_size_y(tmpim),
                                        cpl_image_get_type(tmpim));
        cpl_image_fill_jacobian(correction_map, displacement_x, displacement_y);
        *image = cpl_image_multiply_create(tmpim, correction_map);
        cpl_image_delete(tmpim); tmpim = NULL;
        cpl_image_delete(correction_map); correction_map = NULL;
    }
    if ( rmsmap ) {
        tmpim = cpl_image_new( cpl_image_get_size_x( *rmsmap ),
                cpl_image_get_size_y(*rmsmap), cpl_image_get_type(*rmsmap));
        cpl_image_warp(tmpim,*rmsmap,displacement_x, displacement_y, profile, self->radius, profile, self->radius);
        cpl_image_delete(*rmsmap);*rmsmap = NULL;
        *rmsmap = tmpim; tmpim = NULL;
    }
    if ( ncombmap ) {
        tmpim = cpl_image_new( cpl_image_get_size_x( *ncombmap ),
                cpl_image_get_size_y(*ncombmap), cpl_image_get_type(*ncombmap));
        cpl_image_warp(tmpim,*ncombmap,displacement_x, displacement_y, profile, self->radius, profile, self->radius);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    cpl_vector_delete(profile); profile = NULL;
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}




static
sph_error_code __sph_transform_apply_filter( sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap)
{
    cpl_image*          tmpim = NULL;

    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(fft,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    if ( self->filt_method == SPH_FFT_FILTER_METHOD_TH ) {
        tmpim = sph_fft_freq_filter(fft,*image,1,self->fft_filter_radius,0.0,0.0);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
        if ( *rmsmap ) {
            tmpim = sph_fft_freq_filter(fft,*rmsmap,1,self->fft_filter_radius,0.0,0.0);
            cpl_image_delete(*rmsmap);*rmsmap = NULL;
            *rmsmap = tmpim; tmpim = NULL;
        }
        tmpim = sph_fft_freq_filter(fft,*ncombmap,1,self->fft_filter_radius,0.0,0.0);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( self->filt_method == SPH_FFT_FILTER_METHOD_FERMI ) {
        tmpim = sph_fft_freq_filter(fft,*image,1,
                self->fft_filter_radius,self->fft_fermi_temp,0.0);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
        if ( *rmsmap ) {
            tmpim = sph_fft_freq_filter(fft,*rmsmap,1,
                    self->fft_filter_radius,self->fft_fermi_temp,0.0);
            cpl_image_delete(*rmsmap);*rmsmap = NULL;
            *rmsmap = tmpim; tmpim = NULL;
        }
        tmpim = sph_fft_freq_filter(fft,*ncombmap,1,
                self->fft_filter_radius,self->fft_fermi_temp,0.0);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( self->filt_method == SPH_FFT_FILTER_METHOD_BUTTER ) {
        tmpim = sph_fft_freq_filter(fft,*image,1,
                0.0,self->fft_butter_pass,self->fft_butter_stop);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
        if ( *rmsmap ) {
            tmpim = sph_fft_freq_filter(fft,*rmsmap,1,
                    0.0,self->fft_butter_pass,self->fft_butter_stop);
            cpl_image_delete(*rmsmap);*rmsmap = NULL;
            *rmsmap = tmpim; tmpim = NULL;
        }
        tmpim = sph_fft_freq_filter(fft,*ncombmap,1,
                0.0,self->fft_butter_pass,self->fft_butter_stop);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_transform_shift(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double dx,
        double dy)
{
    cpl_image*          tmpim = NULL;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(fft,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    if ( image ) {
        tmpim = sph_fft_shift_image(fft,*image,dx,dy);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
    }
    if ( ncombmap) {
        tmpim = sph_fft_shift_image(fft,*ncombmap,dx,dy);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( rmsmap ) {
        tmpim = sph_fft_shift_image(fft,*rmsmap,dx,dy);
        cpl_image_delete(*rmsmap);*rmsmap = NULL;
        *rmsmap = tmpim; tmpim = NULL;
    }

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}


static
sph_error_code
__sph_transform_scale(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double scale)
{
    cpl_image*          tmpim = NULL;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(fft,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    if ( image ) {
        tmpim = sph_fft_scale(fft,*image,scale);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
    }
    if ( ncombmap ) {
        tmpim = sph_fft_scale(fft,*ncombmap,scale);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( rmsmap ) {
        tmpim = sph_fft_scale(fft,*rmsmap,scale);
        cpl_image_delete(*rmsmap);*rmsmap = NULL;
        *rmsmap = tmpim; tmpim = NULL;
    }

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_transform_rotate(
        sph_transform* self,
        sph_fft* fft,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap,
        double angle)
{
    cpl_image*          tmpim = NULL;
    cpl_ensure_code(self,CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(fft,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;

    if ( image ) {
        tmpim = sph_fft_rotate_image(fft,*image,-1.0*angle);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
    }
    if ( ncombmap ) {
        tmpim = sph_fft_rotate_image(fft,*ncombmap,-1.0*angle);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( rmsmap ) {
        tmpim = sph_fft_rotate_image(fft,*rmsmap,-1.0 * angle);
        cpl_image_delete(*rmsmap);*rmsmap = NULL;
        *rmsmap = tmpim; tmpim = NULL;
    }

    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

static
sph_error_code
__sph_transform_distort( sph_polynomial_distortion_model* distortionmap,
        cpl_image** image,
        cpl_image** ncombmap,
        cpl_image** rmsmap)
{
    cpl_image*          tmpim = NULL;
    cpl_ensure_code(distortionmap,CPL_ERROR_NULL_INPUT);
    SPH_ERROR_CHECK_STATE_ONERR_RETURN_ERRCODE;
    if ( image ) {
        tmpim = sph_polynomial_distortion_model_apply(distortionmap,
                *image,
                CPL_KERNEL_DEFAULT,
                CPL_KERNEL_DEF_WIDTH);
        cpl_image_delete(*image);*image = NULL;
        *image = tmpim; tmpim = NULL;
    }
    if ( ncombmap ) {
        tmpim = sph_polynomial_distortion_model_apply(distortionmap,
                *ncombmap,
                CPL_KERNEL_DEFAULT,
                CPL_KERNEL_DEF_WIDTH);
        cpl_image_delete(*ncombmap);*ncombmap = NULL;
        *ncombmap = tmpim; tmpim = NULL;
    }
    if ( rmsmap ) {
        tmpim = sph_polynomial_distortion_model_apply(distortionmap,
                *rmsmap,
                CPL_KERNEL_DEFAULT,
                CPL_KERNEL_DEF_WIDTH);
        cpl_image_delete(*rmsmap);*rmsmap = NULL;
        *rmsmap = tmpim; tmpim = NULL;
    }
    SPH_ERROR_CHECK_STATE_RETURN_ERRCODE;
}

/**@}*/
