/* $Id: $
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/
#include <Basic.h>
#include <CUnit.h>
#include "sph_error.h"
#include "sph_test.h"
#include "sph_filter_table.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup A CUnit Test Suite -- representing a collection of testcases
 * @par Synopsis:
 * @code
 * @endcode
 * @par Desciption:
 *
 * This module provides a collection of tests for one specific, distinct
 * module or set-up. The testing code is implemented using the CUnit
 * framework.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*----------------------------------------------------------------------------*/
/*-                            INTERNAL HELPER FUNCTIONS                          */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to initiailise the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_init_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/
    sph_init_test(__FILE__,"moeller@mpia-hd.mpg.de",CPL_MSG_DEBUG);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Function to clean the unit test suite
 */
/*----------------------------------------------------------------------------*/
static
int cutest_clean_testsuite() {
    sph_error_dump( SPH_ERROR_ERROR );
    return sph_end_test();
}
/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_filter_table_new(void) {
    sph_filter_table*        filter_table = NULL;

    filter_table = sph_filter_table_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    CU_ASSERT_EQUAL(sph_filter_table_get_size(filter_table),0);
    sph_filter_table_delete(filter_table); filter_table = NULL;
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_filter_table_add_row(void) {
    sph_filter_table*        filter_table = NULL;

    filter_table = sph_filter_table_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    sph_filter_table_add_row(filter_table,"FILTER1",2.0,3.0);
    CU_ASSERT_EQUAL(sph_filter_table_get_size(filter_table),1);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->left_lambda,0),2.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->right_lambda,0),3.0,0.00001);
    CU_ASSERT_STRING_EQUAL(filter_table->filterNames[0],"FILTER1");
    sph_filter_table_delete(filter_table); filter_table = NULL;
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_filter_table_get(void) {
    sph_filter_table*        filter_table = NULL;
    cpl_vector*         angles = NULL;
    cpl_vector*         xc = NULL;
    cpl_vector*         yc = NULL;
    filter_table = sph_filter_table_new();
    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    sph_filter_table_add_row(filter_table,"FILTER1",2.0,3.0);

    xc = sph_filter_table_get_left_lambdas(filter_table);
    CU_ASSERT_PTR_NOT_NULL_FATAL(xc);
    yc = sph_filter_table_get_right_lambdas(filter_table);
    CU_ASSERT_PTR_NOT_NULL_FATAL(yc);
    CU_ASSERT_EQUAL(cpl_vector_get_size(xc),1);
    CU_ASSERT_EQUAL(cpl_vector_get_size(yc),1);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(xc,0),2.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(yc,0),3.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(sph_filter_table_get_left_lambda(filter_table,"FILTER1"),2.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(sph_filter_table_get_right_lambda(filter_table,"FILTER1"),3.0,0.00001);

    cpl_vector_delete(xc); xc = NULL;
    cpl_vector_delete(yc); yc = NULL;
    sph_filter_table_delete(filter_table); filter_table = NULL;
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_filter_table_load_ascii(void) {
    sph_filter_table*        filter_table = NULL;
    FILE*               f = NULL;

    f = fopen("filter_table.txt","w");
    fprintf(f,"%s\n","# Hello this is a comment");
    fprintf(f,"%s\n","# Hello this is a comment too");
    fprintf(f,"  FILTER1     2.0      11.0\n");
    fprintf(f,"  FILTER2     22.0      111.0\n");
    fprintf(f,"  FILTER3     222.0      1111.0\n");
    fclose(f);
    filter_table = sph_filter_table_load_ascii("filter_table.txt");
    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    CU_ASSERT_EQUAL(sph_filter_table_get_size(filter_table),3);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->left_lambda,0),2.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->right_lambda,0),11.0,0.00001);
    CU_ASSERT_STRING_EQUAL(filter_table->filterNames[0],"FILTER1");
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->left_lambda,1),22.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->right_lambda,1),111.0,0.00001);
    CU_ASSERT_STRING_EQUAL(filter_table->filterNames[1],"FILTER2");
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->left_lambda,2),222.0,0.00001);
    CU_ASSERT_DOUBLE_EQUAL(cpl_vector_get(filter_table->right_lambda,2),1111.0,0.00001);
    CU_ASSERT_STRING_EQUAL(filter_table->filterNames[2],"FILTER3");

    sph_filter_table_delete(filter_table); filter_table = NULL;
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief        A test function of the testsuite
 */
/*----------------------------------------------------------------------------*/
static
void cutest_sph_filter_table_save_ascii(void) {
    sph_filter_table*        filter_table = NULL;
    filter_table = sph_filter_table_new();

    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    sph_filter_table_add_row(filter_table,"FILTER1",1.0,2.0);
    sph_filter_table_add_row(filter_table,"FILTER2",22.0,33.0);
    sph_filter_table_add_row(filter_table,"FITLER3",222.0,333.0);
    CU_ASSERT_EQUAL(sph_filter_table_save_ascii(filter_table,"filter_table.txt"),CPL_ERROR_NONE);
    sph_filter_table_delete(filter_table); filter_table = NULL;

    filter_table = sph_filter_table_load_ascii("filter_table.txt");
    CU_ASSERT_PTR_NOT_NULL_FATAL(filter_table);
    CU_ASSERT_EQUAL(sph_filter_table_get_size(filter_table),3);

    sph_filter_table_delete(filter_table); filter_table = NULL;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test MAIN function
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {
    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("A testsuite", cutest_init_testsuite,
                          cutest_clean_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "Testing constructor",
                             cutest_sph_filter_table_new ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Adding row",
                             cutest_sph_filter_table_add_row ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Loading from ASCII",
                             cutest_sph_filter_table_load_ascii ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "get vectors",
                             cutest_sph_filter_table_get ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "Saving to ASCII",
                             cutest_sph_filter_table_save_ascii ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
