/* $Id: $
 *
 * This file is part of the SPHERE Pipeline
 * Copyright (C) 2007-2010 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: $
 * $Date: $
 * $Revision: $
 * $Name: $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
 Includes
 -----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <cpl.h>

#include "sph_image_grid.h"
#include "sph_common_keywords.h"
#include <math.h>
#include "sph_error.h"
#include "sph_test.h"
#include <Basic.h>
#include <CUnit.h>
#include "sph_test.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup techcal_master_test  Unit test of techcal_master recipe and
 *                                  associated functions.
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/
int cutest_init_image_grid_testsuite() {
    /*--------------------------------------------------------------------
     * -    Prepare CPL and error logging
     * -------------------------------------------------------------------*/

    #if defined CPL_VERSION_CODE && CPL_VERSION_CODE >= CPL_VERSION(4, 0, 0)
        cpl_init(CPL_INIT_DEFAULT);
    #else
        cpl_init();
    #endif

    cpl_msg_set_time_on();
    cpl_msg_set_component_on();
    cpl_msg_set_domain("Testing");
    cpl_msg_set_domain_on();
    cpl_msg_set_level(CPL_MSG_OFF);
    cpl_msg_set_log_level(CPL_MSG_DEBUG);

    sph_error_new();
    return 0;
}

int cutest_clean_image_grid_testsuite() {
    sph_error_dump( SPH_ERROR_ERROR );
    cpl_end();
    return 0;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_new() {
    sph_image_grid*          p   =   NULL;

    p = sph_image_grid_new( );

    CU_ASSERT_PTR_NOT_NULL_FATAL( p );


    CU_ASSERT_EQUAL( p->nx, 0 );
    CU_ASSERT_EQUAL( p->ny, 0 );

    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_rotate_and_shift() {
    sph_image_grid*         p       = NULL;
    sph_image_grid*         target       = NULL;
    cpl_image*                image     = NULL;
    cpl_image*                timage     = NULL;
    cpl_image*                result_image = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_error_code            rerr    = 0;
    sph_polygon*            poly    = NULL;

    image = cpl_image_new( 48, 48, CPL_TYPE_DOUBLE );
    timage = cpl_image_new( 48, 48, CPL_TYPE_DOUBLE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( image );
    CU_ASSERT_PTR_NOT_NULL_FATAL( timage );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            cpl_image_set( image, xx + 1,yy +1, 1.2 );
        }
    }
    p = sph_image_grid_new_from_image( image, 12.0, 260.0,
                                        12.0, 260.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );
    target = sph_image_grid_new_from_image( timage, 0.0, 248.0,
                                        0.0, 248.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_image_grid_rotate_and_shift( target, 25.0, 25.0, 15.0, 00.0, 00.0 );
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    rerr = sph_image_grid_map_to_target( p, target );
    CU_ASSERT_EQUAL_FATAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->polys );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->values );

    CU_ASSERT_EQUAL( target->nx, 48 );
    CU_ASSERT_EQUAL( target->ny, 48 );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            poly = target->polys[ yy * target->nx + xx ];
            CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
            CU_ASSERT_EQUAL( poly->npoints, 4 );
        }
    }
    result_image = sph_image_grid_extract_image( target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result_image );
    cpl_image_save( result_image, "rotated.fits", CPL_TYPE_DOUBLE, NULL, CPL_IO_DEFAULT );
    CU_ASSERT_DOUBLE_EQUAL( cpl_image_get_flux(image), cpl_image_get_flux(result_image), 0.01 );
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_mark_point function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_mark_point() {
    sph_image_grid*         target       = NULL;
    cpl_image*                timage     = NULL;
    cpl_image*                result_image = NULL;
    double                    tv        = 0.0;
    int                        bad        = 0;
    sph_error_code            rerr    = CPL_ERROR_NONE;

    timage = cpl_image_new( 64, 64, CPL_TYPE_DOUBLE );
    target = sph_image_grid_new_from_image( timage, 0.0, 16.0,
                                        0.0, 16.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_image_grid_mark_point( target, 8.0, 8.0, 12.0 );
    result_image = sph_image_grid_extract_image( target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result_image );
    tv = cpl_image_get( result_image, 33, 33, &bad );
    CU_ASSERT_EQUAL( bad, 0 );
    CU_ASSERT_DOUBLE_EQUAL( tv, 12.0, 0.0000000000001 );
    cpl_image_delete( result_image );

    return;

}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_get_xy function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_get_xy() {
    sph_image_grid*         target       = NULL;
    cpl_image*                timage     = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_error_code            rerr    = CPL_ERROR_NONE;

    timage = cpl_image_new( 64, 64, CPL_TYPE_DOUBLE );
    target = sph_image_grid_new_from_image( timage, 0.0, 16.0,
                                        0.0, 16.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    rerr = sph_image_grid_get_xy( target, 0.000001, 0.00001, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 0 );
    CU_ASSERT_EQUAL( yy, 0 );
    rerr = sph_image_grid_get_xy( target, 8.00000, 8.000, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 32 );
    CU_ASSERT_EQUAL( yy, 32 );
    rerr = sph_image_grid_get_xy( target, 7.99999999, 0.0, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 31 );
    CU_ASSERT_EQUAL( yy, 0 );
    rerr = sph_image_grid_get_xy( target, 7.99999999, 7.9999999, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 31 );
    CU_ASSERT_EQUAL( yy, 31 );
    sph_image_grid_rotate_and_shift( target, 0.0, 0.0, -90.0, 0.0, 0.0 );
    rerr = sph_image_grid_get_xy( target, 7.99999999, -7.9999999, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 31 );
    CU_ASSERT_EQUAL( yy, 31 );
    sph_image_grid_rotate_and_shift( target, 0.0, 0.0, 0.0, 8.0, 0.0 );
    rerr = sph_image_grid_get_xy( target, 8.000000, -8.00000, &xx, &yy );
    CU_ASSERT_EQUAL( rerr, CPL_ERROR_NONE );
    CU_ASSERT_EQUAL( xx, 32 );
    CU_ASSERT_EQUAL( yy, 0 );
    sph_image_grid_delete( target );

    return;

}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_map_polygon function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_map_polygon() {
    sph_image_grid*         target       = NULL;
    cpl_image*                timage     = NULL;
    cpl_image*                result_image = NULL;
    sph_polygon*            poly    = NULL;

    timage = cpl_image_new( 48, 48, CPL_TYPE_DOUBLE );
    target = sph_image_grid_new_from_image( timage, -8.0, 8.0,
                                        0.0, 16.0 );
   sph_image_grid_rotate_and_shift( target, -8.0, 0.0, -45.0, 0.0, 0.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );


    poly = sph_polygon_new();
    sph_polygon_add_point( poly, -4.0, -4.0 );
    sph_polygon_add_point( poly, 4.0, -4.0 );
    sph_polygon_add_point( poly, 7.0, 0.0 );
    sph_polygon_add_point( poly, 4.0, 4.0 );
    sph_polygon_add_point( poly, -4.0, 4.0 );
    sph_polygon_add_point( poly, -7.0, 0.0 );
    sph_image_grid_map_polygon( target, poly, 10.0 );
    result_image = sph_image_grid_extract_image( target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result_image );
    cpl_image_save( result_image, "mappedpoly.fits", CPL_TYPE_DOUBLE, NULL, CPL_IO_DEFAULT );
    sph_image_grid_delete( target );
    return;

}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_map_to_target() {
    sph_image_grid*         p       = NULL;
    sph_image_grid*         target       = NULL;
    cpl_image*                image     = NULL;
    cpl_image*                timage     = NULL;
    cpl_image*                result_image = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_polygon*            poly    = NULL;

    image = cpl_image_new( 48, 48, CPL_TYPE_DOUBLE );
    timage = cpl_image_new( 48, 48, CPL_TYPE_DOUBLE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( image );
    CU_ASSERT_PTR_NOT_NULL_FATAL( timage );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            cpl_image_set( image, xx + 1,yy +1, 1.2 );
        }
    }
    p = sph_image_grid_new_from_image( image, 12.0, 260.0,
                                        12.0, 260.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );
    target = sph_image_grid_new_from_image( timage, 0.0, 248.0,
                                        0.0, 248.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );

    sph_image_grid_map_to_target( p, target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->polys );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->values );

    CU_ASSERT_EQUAL( target->nx, 48 );
    CU_ASSERT_EQUAL( target->ny, 48 );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            poly = target->polys[ yy * target->nx + xx ];
            CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
            CU_ASSERT_EQUAL( poly->npoints, 4 );
        }
    }
    result_image = sph_image_grid_extract_image( target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result_image );
    cpl_image_save( result_image, "mapped.fits", CPL_TYPE_DOUBLE, NULL, CPL_IO_DEFAULT );
    sph_image_grid_delete( p );
    sph_image_grid_delete( target );
    return;
}

void cutest_image_grid_map_to_target_caseB() {
    sph_image_grid*         p       = NULL;
    sph_image_grid*         target       = NULL;
    cpl_image*                image     = NULL;
    cpl_image*                result_image = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_polygon*            poly    = NULL;

    image = cpl_image_new( 50, 50, CPL_TYPE_DOUBLE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( image );
    target = sph_image_grid_new_from_image( image, 1.0, 11.0,
                                        1.0, 11.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( target );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            cpl_image_set( image, xx + 1,yy +1, 1.2 );
        }
    }
    p = sph_image_grid_new_from_image( image, 0.0, 10.0,
                                        0.0, 10.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );

    sph_image_grid_map_to_target( p, target );

    CU_ASSERT_EQUAL( target->nx, 50 );
    CU_ASSERT_EQUAL( target->ny, 50 );
    for (xx = 10; xx < 28; ++xx) {
        for (yy = 10; yy < 28; ++yy) {
            poly = target->polys[ yy * target->nx + xx ];
            CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
            CU_ASSERT_EQUAL( poly->npoints, 4 );
        }
    }
    result_image = sph_image_grid_extract_image( target );
    CU_ASSERT_PTR_NOT_NULL_FATAL( result_image );
    cpl_image_save( result_image, "mappedB.fits", CPL_TYPE_DOUBLE, NULL, CPL_IO_DEFAULT );
    sph_image_grid_delete( target );
    sph_image_grid_delete( p );
    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_new_from_image() {
    sph_image_grid*         p       = NULL;
    cpl_image*                image     = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_polygon*            poly    = NULL;

    image = cpl_image_new( 248, 248, CPL_TYPE_DOUBLE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( image );
    for (xx = 20; xx < 228; ++xx) {
        for (yy = 20; yy < 228; ++yy) {
            cpl_image_set( image, xx + 1,yy +1, 1.2 );
        }
    }
    p = sph_image_grid_new_from_image( image, 12.0, 260.0,
                                        12.0, 260.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );

    CU_ASSERT_PTR_NOT_NULL_FATAL( p->polys );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->values );

    CU_ASSERT_EQUAL( p->nx, 248 );
    CU_ASSERT_EQUAL( p->ny, 248 );
    for (xx = 20; xx < 228; ++xx) {
        for (yy = 20; yy < 228; ++yy) {
            poly = p->polys[ yy * p->nx + xx ];
            CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
            CU_ASSERT_EQUAL( poly->npoints, 4 );
        }
    }
    CU_ASSERT_DOUBLE_EQUAL( p->polys[0]->points[0]->x, 12.0, 1.0e-15 );
    CU_ASSERT_DOUBLE_EQUAL( p->polys[0]->points[0]->y, 12.0, 1.0e-15 );
    return;
}
/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_new function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_extract_image() {
    sph_image_grid*         p       = NULL;
    cpl_image*                image     = NULL;
    cpl_image*                result    = NULL;
    int                        xx        = 0;
    int                        yy        = 0;
    sph_polygon*            poly    = NULL;
    double                    sum        = 0.0;

    image = cpl_image_new( 248, 248, CPL_TYPE_DOUBLE );
    CU_ASSERT_PTR_NOT_NULL_FATAL( image );
    for (xx = 20; xx < 228; ++xx) {
        for (yy = 20; yy < 228; ++yy) {
            cpl_image_set( image, xx + 1,yy +1, 1.2 );
        }
    }
    p = sph_image_grid_new_from_image( image, 12.0, 260.0,
                                        12.0, 260.0 );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p );
    sum = cpl_image_get_flux( image );
    CU_ASSERT_DOUBLE_EQUAL( sum, 208 * 208 * 1.2, 1.0e-6 );

    CU_ASSERT_PTR_NOT_NULL_FATAL( p->polys );
    CU_ASSERT_PTR_NOT_NULL_FATAL( p->values );

    CU_ASSERT_EQUAL( p->nx, 248 );
    CU_ASSERT_EQUAL( p->ny, 248 );
    for (xx = 20; xx < 228; ++xx) {
        for (yy = 20; yy < 228; ++yy) {
            poly = p->polys[ yy * p->nx + xx ];
            CU_ASSERT_PTR_NOT_NULL_FATAL( poly );
            CU_ASSERT_EQUAL( poly->npoints, 4 );
        }
    }
    CU_ASSERT_DOUBLE_EQUAL( p->polys[0]->points[0]->x, 12.0, 1.0e-6 );
    CU_ASSERT_DOUBLE_EQUAL( p->polys[0]->points[0]->y, 12.0, 1.0e-6 );

    result = sph_image_grid_extract_image( p );
    CU_ASSERT_EQUAL( cpl_image_get_size_x(result),248);
    CU_ASSERT_EQUAL( cpl_image_get_size_y(result),248);
    sum = cpl_image_get_flux( result );
    CU_ASSERT_DOUBLE_EQUAL( sum, 208 * 208 * 1.2, 1.0e-7 );
    cpl_image_subtract( result, image );
    sum = cpl_image_get_flux( result );
    CU_ASSERT_DOUBLE_EQUAL( sum, 0.0, 1.0e-12 );
    return;
}


/*----------------------------------------------------------------------------*/
/**
 @brief    Unit test for the sph_image_grid_delete function.
 */
/*----------------------------------------------------------------------------*/
void cutest_image_grid_delete() {
    sph_image_grid*          p   = NULL;
    p = sph_image_grid_new( );

    CU_ASSERT_PTR_NOT_NULL_FATAL( p );

    sph_image_grid_delete( p );

    return;
}

/*----------------------------------------------------------------------------*/
/**
 @brief    Unit tests of techcal_master_dark recipe and associated functions
 */
/*----------------------------------------------------------------------------*/
int main(int argc, char* argv[]) {

    int         result      = 0;
    CU_pSuite   pSuite      = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
       return CU_get_error();


    /* add a suite to the registry */
    pSuite = CU_add_suite("image_grid_test", cutest_init_image_grid_testsuite,
                          cutest_clean_image_grid_testsuite );
    if (NULL == pSuite) {
       CU_cleanup_registry();
       return CU_get_error();
    }


    /* add the tests to the suite */
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_new",
                             cutest_image_grid_new ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_new_from_image",
                             cutest_image_grid_new_from_image ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_extract_image",
                             cutest_image_grid_extract_image ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_get_xy",
                             cutest_image_grid_get_xy ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_mark_point",
                             cutest_image_grid_mark_point ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    if ( NULL == CU_add_test(pSuite, "sph_image_grid_map_to_target",
                             cutest_image_grid_map_to_target ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_map_polygon",
                             cutest_image_grid_map_polygon ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_map_to_target_caseB",
                             cutest_image_grid_map_to_target_caseB ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_rotate_and_shift",
                             cutest_image_grid_rotate_and_shift ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }
    if ( NULL == CU_add_test(pSuite, "sph_image_grid_delete",
                             cutest_image_grid_delete ) )
    {
       CU_cleanup_registry();
       return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode( CU_BRM_VERBOSE );
    CU_basic_run_tests();
    sph_test_create_results_file( "results.txt" );
    result = sph_test_check_result();
    CU_cleanup_registry();
    return result;
}

/**@}*/
